unit BDEFiltr;
(*

This code shows how compiled Delphi code may be used to determine inclusion or
exclusion of records in a data table.  It is quite fast.  This code is based on
information downloaded from the Compuserve BDEVTOOL forum in the file
CBFILTER.ZIP.

This sample will only set one filter active at a time, and only disables that
filter.  You should also be aware that you could set as many code filters as you
wish (although logic limits your filtering) and either clear them all by passing
nil as the filter handle to DbiDropFilter(), or clear a specific one as I have
in this code sample.

I am currently using DBDEMOS and the CUSTOMER.DB for the test. However, you
should be able to use any Paradox or dBASE table with this code to do timing
tests or determine that it does indeed work.

Subjectively, implementing a filter via code appears to be extremely fast.  I
would be interested in finding out exact timings from anyone who determines the
actual performance differences between an interpreted filter and the code one.

When I was designing a component to facilitate this feature, I found that
Delphi's TDataset component already implemented this feature in a slightly
different fashion with its OnFilterRecord event.

Although this code compiles and runs with TQuery components, it appears that
the filter callback is ignored for it.  The documentation says that Filter
callbacks don't work with pass-through SQL cursors, but it evidently doesn't
work with local SQL tables either.

John F. Kaster
Compuserve 71510,3321
Internet jfk@earthlink.net
http:/www.earthlink.net/~jfk

Below is the original text file downloaded from CompuServe that was the basis of
the information used to create this demo project, with spaces added to prevent
the characters '*' and ')' from being the end-of-comment characters in the
inappropriate place.

Callback filters with BDE

Have you needed to filter information retrieved from BDE by criteria that
prevents you from using a normal filter, or are you simply tired of
creating overly complex filter trees?

During the development of one of my projects, I ran into a situation that
really presented a problem for me.  I needed to display the contents of a
table in a grid based upon the criteria that the user entered.  Not a really
big problem until I realized that the criteria that the user was searching for
could be anywhere in the specified field, not just the beginning.  In addition,
the table needed to be sorted and displayed by another indexed field.  While
it's pretty easy to create a BDE filter that limits the rows retrieved by the
beginning of the field, it's impossible to add a filter that limits information
by something in the middle of the field.

While I could use SQL against the table to retrieve the desired information,
I'm using dBASE tables stored on a network drive.  A good solution for small
tables, but the performance was too poor for distribution once the table
got larger then 20 or 25 thousand rows.

Looking through the BDE documentation provided no possible solution to my
problem until I stumbled upon the DbiAddFilter function declaration in the
IDAPI.H file.  Surprisingly enough, there is an undocumented option to add
a user defined callback function.  Using this undocumented option, you can
create your own filters that allow you to filter by just about any criteria
that you want.  All you have to do is tell the BDE if the current record
should be returned to the application by returning TRUE (1) or not by returning
FALSE (0).

The steps for implementing this are simple.

1. Create a structure to pass information to the callback function.  Make
  sure that you include the cursor to the table or you won't be able to
  retrieve field information.
2. Create a callback function for the BDE to call.  See the declaration below
  for SelectionCallback for the layout.  You can name this callback function
  anything that you want.
3. In your callback function, cast the ulClientData parameter to a pointer
  of your defined callback structure.
4. Create your comparison code.  Return TRUE (1) for each record that you
  want to return to your program, or FALSE (0) for records that you don't want
  returned.
5. In the body of your program, instantiate an instance of your callback
  structure.
6. Populate your callback structure with the needed information.  don't forget
  to pass the cursor handle.
7. Create a hDBIFilter handle.
8. Call DbiAddFilter with a pointer to your callback structure as the second
  parameter.  Cast this pointer to a UINT32.  The 6th parameter should be
  the name of the callback function, instead of the 0 indicated in the BDE
  documentation.  Make sure that the 5th parameter reserved for the filter
  expression is 0 or your callback function won't be called.
9. Activate your filter.

That's pretty much all there is too it.

The following is an example of a callback filter that searches the 2nd column
in a table for the word 'RAM'.  This would allow me to return records to my
app that have fields that contain strings like 'RAM Chip', '1 Meg RAM chip', or
'STEALTH-24, 2MB RAM'.  Normal BDE filters would only allow me to retrieve the
first example.


//beginning of example

//declarations
struct CallbackInfo {
	hDBICur hCur;
	char Criteria[50];
	};
INT16 DBIFN SelectionCallback(UINT32 ulClientData, pBYTE pRecBuf, UINT32 iPhyRecNum);

//implementation
INT16 DBIFN SelectionCallback(UINT32 ulClientData, pBYTE pRecBuf, UINT32 iPhyRecNum)
{
char TempInfo[255], szTemp[255];  //create working storage buffers
CallbackInfo *myCBInfo = ( CallbackInfo* )ulClientData;  //get a usable pointer from the client data
BOOL isBlank;

DBIResult Result = DbiGetField(myCBInfo->hCur, 2, pRecBuf, (pBYTE) szTemp,
   &isBlank);  //get the field
if(!Result && !isBlank) strcpy(TempInfo, szTemp);  //if it's not blank, copy the field over
else strcpy(TempInfo, "");  //otherwise blank out the comparison field

if(!strstr(TempInfo, myCBInfo->Criteria)) return FALSE;  //could not find the field, return false
return TRUE;  //match found, return true to indicate that the record should be returned
}

/*main body of the package... This assumes that the table you want to filter
  has already been opened using a cursor handle called hCur;
.
.
.
hDBIFilter secondaryFilter;        // create a filter handle.
CallbackInfo myCBInfo;
	myCBInfo.hCur = hCur;
	strcpy(myCBInfo.Criteria, "RAM");
DbiAddFilter(hCur, (UINT32)&myCBInfo, 1, FALSE, 0, SelectionCallback, &secondaryFilter);
DbiActivateFilter(hCur, secondaryFilter);
.
.
.
*/

//end of example

This information is provided for you to use as you desire without any warranty
or guarantee.  As the functions described are not documented by Borland, support
for these functions may be changed or removed without notice from Borland.

I ask for no monetary renumeration or recognition for the usage of this
information.  All that I ask of you is that if this information benefits you,
please take the time to send a message to Borland to make this a supported
feature.

Terry Swiers, President
Millennium Software, LLC
web: http://www.1000years.com
e-mail: millennium@1000years.com

(**)

interface

uses
   {$IFDEF WIN32}
  Windows, BDE,
  {$ELSE}
  Winprocs, DbiProcs, DbiTypes, DbiErrs,
  {$ENDIF}
  Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, DB, DBTables, Grids, DBGrids, ExtCtrls, DBCtrls;

type

   {$IFNDEF WIN32}
   pByte = ^Byte;
   {$ENDIF}

   { As long as the cursor is passed, anything else can also be declard in this
      structure.  Just be sure to initialize the values first. }
   dbiFilterCB = packed record
      hCursor        : hDBICur;  { Data table cursor }
	   sFind          : string;   { Text to find }
      bCaseSensitive : boolean;  { Do a case sensitive search? }
      oCargo         : TObject;  { Any object to attach to the filter }
	end;

   dbiCodeFilter = function(
      lFilterCB      : longint;  { Pointer to the Client data }
      pRecBuf        : pByte;    { Record buffer pointer to examine }
      lPhyRecNum     : longint   { Physical record number for dBASE tables
                                   Sequential record number for Paradox }
      )              : smallint {$IFDEF WIN32}stdcall{$ENDIF}; { 0 for False, 1 for True }

  TFormFilterCallback = class(TForm)
    DataSource1: TDataSource;
    DBGrid1: TDBGrid;
    Panel1: TPanel;
    ButtonFind: TButton;
    ButtonDrop: TButton;
    DBNavigator1: TDBNavigator;
    EditFind: TEdit;
    ButtonEven: TButton;
    ButtonOdd: TButton;
    CheckBoxCase: TCheckBox;
    Table1: TTable;
    procedure ButtonFindClick(Sender: TObject);
    procedure ButtonDropClick(Sender: TObject);
    procedure ButtonEvenClick(Sender: TObject);
    procedure ButtonOddClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  FormFilterCallback: TFormFilterCallback;

function CodeFilterApply(        { Apply a code filter to a data table }
   oDataset    : TDataset;       { Dataset to filter }
   codeFilter  : dbiCodeFilter;  { Pointer to the filtering function }
   var hFilter : hDBIFilter      { Filter handle to set }
   )           : boolean;

function CodeFilterDrop(         { Drop a (code) filter from a dataset }
   oDataset    : TDataset;       { Dataset that is using passed filter }
   var hFilter : hDBIFilter      { Filter to drop (could be any filter) }
   )           : boolean;        { True if the passed filter was dropped }

function TextInRecord(        { Only display records with specified text }
   lFilterCB      : longint;  { Pointer to the Client data }
   pRecBuf        : pByte;    { Record buffer pointer to examine }
   lPhyRecNum     : longint   { Physical record number (for dBASE tables) }
   )              : smallint {$IFDEF WIN32}stdcall{$ENDIF}; { 0 for False, 1 for True }

function OnlyEvenRecords(     { Only display even-numbered records }
   lFilterCB      : longint;  { Pointer to the Client data }
   pRecBuf        : pByte;    { Record buffer pointer to examine }
   ulPhyRecNum    : longint   { Physical record number (for dBASE tables) }
   )              : smallint {$IFDEF WIN32}stdcall;{$ELSE}; far;{$ENDIF} { 0 for False, 1 for True }

function OnlyOddRecords(      { Only display odd-numbered records }
   lFilterCB      : longint;  { Pointer to the Client data }
   pRecBuf        : pByte;    { Record buffer pointer to examine }
   ulPhyRecNum    : longint   { Physical record number (for dBASE tables) }
   )              : smallint {$IFDEF WIN32}stdcall;{$ELSE}; far;{$ENDIF} { 0 for False, 1 for True }

implementation

{$R *.DFM}

var
   hFilter     : hDBIFilter;     { Filter handle }
   FilterInfo  : dbiFilterCB ;   { Filter information for the specific entry }

{ This function demonstrates accessing information from the callback without
   using the structure -- the physical record number is used instead }
function OnlyEvenRecords(     { Only display even-numbered records }
   lFilterCB      : longint;  { Pointer to the Client data }
   pRecBuf        : pByte;    { Record buffer pointer to examine }
   ulPhyRecNum    : longint   { Physical record number (for dBASE tables) }
   )              : smallint; { 0 for False, 1 for True }
begin
   Result := Smallint( not Odd( ulPhyRecNum ) );  { Will return 0 or 1 }
end; { OnlyEvenRecords() }

{ This function demonstrates accessing information from the callback without
   using the structure -- the physical record number is used instead }
function OnlyOddRecords(      { Only display odd-numbered records }
   lFilterCB      : longint;  { Pointer to the Client data }
   pRecBuf        : pByte;    { Record buffer pointer to examine }
   ulPhyRecNum    : longint   { Physical record number (for dBASE tables) }
   )              : smallint; { 0 for False, 1 for True }
begin
   Result := Smallint( Odd( ulPhyRecNum ) ); { Will return 0 or 1 }
end; { OnlyOddRecords() }

{$IFNDEF WIN32}
procedure SetLength( var s : string; iLen : integer );
begin
   s[ 0 ] := Chr( iLen );
end; { SetLength() }
{$ENDIF}

{ This function demonstrates accessing information from the callback structure
   by calling dbiGetCursorProps to determine the length of the record buffer,
   and by moving the record buffer into a string for the Pos() function }
function TextInRecord(        { Only display records with specified text }
   lFilterCB      : longint;  { Pointer to the Client data }
   pRecBuf        : pByte;    { Record buffer pointer to examine }
   lPhyRecNum     : longint   { Physical record number (for dBASE tables) }
   )              : smallint; { 0 for False, 1 for True }
var
   cbInfo   : dbiFilterCB;
   curProp  : CURProps;
   sBuf     : string;
   iRecSize : integer;
begin
   cbInfo := dbiFilterCB( Pointer( lFilterCB )^ );
   {$IFDEF WIN32} { exceptions cannot be used in Delphi 1.0 callbacks }
   try
   {$ENDIF}
      Check( dbiGetCursorProps( cbInfo.hCursor, curProp ) );
      iRecSize := curProp.iRecSize;
      SetLength( sBuf, iRecSize );
      { Move memory because of binary characters embedded in pRecBuf }
      Move( pRecBuf^, sBuf[1], iRecSize );
      if cbInfo.bCaseSensitive then
         Result := Smallint( Pos( cbInfo.sFind, sBuf ) > 1 )
      else
         Result := Smallint( Pos( UpperCase( cbInfo.sFind ),
            UpperCase( sBuf ) ) > 1 );
   {$IFDEF WIN32}
   except
      Result := 1; { Default to showing the record if an error occurs }
   end; { try .. except }
   {$ENDIF}
end; { TextInRecord() }

function GetTableCursor( oDataset : TDataset ) : hDBICur;
var
	szTable : Array[0..78] of Char;
   iResult : DBIResult;
begin
   iResult := DBIERR_NONE;
   if oDataset is TTable then
   begin
      StrPCopy( szTable, TTable( oDataset ).TableName );
      iResult := DbiGetCursorForTable( TTable( oDataset ).DBHandle, szTable, nil, Result );
   end
   else if oDataset is TQuery then
      iResult := DBIQExec( TQuery( oDataset ).StmtHandle, @Result );

   if iResult <> DBIERR_NONE then
   begin
      Result := nil;
      Check( iResult );
   end;
end; { GetTableCursor() }

function CodeFilterDrop(      { Drop a (code) filter from a dataset }
   oDataset    : TDataset;    { Dataset that is using passed filter }
   var hFilter : hDBIFilter   { Filter to drop from table (could be any filter) }
   )           : boolean;     { True if the passed filter was dropped }
begin
   Result := DbiDropFilter( GetTableCursor( oDataset ), hFilter )
      = DBIERR_NONE;
   oDataset.First;
end; { CodeFilterDrop() }

function CodeFilterApply(        { Attach a code filter to a data table }
   oDataset    : TDataset;       { Dataset to filter }
   codeFilter  : dbiCodeFilter;  { Pointer to the filtering function }
   var hFilter : hDBIFilter      { Filter handle to set }
   )           : boolean;
var
   hCursor  : hDBICur;
begin
   try
      if Assigned( hFilter ) then
      begin
         CodeFilterDrop( oDataset, hFilter );
         hFilter := nil;
      end; { Drop current filter }
      hCursor := GetTableCursor( oDataset );
      (*
      function DbiAddFilter (           { Add a filter to the cursor }
      hCursor       : hDBICur;          { Cursor handle }
      iClientData   : Longint;          { Client supplied data      (opt) }
      iPriority     : Word;             { 1..N with 1 being highest (opt) }
      bCanAbort     : Bool;             { TRUE if pfFiltercan return ABORT (opt) }
      pcanExpr      : pCANExpr;         { Expression tree        (opt) }
      pfFilter      : pfGENFilter;      { ptr to filter function (opt) }
      var   hFilter : hDBIFilter        { Returns filter handle }
      ): DBIResult;

      For Delphi 2.0, DbiAddFilter() is a stdcall routine.
      (**)
      FilterInfo.hCursor   := hCursor;
      Check( DbiAddFilter( hCursor, Longint( @FilterInfo ), 1, False, nil,
         pfGENFilter(@codeFilter), hFilter ) );
      Check( DbiActivateFilter( hCursor, hFilter ) );
      oDataset.Refresh;
      Result := True;
   except
      Result := False;
   end; { try .. except }
end; { CodeFilterApply() }

procedure TFormFilterCallback.ButtonFindClick(Sender: TObject);
var
   timing : TDateTime;
begin
   try
      Screen.Cursor := crHourglass;
      timing := Now;
      FilterInfo.sFind := EditFind.Text;
      FilterInfo.bCaseSensitive := CheckBoxCase.Checked;
      CodeFilterApply( Table1, TextInRecord, hFilter );
      timing := Now - timing;
   finally
      Screen.Cursor := crDefault;
   end;
end;

procedure TFormFilterCallback.ButtonDropClick(Sender: TObject);
begin
   CodeFilterDrop( Table1, hFilter );
end;

procedure TFormFilterCallback.ButtonEvenClick(Sender: TObject);
begin
   try
      Screen.Cursor := crHourglass;
      CodeFilterApply( Table1, OnlyEvenRecords, hFilter );
   finally
      Screen.Cursor := crDefault;
   end;
end;

procedure TFormFilterCallback.ButtonOddClick(Sender: TObject);
begin
   try
      Screen.Cursor := crHourglass;
      CodeFilterApply( Table1, OnlyOddRecords, hFilter );
   finally
      Screen.Cursor := crDefault;
   end;
end;

procedure TFormFilterCallback.FormCreate(Sender: TObject);
begin
   hFilter := nil;
end;

procedure TFormFilterCallback.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
   CodeFilterDrop( Table1, hFilter );
end;

end.
