unit FastUnit;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, MMSystem ;

type
  TFastForm = class(TForm)
    Panel1: TPanel;
    RepaintAllBtn: TButton;
    OptimisedRedraw: TCheckBox;
    Time: TLabel;
    EllipsesBtn: TButton;
    RepaintOneBtn: TButton;
    Panel2: TPanel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormDumbPaint(Sender: TObject);
    procedure FormSmartPaint(Sender: TObject);
    procedure RepaintAllBtnClick(Sender: TObject);
    procedure EllipsesBtnClick(Sender: TObject);
    procedure RepaintOneBtnClick(Sender: TObject);
    procedure OptimisedRedrawClick(Sender: TObject);
  private
    { Private declarations }
    nEllipses : word ;
    Ellipses  : TList ;
    procedure CreateEllipses( Number : word ) ;
  public
    { Public declarations }
  end;

var
  FastForm: TFastForm;

implementation

uses EllpsDlg ;

{$R *.DFM}

type
  TEllipse = class( TObject )
  protected
    LineColor : TColor ;
    FillColor : TColor ;
  public
    Rect      : TRect ;
    constructor Create( const ARect : TRect ;
                        ALineColor  : TColor ;
                        AFillColor  : TColor ) ;
    procedure Paint( ACanvas : TCanvas ) ; virtual ;
  end ;

constructor TEllipse.Create( const ARect : TRect ;
                               ALineColor  : TColor ;
                               AFillColor  : TColor ) ;

begin
  inherited Create ;
  Rect := ARect ;
  LineColor := ALineColor ;
  FillColor := AFillColor ;
end ;


procedure TEllipse.Paint( ACanvas : TCanvas ) ;

begin
  with ACanvas do begin
    Pen.Color := LineColor ;
    Pen.Width := 5 ;
    Pen.Style := psInsideFrame ;
    Brush.Color := FillColor ;
    with Rect do Ellipse( Left, Top, Right, Bottom ) ;
  end ;
end ;


{ TDemoForm }


procedure TFastForm.CreateEllipses( Number : word ) ;

const Colors : array[ 0..10 ] of TColor =
               ( $0000FF, $00FF00, $FF0000,
                 $00FFFF, $FF00FF, $FFFF00,
                 $000080, $008000, $800000,
                 $800080, $808000 ) ;

var i : integer ;

begin
  if not Assigned( Ellipses ) then Ellipses := TList.Create ;
  if Number = nEllipses then exit ;
  nEllipses := Number ;

  { free any existing ellipses }
  for i := Ellipses.Count - 1 downto 0 do
    TEllipse( Ellipses[ i ] ).Free ;
  Ellipses.Clear ;

  { create the ellipses }
  for i := 1 to nEllipses do
    Ellipses.Add( TEllipse.Create( Bounds( Random( Screen.Width - 50 ),
                                             Random( Screen.Height - 50 -
                                                     Panel1.Height ) + Panel1.Height,
                                             Random( 100 ),
                                             Random( 100 ) ),
                                     Colors[ Random( 11 ) ],
                                     Colors[ Random( 11 ) ] ) ) ;

  { invalidate to cause a redraw }
  Invalidate ;
end ;


procedure TFastForm.FormCreate(Sender: TObject);

begin
  CreateEllipses( 100 ) ;
end;

procedure TFastForm.FormDestroy(Sender: TObject);

var i : integer ;

begin
  Ellipses.Free ;
end;

procedure TFastForm.RepaintAllBtnClick(Sender: TObject);
begin
  Invalidate ;
end;


procedure TFastForm.FormDumbPaint(Sender: TObject);

var i : integer ;
    sTime      : string[ 15 ] ;
    OldCursor  : HCursor ;
    StartTime  : longint ;
    FinishTime : longint ;

begin
  { make sure the panel and its components are updated first }
  Panel1.Update ;

  { set the hourglass cursor and note the time }
  OldCursor := SetCursor( LoadCursor( 0, IDC_WAIT ) ) ;
  StartTime := timeGetTime ;

  { now draw the ellipses - "dumb" version blindly paints them all }
  for i := 0 to Ellipses.Count - 1 do
    TEllipse( Ellipses[ i ] ).Paint( Canvas ) ;

  { format the time and update the label }
  FinishTime := timeGetTime ;
  Str( ( FinishTime - StartTime ) / 1000:1:2, sTime ) ;
  Time.Caption := sTime + ' secs' ;

  { reset the cursor }
  SetCursor( OldCursor ) ;
end;


procedure TFastForm.FormSmartPaint(Sender: TObject);

var i : integer ;
    sTime      : string[ 15 ] ;
    OldCursor  : HCursor ;
    StartTime  : longint ;
    FinishTime : longint ;

    { add a clip rect and temporary rect for optimisation }
    ClipRect   : TRect ;
    ARect      : TRect ;

begin
  { make sure the panel and its components are updated first }
  Panel1.Update ;

  { set the hourglass cursor and note the time }
  OldCursor := SetCursor( LoadCursor( 0, IDC_WAIT ) ) ;
  StartTime := timeGetTime ;

  { now draw the ellipses - "smart" version checks whether they need painting }
  { only paints if part of it's rect intersects with the clipping rect }
  GetClipBox( Canvas.Handle, ClipRect ) ;
  for i := 0 to Ellipses.Count - 1 do
    with TEllipse( Ellipses[ i ] ) do
      if IntersectRect( ARect, Rect, ClipRect ) <> 0 then
        Paint( Canvas ) ;

  { format the time and update the label }
  FinishTime := timeGetTime ;
  Str( ( FinishTime - StartTime ) / 1000:1:2, sTime ) ;
  Time.Caption := sTime + ' secs' ;

  { reset the cursor }
  SetCursor( OldCursor ) ;
end;


procedure TFastForm.EllipsesBtnClick(Sender: TObject);
begin
  with EllipsesForm do begin
    NumberOfEllipses.Text := IntToStr( nEllipses ) ;
    if EllipsesForm.ShowModal = mrOK then
      CreateEllipses( StrToInt( NumberOfEllipses.Text ) ) ;
  end ;
end;


procedure TFastForm.RepaintOneBtnClick(Sender: TObject);

  function  CreateFrameRegion( const ARect : TRect ) : HRgn ;

  var Region1, Region2 : HRgn ;

  begin
    { creates a "frame" area using regions as an illustration -
      also illustrates protecting code with try blocks }
    with ARect do begin
      Region1 := CreateRectRgn( Left - 6, Top - 6, Right + 6, Bottom + 6 ) ;
      try
        Region2 := CreateRectRgn( Left, Top, Right, Bottom ) ;
        try
          Result := CreateRectRgn( 0, 0, 0, 0 ) ;
          try
            { remove region 2 from region 1 and delete the source regions }
            CombineRgn( Result, Region1, Region2, RGN_DIFF ) ;
          except
            DeleteObject( Result ) ;
            Raise ;
          end ;
        finally
          DeleteObject( Region2 ) ;
        end ;
      finally
        DeleteObject( Region1 ) ;
      end ;
    end ;
  end ;

const sMessage = 'Updating this rect' ;

var InvalidRect : TRect ;
    FrameRegion : HRgn ;
    ABitmap     : TBitmap ;
    SourceRect  : TRect ;
    DestRect    : TRect ;
    ClientRect  : TRect ;
    i           : integer ;
    TempList    : TList ;

begin
  { invalidate one of the ellipses at random }
  if Assigned( Ellipses ) and ( Ellipses.Count > 0 ) then begin

    { get all ellipses that are at least partly on the screen into a temp list }
    TempList := TList.Create ;
    try
      ClientRect := Bounds( 0, 0, Width, Height ) ;
      for i := Ellipses.Count - 1 downto 0 do
        if IntersectRect( DestRect, TEllipse( Ellipses[ i ] ).Rect,
                          ClientRect ) <> 0 then
          TempList.Add( Ellipses[ i ] ) ;

      { see if we've actually got any ellipses }
      if TempList.Count > 0 then begin

        { get an ellipse at random and use the rect bounding the ellipse to
          make the invalid rect }
        InvalidRect := TEllipse( TempList[ Random( TempList.Count ) ] ).Rect ;
        InvalidateRect( Handle, @InvalidRect, true ) ;

        FrameRegion := CreateFrameRegion( InvalidRect ) ;
        try
          { invert the frame region }
          InvertRgn( Canvas.Handle, FrameRegion ) ;

          { we're going to grab an area of the window for a message rectangle
            so it can be put back later on. First, create dest and source rects. }
          DestRect := Rect( 0, 0, Canvas.TextWidth( sMessage ) + 8,
                            Canvas.TextHeight( sMessage ) + 8 ) ;
          SourceRect := Bounds( InvalidRect.Left - 6, InvalidRect.Bottom + 8,
                                DestRect.Right, DestRect.Bottom ) ;

          { now create a bitmap that we're going to use to hold a copy of the
            area on the screen and use TCanvas.CopyRect to copy it. }
          ABitmap := TBitmap.Create ;
          try
            ABitmap.Width := DestRect.Right ;
            ABitmap.Height := DestRect.Bottom ;
            ABitmap.Canvas.CopyRect( DestRect, Canvas, SourceRect ) ;

            { now paint on the message the message }
            with Canvas do begin
              { setup pens and brushes etc. }
              Brush.Color := clRed ;
              Font.Color := clWhite ;
              Pen.Style := psInsideFrame ;
              Pen.Width := 3 ;
              Pen.Color := clNavy ;

              { draw the rectangle and write the text }
              with SourceRect do Rectangle( Left, Top, Right, Bottom ) ;
              TextOut( SourceRect.Left + 4, SourceRect.Top + 4, sMessage ) ;
            end ;

            { call update to make Windows send a paint message right now }
            Update ;

            { replace the message are from the bitmap and free it }
            Self.Canvas.CopyRect( SourceRect, ABitmap.Canvas, DestRect ) ;
          finally
            ABitmap.Free ;
          end ;

          { invert the frame again to remove it }
          InvertRgn( Canvas.Handle, FrameRegion ) ;
        finally
          { ensure we free the frame region }
          DeleteObject( FrameRegion ) ;
        end ;
      end ;
    finally
      { ensure we free the temp list }
      TempList.Free ;
    end ;
  end ;
end;

procedure TFastForm.OptimisedRedrawClick(Sender: TObject);
begin
  { select the paint handler depending on the state of the optimised checkbox }
  if OptimisedRedraw.Checked then OnPaint := FormSmartPaint else
    OnPaint := FormDumbPaint ;
end;

end.
