unit Arrayu;
{$X+}

interface

uses
  SysUtils;

const
  { Normal error code for RTL range error }
  reRTLRangeError = 201;
  { Code used for range error exception in SysUtils and hidden in RTL }
  reErrorProcRangeError = 4;

type
  TErrorProc = procedure(ErrorCode: Integer; ErrorAddr: Pointer);

  PItemType = ^TItemType;
  { This is the data type stored in the array. Change at will }
  TItemType = Longint;

  TArray = class
  private
    { Upper bound }
    FUpper,
    { Lower bound }
    FLower,
    { Number of elements in the array }
    FNumItems,
    { Number of elements that fit in one segment }
    FNumPerSeg,
    { Number of 64k segments used - there may additionally be a <64k segment }
    FNumWholeSegs: Longint;
    { Pointer to the memory block }
    FPtr: Pointer;
    { ErrorProc is just a pointer. This will be a proper procedure }
    FErrProc: TErrorProc;
    function GetPtr(Index: Longint): PItemType;
    function GetItem(Index: Longint): TItemType;
    procedure PutItem(Index: Longint; const Value: TItemType);
  public
    constructor Create(Lower, Upper: Longint); virtual;
    destructor Destroy; override;
    property Items[Index: Longint]: TItemType read GetItem write PutItem;
      default;
    property Size: Longint read FNumItems;
    Property High: Longint read FUpper;
    Property Low: Longint read FLower;
  end;

implementation

uses
  WinProcs;

const
  SixtyFourK = 64 * 1024;

constructor TArray.Create(Lower, Upper: Longint);
var
  AllocSize: Longint;
begin
  { Assign our error proc pointer to a callable routine }
  { namely SysUtils exception generating handler (ErrorHandler) }
  FErrProc := TErrorProc(ErrorProc);
  FLower := Lower;
  FUpper := Upper;
  FNumItems := FUpper - FLower + 1;
  FNumPerSeg := SixtyFourK div SizeOf(TItemType);
  FNumWholeSegs := FNumItems div FNumPerSeg;
  AllocSize := FNumWholeSegs * SixtyFourK +
    FNumItems mod FNumPerSeg * SizeOf(TItemType);
  FPtr := nil;
  FPtr := GlobalAllocPtr(HeapAllocFlags, AllocSize);
  { If no memory allocated, raise exception }
  if FPtr = nil then
  begin
    FLower := 0;
    FUpper := 0;
    FNumItems := 0;
    OutOfMemoryError;
  end;
end;

destructor TArray.Destroy;
begin
  if FPtr <> nil then
    GlobalFreePtr(FPtr);
  inherited Destroy;
end;

function TArray.GetPtr(Index: Longint): PItemType;
var
  P: PtrRec;

  { Get return address of caller }
  { This has been calculated to return the address of the instruction after }
  { the call to Get/PutItem which in turn calls GetPtr which calls this code }
  function ReturnAddr: Pointer; assembler;
  asm
    mov ax,[bp].Word[36]
    mov dx,[bp].Word[38]
  end;

begin
  { Check for invalid array bound }
  if (Index < FLower) or (Index > FUpper) then
    { This could be replaced with RunError(reRTLRangeError) for a standard }
    { run-time error behaviour but that would terminate the app, not raise }
    { an RTL exception }
    { The reason we pass the address, even though it isn't used for the }
    { exception message (unlike GP fault exception) is that if debugging is }
    { on, the IDE will catch this exception and highlight the line which }
    { called the array routine }
    FErrProc(reErrorProcRangeError, ReturnAddr);
  { Get base address }
  P := PtrRec(FPtr);
  { Increment the selector value appropriately }
  Inc(P.Seg, (Index - FLower) div FNumPerSeg * SelectorInc);
  { Increment the offset value appropriately }
  Inc(P.Ofs, (Index - FLower) mod FNumPerSeg * SizeOf(TItemType));
  Result := PItemType(P);
end;

function TArray.GetItem(Index: Longint): TItemType;
begin
  Result := GetPtr(Index)^;
end;

procedure TArray.PutItem(Index: Longint; const Value: TItemType);
begin
  GetPtr(Index)^ := Value;
end;

end.
