unit Snglins9;

{ This unit defines the single instance classes TSLSingle and         }
{ TSLSingleNum (which is derived from TSLSingle).                     }

interface

uses Classes, Dialogs;

{ Type declaration for the registered OnChanged events. No parameters }
{ are passed.                                                         }

type
  TDataChangedEvent = procedure of object;

{ Type declaration for TSLSingle }

{ Note: Although this class does make use of typed constants (e.g.    }
{ FUsage) it also has an object field FOnChangedList. This is         }
{ initialised in the RealCreate constructor and free'd in the         }
{ RealDestroy constructor, as for a standard class.                   }

{ The Changed method has been overriden. This method will call all    }
{ registered OnChanged event handlers whenever data is added or       }
{ removed from the list.                                              }

{ Free and Destroy have been defined as virtual to provide            }
{ polymorphism for derived classes. See TClientForm.FormDestroy where }
{ the Free method is called.                                          }

{ The class function Usage has been implemented for demonstration     }
{ purposes only. It allows the current value for usage to be returned }
{ and displayed on TMainWindow. Note that this still works when no    }
{ objects have been created.                                          }

{ If compiled in Delphi 2.0 a warning will be generated that the      }
{ Destroy method hides the virtual method of base type 'TStringList'. }
{ This is exactly what we are doing so that warning can be ignored.   }

type
  TSLSingle = class(TStringList)
    FOnChangedList: TList;
  protected
    procedure Changed; override;
    constructor RealCreate; virtual;
    destructor RealDestroy; virtual;
  public
    class function Create: TSLSingle;
    procedure Destroy; virtual;
    procedure Free; virtual;
    procedure RegisterOnChangedEvent(Routine: TDataChangedEvent); virtual;
    procedure UnRegisterOnChangedEvent(Routine: TDataChangedEvent); virtual;
    class function Usage: Integer;
  end;

{ Note: The Destroy and Free methods override the virtual methods in  }
{ the parent class.                                                   }

{ The class function Usage has been implemented for demonstration     }
{ purposes only. It allows the current value for usage to be returned }
{ and displayed on TMainWindow. Note that this still works when no    }
{ objects have been created.                                          }

{ The Add method here simply checks that the first character of the   }
{ string is a digit, and ignores it if it is not.                     }

type
  TSLSingleNum = class(TSLSingle)
  protected
    constructor RealCreate; override;
    destructor RealDestroy; override;
  public
    class function Create: TSLSingleNum;
    procedure Destroy; override;
    procedure Free; override;
    class function Usage: Integer;
    function Add(const s: string): Integer; override;
  end;

implementation

uses WinProcs, SnglIns2, SysUtils, Forms;

const

  { TSLSingle }

  FUsage: Integer = 0;
  FInstance: TSLSingle = nil;

  { Form which provides a view of the data, allowing it to be amended }
  { This has been implemented simply to view/amend the data           }

  FDataView: TDataView = nil;

class function TSLSingle.Create;
begin
  Inc(FUsage);
  If FUsage = 1 then
    begin
      FInstance := RealCreate;
      { This creates the window to allow uses to view/amend central data }
      FDataView := TDataView.Create(nil);
      FDataView.Caption := 'Data View (All)';
      FDataView.ObjectAddr := FInstance;
      FDataView.Show;
    end;
  Result := FInstance
end;

{ This constructor will only be called for the first instance (or     }
{ from derived classes).                                              }

constructor TSLSingle.RealCreate;
begin
  inherited Create;
  Sorted := True;
  Duplicates := dupAccept;
  FOnChangedList := TList.Create;
  LoadFromFile(ExtractFilePath(Application.ExeName)+'\TESTDATA.TXT');
end;

{ The Free/Destroy methods have been implemented as we don't want to  }
{ call the destructor until the last instance has been freed.         }

procedure TSLSingle.Free;
begin
  Destroy;
end;

{ The destroy method implements the check automatically made by the   }
{ compiler for standard classes (i.e. Self <> nil). Due to the way we }
{ have implemented this class we are able to validate the pointers    }
{ passed to the method. Here we check to ensure that it contains the  }
{ value we expect and raise an exception if it is incorrect.          }

procedure TSLSingle.Destroy;

  procedure Error;
  begin
     raise Exception.Create(Format('The Destroy method of the TSLSingle class ' +
       'was called using a pointer value'#13#10'of %p. This is invalid, it should be %p',
       [Pointer(Self),Pointer(FInstance)]));
  end;

begin
  if Self <> nil then
    begin
      { Error if there is not currently an instance of this object   }
      { or the pointer passed is invalid                             }
      if (not Assigned(FInstance)) or (Self <> FInstance) then
        Error;

      Dec(FUsage);

      if FUsage = 0 then
        begin
          RealDestroy;
          FInstance := nil;
          { Destroy the associated window now that the instance has gone }
          FDataView.Free;
          FDataView := nil;
        end;

    end;

end;

destructor TSLSingle.RealDestroy;
begin
  FOnChangedList.Free;
  inherited Destroy;
end;

{ This overrides the method in TStringList to call the registered     }
{ event procedures. By doing this, we have removed the normal         }
{ OnChaned event property. That is, if you set the OnChanged          }
{ property it will have no effect. To re-implement this processing    }
{ you would need to add a call to inherited Changed.                  }

procedure TSLSingle.Changed;
var
  I: Integer;

begin
  { Call each registered callback in turn }
  for I := 0 to FOnChangedList.Count - 1 do
    TDataChangedEvent(FOnChangedList[I]^);
end;

class function TSLSingle.Usage;
begin
  Result := FUsage;
end;

procedure TSLSingle.RegisterOnChangedEvent(Routine: TDataChangedEvent);
var
  DataChangedEvent: ^TDataChangedEvent;

begin
  New(DataChangedEvent);
  DataChangedEvent^ := Routine;
  FOnChangedList.Add(DataChangedEvent);
end;

procedure TSLSingle.UnRegisterOnChangedEvent(Routine: TDataChangedEvent);
var
  I: Integer;
  DataChangedEvent: ^TDataChangedEvent;

begin
  for I := 0 to FOnChangedList.Count - 1 do
    begin
      DataChangedEvent := FOnChangedList[I];
      if (TMethod(DataChangedEvent^).Code = TMethod(Routine).Code) and
        (TMethod(DataChangedEvent^).Data = TMethod(Routine).Data) then
      begin
        Dispose(DataChangedEvent);
        FOnChangedList.Delete(I);
        Break;
      end;
    end;
end;

{ TSLSingleNum }

const

  FNumUsage: Integer = 0;
  FNumInstance: TSLSingleNum = nil;

  { Form which provides a view of the TSLSingleNum data, allowing it  }
  { to be amended. This has been implemented simply to view/amend the }
  { data.                                                             }

  FDataViewNumeric: TDataView = nil;

class function TSLSingleNum.Create: TSLSingleNum;
begin
  Inc(FNumUsage);
  if FNumUsage = 1 then
    begin
      FNumInstance := RealCreate;
      FDataViewNumeric := TDataView.Create(nil);
      FDataViewNumeric.Caption := 'Data View (Numeric)';
      FDataViewNumeric.Top := 290;
      FDataViewNumeric.ObjectAddr := FNumInstance;
      FDataViewNumeric.Show;
    end;
  Result := FNumInstance;
end;

{ Note: This constructor calls inherited RealCreate - not Create.     }

constructor TSLSingleNum.RealCreate;
begin
  inherited RealCreate;
  { Add any class initialisation here }
end;

procedure TSLSingleNum.Free;
begin
 Destroy;
end;

procedure TSLSingleNum.Destroy;

  procedure Error;
  begin
     raise Exception.Create(Format('The Destroy method of the TSLSingleNum class ' +
       'was called using a pointer value'#13#10'of %p. This is invalid, it should be %p',
       [Pointer(Self),Pointer(FNumInstance)]));
  end;

begin
  if Self <> nil then
    begin
      { Error if there is not currently an instance of this object   }
      { or the pointer passed is invalid                             }
      if (not Assigned(FNumInstance)) or (Self <> FNumInstance) then
        Error;

      Dec(FNumUsage);

      if FNumUsage = 0 then
        begin
          RealDestroy;
          FNumInstance := nil;
          FDataViewNumeric.Free;
          FDataViewNumeric := nil;
        end;
    end;
end;

{ Note: This destructor calls inherited RealDestroy - not Destroy.    }

destructor TSLSingleNum.RealDestroy;
begin
  { Add class specific clean up here }
  inherited RealDestroy;
end;

class function TSLSingleNum.Usage: Integer;
begin
  Result := FNumUsage;
end;

{ A simple method to ignore any strings which do not start with a     }
{ digit. Note: Due to polymorphism this is called for each string     }
{ read during the LoadFromFile method. This allows the non-numeric    }
{ data to be stripped out from the same data source                   }

function TSLSingleNum.Add(const S: string): Integer;
begin
  if (Length(S) > 0) and (S[1] in ['0'..'9']) then
    Result := inherited Add(S)
  else
    Result := -1;
end;

end.
