unit HeapStat;

{ Heap statistics unit. Supports both non-OOP callbacks and OOP events.

  Implemented by using the scarcely documented HeapCheck callback in
  the System unit. Delphi 1.0 only!

  Written by Hallvard Vassbotn, December 1995 (hallvard@falcon.no) }

interface

{$IFNDEF VER80}
  Can only be used with Delphi 1.0
{$ENDIF}

type
  THeapOp   = (hoAlloc, hoDealloc, hoBigAlloc);
  THeapStat = record
    ThisOp       : THeapOp;
    ThisSize     : longint;
    FreePtr      : pointer;
    AllocBytes   : longint;
    AllocCount   : longint;
    DeallocBytes : longint;
    DeallocCount : longint;
    DiffBytes    : longint;
    DiffCount    : longint;
    BigBytes     : longint;
    BigCount     : longint;
    Recursive    : boolean;
    IsInCallBack : boolean;
  end;
  THeapStatEvent = procedure(const HeapStat: THeapStat) of object;
  THeapStatProc  = procedure(const HeapStat: THeapStat);

procedure HeapStatInit(aHeapStatEvent: THeapStatEvent;
                       aHeapStatProc : THeapStatProc);

function HeapStatDone: boolean;

implementation

uses
  SysUtils;

type
  PtrRec = record
    Ofs, Seg: word;
  end;
  TProcedure = procedure;

const
  HeapStatEvent  : THeapStatEvent = nil;
  HeapStatProc   : THeapStatProc = nil;
  SaveHeapCheck  : pointer = nil;
  HeapCheckMagic : word = 0;
var
  HeapStatistics: THeapStat;
  LocalHS       : THeapStat;

{$IFOPT S+} {$DEFINE STACKCHK} {$S-} {$ENDIF}

{ Note: Stack-checking cannot be turned on here. It would overwrite AX!

  The HeapCheck callback doesn't really take any parameters, so we have
  to do some checks of registers and stack contents to find out were
  we are called from (allocation or deallocation) and what the
  block size is.

  Use Cdecl calling convention to look at the stack without changing
  the stack pointer. }

procedure HeapCheckProc(AllocSize, FreeSeg, FreeSize: word); cdecl; far;
var
  FreeOfs : word absolute AllocSize;
  AX_, BX_, CX_, ES_, SP_, BP_: Word;
  Allocate  : boolean;
  Deallocate: boolean;
begin
  { Save register values to variables }
  asm
    MOV AX_, AX
    MOV BX_, BX
    MOV CX_, CX
    MOV ES_, ES
    MOV BP_, BP
  end;
  { Do some magic checks to see if we are called from NewMemory or DisMemory in WMEM.ASM }
  Deallocate := (FreeSeg = ES_) and (FreeOfs = BX_) and (HeapCheckMagic = CX_);
  Allocate   := (not Deallocate) and (AllocSize = AX_);

  with HeapStatistics do
  begin
    if Allocate then
    begin
      FreePtr := nil;
      ThisSize := AllocSize;
      if ThisSize < System.HeapLimit then
      begin
        ThisOp := hoAlloc;
        Inc(AllocBytes, ThisSize);
        Inc(AllocCount);
      end
      else
      begin
        ThisOp := hoBigAlloc;
        Inc(BigBytes, ThisSize);
        Inc(BigCount);
      end;
    end
    else
    begin
      PtrRec(FreePtr).Seg := FreeSeg;
      PtrRec(FreePtr).Ofs := FreeOfs;
      ThisSize := FreeSize;
      ThisOp   := hoDealloc;
      Inc(DeallocBytes, ThisSize);
      Inc(DeallocCount);
    end;
    DiffBytes := AllocBytes - DeallocBytes;
    DiffCount := AllocCount - DeallocCount;

    { Call the callback/event handler, but look out for recursiveness }
    if not IsInCallBack then
    begin
      IsInCallBack := true;
      LocalHS := HeapStatistics;
      if Assigned(HeapStatProc) then
        HeapStatProc(LocalHS);
      if Assigned(HeapStatEvent) then
        HeapStatEvent(LocalHS);
      IsInCallBack := false;
    end
    else
      Recursive := true;
  end;

  { Call any previous HeapCheck callback that might be installed.}
  if Assigned(SaveHeapCheck) then
    TProcedure(SaveHeapCheck);
end;

{$IFDEF STACKCHK} {$UNDEF STACKCHK} {$S+} {$ENDIF}

procedure HeapStatInit(aHeapStatEvent: THeapStatEvent;
                       aHeapStatProc : THeapStatProc);
begin
  if System.HeapCheck <> @HeapCheckProc then
  begin
    FillChar(HeapStatistics, SizeOf(HeapStatistics), 0);
    SaveHeapCheck    := System.HeapCheck;
    System.HeapCheck := @HeapCheckProc;
    HeapCheckMagic   := LongRec(HeapCheck).Hi or LongRec(HeapCheck).Lo;
  end;
  HeapStatEvent    := aHeapStatEvent;
  HeapStatProc     := aHeapStatProc;
end;

function HeapStatDone: boolean;
begin
  System.HeapCheck := SaveHeapCheck;
  Result := not HeapStatistics.Recursive;
end;

end.
