unit PtInvoce;

interface
uses Windows, Messages, Dialogs, SysUtils, Classes;

const
  Heading1 = 'Marine Adventures Inc.';
  Heading2 = '1234 N. Ocean Blvd.';
  Heading3 = 'Santa Cruz, CA 95066';
  Footer = 'All sales are final - no returns or exchanges!!';
  RightBorderMargin = 7;

type

  TInvoiceReport = class
  private
    FPageNum: Integer;    // Keeps track of pages
    FPixInInchX,          // # of pixels per horizontal inch
    FPixInInchY: Integer; // # of pixels per vertical inch
    FAmountPrintedY: Integer; // Keeps track of pixels used along Y axis
    FPageCenterX: Integer;    // Center of the page, used for alignment
    FLeftMargin: Integer;     // Left Border Margin
    FDiscountAlignPos: Integer;  // Used for alignment
    FSellPriceAlignPos: Integer; // Used for alignment
    FExtPriceAlignPos: Integer;  // Used for alignment
    procedure StartPrinting;     // Starts the printing process
    procedure PrintBorders;      // Prints the border
    procedure PrintHeading;      // Prints heading information
    procedure PrintCustomerInfo; // Prints Bill To information
    procedure PrintShipToInfo;   // Prints Ship To information
    procedure PrintOrderInfo;    // Prints Order data
    procedure PrintTitles;       // Prints item titles
    procedure PrintRecords;      // Prints ordered items
    procedure PrintTotals;       // Prints totals
    procedure PrintFooter;       // Prints footer and page number
    function PageDone: Boolean;  // Detemines if there is vertical space left
    function GetLineHeight: Integer; // Calculates a line height
    procedure SetFontSize(Size: Integer); // Sets the font size
    procedure WriteDecimalAlign(R: TRect; S: String; RAlignPos: Integer);
    { Writes a floating point number based on a specific decimal alignment
      position }
  public
    procedure PrintInvoice;
    { Public procedure called by TInvoiceRecord user to print the invoice }
  end;

implementation
uses Printers, EdOrders, DataMod;

function TInvoiceReport.GetLineHeight: Integer;
{ This function determines a line height based on the currently rendered font }
var
  TM: TTextMetric;
begin
  GetTextMetrics(Printer.Canvas.Handle, TM);
  Result := TM.tmHeight;
end;

procedure TInvoiceReport.SetFontSize(Size: Integer);
{ This procedure sets the printer font size. }
begin
  { The line below must be executed because it has been experienced that
    Delphi sometime does not set this value correctly }
  Printer.Canvas.Font.PixelsPerInch:=
      GetDeviceCaps(Printer.Canvas.Handle, LOGPIXELSY);
  Printer.Canvas.Font.Size := Size;
end;

procedure TInvoiceReport.PrintInvoice;
{ This function starts the print job. It is called by the TInvoiceReport user }
begin
  Printer.BeginDoc; // Start the print job.
  try
    StartPrinting ;  // Perform the printing operations.
  finally
    Printer.EndDoc; // End the print job.
  end;
end;

procedure TInvoiceReport.StartPrinting;
{ This function calls all of the invoice print function after setting up the
  initial global values. }
begin
  FAmountPrintedY := 0; // Y Position starts a zero
  FPageNum := 1;        // Page 1
  { Get the number of pixels along the X and Y axis by calling the
    GetDeviceCaps Win32 API function }
  FPixInInchX := GetDeviceCaps(Printer.Canvas.Handle, LOGPIXELSX);
  FPixInInchY := GetDeviceCaps(Printer.Canvas.Handle, LOGPIXELSY);
  { Calculate the center of the page }
  FPageCenterX := Round(Printer.PageWidth / 2);
  { The bounding border is 7 inches wide. Therefore, center its Left position
    based on the center of the page }
  FLeftMargin := FPageCenterX - Round(3.5*FPixInInchX);
  { Set a standard font for the report }
  Printer.Canvas.Font.Name := 'Arial';
  SetFontSize(8);    // Set the font size to 8
  PrintBorders;      // Print report borders.
  PrintHeading;      // Print heading information.
  PrintCustomerInfo; // Print Bill To information.
  PrintShipToInfo;   // Print Ship To information.
  PrintOrderInfo;    // Print Order data.
  PrintTitles;       // Print items titles.
  PrintRecords;      // Print sale items.
  PrintTotals;       // Print cost totals.
  PrintFooter;       // Print footer information.
end;

procedure TInvoiceReport.PrintBorders;
{ This procedure prints out the borders for the report }
begin
  with Printer.Canvas do begin
    Pen.Width := Round(FPixInInchX / 40); // Set a wider pen width.
    { Draw a rectangle in which the Bill To, Send To and order information
      will be drawn }
    Rectangle(FLeftMargin, 1*FPixInInchY,
      FLeftMargin+RightBorderMargin*FPixInInchX, trunc(2.5*FPixInInchY));
    { Draw a recangle in which the sales items will be drawn. }
    Rectangle(FLeftMargin, trunc(2.5*FPixInInchY),
      FLeftMargin+RightBorderMargin*FPixInInchX,
      Printer.PageHeight-FPixInInchY);
    Pen.Width := Round(FPixInInchX / 80); // Restore the pen width.
    { Draw a vertical line to separate the totals from the rest of the sales
      item information }
    MoveTo(FLeftMargin+FPixInInchX*6, trunc(2.5*FPixInInchY));
    LineTo(FLeftMargin+FPixInInchX*6, Printer.PageHeight-FPixInInchY);
  end;
end;

procedure TInvoiceReport.PrintHeading;
{ This procedure prints the heading information for the invoice report }
var
  OldSize: Integer;
begin
  FAmountPrintedY := 0; // Set Y printing position to zero.
  with Printer.Canvas do begin
    Font.Style := [fsBold];  // Heading is printed in bold face.
    OldSize := Font.Size;    // Save the original font size.
    SetFontSize(12);         // Set a new font size to 12.

    { Print the heading constants by centering them on the page. }
    TextOut(FPageCenterX-(Round(TextWidth(Heading1) / 2)),
      FAmountPrintedY, Heading1);
    { Increment FAmountPrintedY by 1+1/20th of a line height in the current
      font. }
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+(round(GetLineHeight /20));
    { Draw remaining header lines. }
    TextOut(FPageCenterX-(Round(TextWidth(Heading2) / 2)),
      FAmountPrintedY, Heading2);
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+(Round(GetLineHeight / 20));
    TextOut(FPageCenterX-(Round(TextWidth(Heading3) / 2)),
      FAmountPrintedY, Heading3);
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+(Round(GetLineHeight / 20));
    Font.Style := [];     // Restore font style.
    SetFontSize(OldSize);  // Restore font size.
  end;
end;

procedure TInvoiceReport.PrintCustomerInfo;
{ This procedure prints the Bill To customer information }
var
  LeftTextMargin: Integer;  // Keep track of left text margin
  OldSize: Integer;
begin
  with Printer.Canvas, EdOrderForm do begin
    OldSize := Printer.Canvas.Font.Size;  // Save the print font size.
    SetFontSize(10);                      // Set printer font size to 10
    { Calculate a left margin from which to start printing text. }
    LeftTextMargin := FLeftMargin + (Round(FPixInInchX / 4));
    FAmountPrintedY := 1*FPixInInchY + (round(FPixInInchY / 20));
    Printer.Canvas.Font.Style := [fsBold];
    TextOut(LeftTextMargin, FAmountPrintedY, 'BILL TO:');
    { Double the distance between the title and the information }
    FAmountPrintedY := FAmountPrintedY+(2*GetLineHeight)+
      Round(GetLineHeight / 20);
    { Restore font style }
    Printer.Canvas.Font.Style := [];
    { Draw customer information }
    TextOut(LeftTextMargin, FAmountPrintedY, CompanyCombo.Text);
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
     Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, CustAdd1Edit.Text);
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, CustAdd2Edit.Text);
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    { The following if statement prevents a comma from being printed if there
      is no city }
    if CustCityEdit.Text <> '' then
      TextOut(LeftTextMargin, FAmountPrintedY, CustCityEdit.Text+',');
    LeftTextMargin := LeftTextMargin + TextWidth(CustCityEdit.Text+',') +
      Round(FPixInInchX / 8);
    TextOut(LeftTextMargin, FAmountPrintedY, CustStateEdit.Text);
    LeftTextMargin := LeftTextMargin + TextWidth(CustStateEdit.Text) +
      Round(FPixInInchX / 4);
    TextOut(LeftTextMargin, FAmountPrintedY, CustZipEdit.Text);
    SetFontSize(OldSize); // Restore font size
  end;
end;

procedure TInvoiceReport.PrintShipToInfo;
{ This procedure prints the Ship To information }
var
  LeftTextMargin: Integer;
  OldSize: Integer;
begin
  with Printer.Canvas, EdOrderForm do begin
    OldSize := Printer.Canvas.Font.Size; // Save Font size
    SetFontSize(10); // Set Font size to 10
    { Set next margin at 3 inches from the original border position. }
    LeftTextMargin := RightBorderMargin+(FPixInInchX*3);
    FAmountPrintedY := 1*FPixInInchY + Round(FPixInInchY / 20);
    Printer.Canvas.Font.Style := [fsBold]; // Change font to bolded
    TextOut(LeftTextMargin, FAmountPrintedY, 'SHIP TO:');
    { Double the distance between the title and the information }
    FAmountPrintedY := FAmountPrintedY+(2*GetLineHeight)+
      Round(GetLineHeight / 20);
    Printer.Canvas.Font.Style := []; // Restore font style
    { Print the Ship To information }
    TextOut(LeftTextMargin, FAmountPrintedY, ShipToCompanyEdit.Text);
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, ShipToAdd1Edit.Text);
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, ShipToAdd2Edit.Text);
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    { The following if statement prevents a comma from being printed if there
      is no city }
    if ShipToCityEdit.Text <> '' then
      TextOut(LeftTextMargin, FAmountPrintedY, ShipToCityEdit.Text+',');
    LeftTextMargin := LeftTextMargin + TextWidth(ShipToCityEdit.Text+',') +
      Round(FPixInInchX / 8);
    TextOut(LeftTextMargin, FAmountPrintedY, ShipToStateEdit.Text);
    LeftTextMargin := LeftTextMargin + TextWidth(ShipToStateEdit.Text) +
      Round(FPixInInchX / 4);
    TextOut(LeftTextMargin, FAmountPrintedY, ShipToZipEdit.Text);
    SetFontSize(OldSize); // Restore the font size.
  end;
end;

procedure TInvoiceReport.PrintOrderInfo;
{ This procedure prints other miscellaneous order information. }
var
  OldSize: Integer;
  LeftTextMargin: Integer;
  LeftTextMargin2: Integer;
begin
  with Printer.Canvas, EdOrderForm, MastData do begin
    OldSize := Printer.Canvas.Font.Size; // Save font size.
    SetFontSize(10); // Set font size to 10
    { Set next margin at 5 inches from the original border position. }
    LeftTextMargin := RightBorderMargin+(FPixInInchX*5);
    { The string "Payment Method:" is the longest label, so align the data
      along with it's data }
    LeftTextMargin2 := LeftTextMargin+TextWidth('Payment Method: ');
    FAmountPrintedY := 1*FPixInInchY + Round(FPixInInchY / 20);
    { Print the additional order information }
    TextOut(LeftTextMargin, FAmountPrintedY, 'Date: ');
    TextOut(LeftTextMargin2, FAmountPrintedY, SaleDateEdit.Text);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
       Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'Cust No: ');
    TextOut(LeftTextMargin2, FAmountPrintedY, CustNoEdit.Text);

    { OrdersOrderNo exists on the MastData data module }
    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'Order No: ');
    TextOut(LeftTextMargin2, FAmountPrintedY, FloatToStr(OrdersOrderNo.Value));

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'Sold By: ');
    TextOut(LeftTextMargin2, FAmountPrintedY, SoldByCombo.Text);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'Terms: ');
    TextOut(LeftTextMargin2, FAmountPrintedY, TermsCombo.Text);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'Payment Method: ');
    TextOut(LeftTextMargin2, FAmountPrintedY, PaymentCombo.Text);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'Ship Via: ');
    TextOut(LeftTextMargin2, FAmountPrintedY, ShipViaCombo.Text);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'PO#: ');
    TextOut(LeftTextMargin2, FAmountPrintedY, POEdit.Text);

    SetFontSize(OldSize); // Restore font size.
  end;
end;

procedure TInvoiceReport.PrintTitles;
{ This procedure prints the sales item titles. }
var
  OldSize: Integer;
begin
  with Printer.Canvas do begin
    OldSize := Printer.Canvas.Font.Size; // Save font size
    SetFontSize(8); // Set font size to 8
    FAmountPrintedY := trunc(2.5*FPixInInchY)+GetLineHeight+
      Round(GetLineHeight / 20);
    Printer.Canvas.Font.Style := [fsBold]; // font style to bolded.
    { Print out the titles. }
    TextOut(FLeftMargin+
      Round(FPixInInchX / 10), FAmountPrintedY, 'PART NO');
    TextOut(FLeftMargin+FPixInInchX, FAmountPrintedY, 'DESCRIPTION');
    TextOut(FLeftMargin+trunc(FPixInInchX*3.5), FAmountPrintedY, 'DISCOUNT');
    { FDiscountAlignPos will hold alignment value to be used in printing
      records }
    FDiscountAlignPos := FLeftMargin+trunc(FPixInInchX*3.5)+
       TextWidth('DISCOUNT');
    TextOut(FLeftMargin+trunc(FPixInInchX*4.5), FAmountPrintedY, 'QTY');
    TextOut(FLeftMargin+FPixInInchX*5, FAmountPrintedY, 'SELL PRICE');
    { FSellPriceAlignPos will hold alignment value to be used in printing
      records }
    FSellPriceAlignPos := FLeftMargin+FPixInInchX*5+TextWidth('SELL PRICE');

    TextOut(FLeftMargin+FPixInInchX*6+Round(FPixInInchX / 10), FAmountPrintedY,
       'EXT PRICE');
    { FExtPriceAlignPos will hold alignment value to be used in printing
      records }
    FExtPriceAlignPos := FLeftMargin+FPixInInchX*6+
      Round(FPixInInchX / 10)+ TextWidth('EXT PRICE');
    { Restore printer information }
    Printer.Canvas.Font.Style := [];
    SetFontSize(OldSize);
  end;

end;

procedure TInvoiceReport.WriteDecimalAlign(R: TRect; S: String;
          RAlignPos: Integer);
{ This procedure is used to decimal align columns of floating point values, the
  alignment position is specified by the RAlignPos parameter }
var
  OldAlign: Integer;
  P: Integer;
begin
  with Printer.Canvas, MastData do begin
    { Save the original alignment flag }
    OldAlign := SetTextAlign(Handle, TA_RIGHT);
    P := Pos('.', S); // Determine the position of the decimal
    { Draw the integer portion of the value }
    TextRect(R, R.Right, R.Top, Copy(S, 1, P - 1));
    { Recalculate the alignment position }
    R.Left := R.Right;
    R.Right := RAlignPos;
    { Change alignment to left align }
    SetTextAlign(Handle, TA_LEFT);
    { Draw the fractional portion of the decimal along with the decimal }
    TextRect(R, R.Left, R.Top, Copy(S, P, 255));
    SetTextAlign(Handle, OldAlign); // Restore the original alignment flag
  end;
end;

procedure TInvoiceReport.PrintRecords;
{ This procedure prints each sales item record }
var
  R: TRect;
  S: String;
begin
  { Recalculate the Y position from which to begin drawing }
  FAmountPrintedY := FAmountPrintedY+(GetLineHeight*2)+
    Round(GetLineHeight / 20);
  with Printer.Canvas, MastData do begin
    Items.First; // go to first item in the database.
    { Loop through sales items and print each record. }
    while (not Items.EOF) and (not PageDone) do begin
      { Print each field from the sales item. Align the data accordingly. }
      TextOut(FLeftMargin+Round(FPixInInchX / 10), FAmountPrintedY,
              FloatToStr(ItemsPartNo.Value));
      TextOut(FLeftMargin+FPixInInchX, FAmountPrintedY, ItemsDescription.Value);

      { Determine the TRect values for use with the WriteDecimalAlign
        procedure }
      R := Rect(FLeftMargin+trunc(FPixInInchX*3.5), FAmountPrintedY,
                FDiscountAlignPos-TextWidth('.00%'),
                FAmountPrintedY+GetLineHeight);
      { Format the decimal string to pass to WriteDecimalAlign }
      S := FormatFloat('0.00%', ItemsDiscount.Value);
      WriteDecimalAlign(R, S, FDiscountAlignPos);

      { Continue printing sales item data }
      TextOut(FLeftMargin+trunc(FPixInInchX*4.5), FAmountPrintedY,
              IntToStr(ItemsQty.Value));
      R := Rect(FLeftMargin+FPixInInchX*5, FAmountPrintedY,
                 FSellPriceAlignPos-TextWidth('.00'),
                 FAmountPrintedY+GetLineHeight);
      S := FormatFloat('$0.00', ItemsSellPrice.Value);
      WriteDecimalAlign(R, S, FSellPriceAlignPos);

      R := Rect(FLeftMargin+FPixInInchX*
                6+Round(FPixInInchX / 10), FAmountPrintedY,
                FExtPriceAlignPos-TextWidth('.00'),
                FAmountPrintedY+GetLineHeight);
      S := FormatFloat('$0.00', ItemsExtPrice.Value);
      WriteDecimalAlign(R, S, FExtPriceAlignPos);
      FAmountPrintedY := FAmountPrintedY+GetLineHeight+
        Round(GetLineHeight / 20);
      Items.Next;
      { If the next record will not fit on the current page, then start a new
        page. The report information like borders, heading etc must be printed
        on the new page.  }
      if (not Items.Eof) and PageDone then begin
        PrintFooter;
        Printer.NewPage;
        PrintBorders;
        PrintHeading;
        PrintCustomerInfo;
        PrintShipToInfo;
        PrintOrderInfo;
        PrintTitles;
        FAmountPrintedY := FAmountPrintedY+(GetLineHeight*2)+
           Round(GetLineHeight / 20);
        inc(FPageNum); // Increment the page number. 
      end;
    end;
  end;
end;

procedure TInvoiceReport.PrintTotals;
{ This procedure prints the sales totals }
var
  LeftTextMargin: Integer;
  R: TRect;
  S: String;
begin
  { If the sales totals information + 1 inch will not fit on the current page
    go to a new page. }
  if FAmountPrintedY+FPixInInchY+(GetLineHeight*5) >=
    Printer.PageHeight - FPixInInchY then begin
      PrintFooter;
      Printer.NewPage;
      PrintBorders;
      PrintHeading;
      PrintCustomerInfo;
      PrintShipToInfo;
      PrintOrderInfo;
      PrintTitles;
  end;
  with EdOrderForm, Printer.Canvas do begin
    { Print the totals information }
    LeftTextMargin := FLeftMargin+FPixInInchX*5;
    FAmountPrintedY := FAmountPrintedY+FPixInInchY;
    TextOut(LeftTextMargin, FAmountPrintedY, 'Subtotal');
    { Calculate TRect values to pass to WriteDecimalAlign }
    R := Rect(FLeftMargin+FPixInInchX*6+Round(FPixInInchX / 10),
              FAmountPrintedY, FExtPriceAlignPos-TextWidth('.00'),
              FAmountPrintedY+GetLineHeight);
    { If the control on the form has focus the "$" and "%" characters are
      removed. Use the Pos function to ensure these characters are written }
    S := TotalEdit.Text;
    if Pos('$', S) = 0 then
      S := '$'+S;

    WriteDecimalAlign(R, S, FExtPriceAlignPos);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
     Round(GetLineHeight / 20);
    S := TaxRateEdit.Text;
    if Pos('%', S) = 0 then
      S := S+'%';
    TextOut(LeftTextMargin, FAmountPrintedY, 'Tax @'+S);

    R := Rect(FLeftMargin+FPixInInchX*6+
              Round(FPixInInchX / 10), FAmountPrintedY,
              FExtPriceAlignPos-TextWidth('.00'),
              FAmountPrintedY+GetLineHeight);
    S := TaxTotalEdit.Text;
    if Pos('$', S) = 0 then
      S := '$'+S;
    WriteDecimalAlign(R, S, FExtPriceAlignPos);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
       Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'Freight');
    R := Rect(FLeftMargin+FPixInInchX*6+
      Round(FPixInInchX / 10), FAmountPrintedY,
              FExtPriceAlignPos-TextWidth('.00'),
              FAmountPrintedY+GetLineHeight);
    S := FreightEdit.Text;
    if Pos('$', S) = 0 then
      S := '$'+S;
    WriteDecimalAlign(R, S, FExtPriceAlignPos);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    TextOut(LeftTextMargin, FAmountPrintedY, 'Amount Paid');
    R := Rect(FLeftMargin+FPixInInchX*6+
      Round(FPixInInchX / 10), FAmountPrintedY,
      FExtPriceAlignPos-TextWidth('.00'), FAmountPrintedY+GetLineHeight);
    S := AmountPaidEdit.Text;
    if Pos('$', S) = 0 then
      S := '$'+S;
    WriteDecimalAlign(R, S, FExtPriceAlignPos);

    FAmountPrintedY := FAmountPrintedY+GetLineHeight+
      Round(GetLineHeight / 20);
    Printer.Canvas.Font.Style := [fsBold];
    TextOut(LeftTextMargin, FAmountPrintedY, 'Amount Due');
    R := Rect(FLeftMargin+FPixInInchX*6+
      Round(FPixInInchX / 10), FAmountPrintedY,
      FExtPriceAlignPos-TextWidth('.00'), FAmountPrintedY+GetLineHeight);
    S := AmountDueEdit.Text;
    if Pos('$', S) = 0 then
      S := '$'+S;
    WriteDecimalAlign(R, S, FExtPriceAlignPos);
    Printer.Canvas.Font.Style := [];
  end;
end;

procedure TInvoiceReport.PrintFooter;
{ This procedure prints footer information and page number }
begin
  with Printer.Canvas do begin
    TextOut(FLeftMargin, Printer.PageHeight-
      Round(FPixInInchX / 2), Footer);
    TextOut(RightBorderMargin*FPixInInchX -
       TextWidth('  Page '+IntToStr(FPageNum)),
       Printer.PageHeight-Round(FPixInInchX / 2), 'Page '+IntToStr(FPageNum));
  end;
end;

function TInvoiceReport.PageDone: Boolean;
{ This procedure determines if additional information can be printed on the
  current page. }
begin
  Result := FAmountPrintedY+GetLineHeight >= Printer.PageHeight - FPixInInchY;
end;

end.
