unit Login;

interface

uses
  Classes, Forms, SysUtils, DB, DBTables;

const
  cNoUserID         = -1;   { Token for no user connected }
  cNoAppID          = -1;   { Token for no application defined }
  DefNumAttempts    = 3;    { Default number of login retry attempts }

type
  TLoginEvent = procedure(Sender: TObject; UserName: string; Password: string) of object;
  TLoggingInEvent = procedure(Sender: TObject; UserName: string; Password: string; var Cancel: Boolean) of object;
  TLoggingOutEvent = procedure(Sender: TObject; var Cancel: Boolean) of object;

  TLoginManager = class(TComponent)
  protected
    FApplicationID: LongInt;       { Identifier for the application }
    FDateLastLogin: TDateTime;     { Date/time of the last login for this user }
    FPasswordExpired: Boolean;     { True if user's password has expired on this login }
    FMainDB: TDatabase;            { Pointer to the application's TDatabase component }
    FNumAttemptsAllowed: Integer;  { Number of login retries allowed }
    FNumFailedAttempts: Integer;   { Number of failed login attempts so far }
    FPassword: string;             { User's password }
    FUserFirstName: string;        { Proper first name for user }
    FUserFullName: string;         { Proper full name for user }
    FUserLastName: string;         { Proper last name for user }
    FUserID: LongInt;              { System ID for user }
    FUsername: string;             { Login username for user }
    FOnLogin: TLoginEvent;         { Event-handler }
    FOnLoggingIn: TLoggingInEvent; { Event-handler }
    FOnLogout: TNotifyEvent;       { Event-handler }
    FOnLoggingOut: TLoggingOutEvent; { Event-handler }
    FOnBadLogin: TNotifyEvent;     { Event-handler }

    procedure Connect;
    procedure ConnectDB(DB: TDatabase; Username, Password: string);
    procedure Disconnect;
    procedure DisconnectDB(DB: TDatabase);
    procedure SetMainDB(Value: TDatabase);
  public
    constructor Create(AOwner: TComponent); override;
    procedure Login(UserName, Password: string);
    procedure Logout;

    property ApplicationID: LongInt read FApplicationID write FApplicationID;
    property MainDB: TDatabase read FMainDB write SetMainDB;
    property NumAttemptsAllowed: Integer read FNumAttemptsAllowed write FNumAttemptsAllowed default defNumAttempts;
    property Password: string read FPassword;
    property UserFirstName: string read FUserFirstName;
    property UserFullName: string read FUserFullName;
    property UserLastName: string read FUserLastName;
    property UserID: LongInt read FUserID;
    property Username: string read FUsername;

    property OnLogin: TLoginEvent read FOnLogin write FOnLogin;
    property OnLoggingIn: TLoggingInEvent read FOnLoggingIn write FOnLoggingIn;
    property OnLogout: TNotifyEvent read FOnLogout write FOnLogout;
    property OnLoggingOut: TLoggingOutEvent read FOnLoggingOut write FOnLoggingOut;
    property OnBadLogin: TNotifyEvent read FOnBadLogin write FOnBadLogin;
  end;

var
  LoginManager: TLoginManager;

implementation

uses
  Controls, Dialogs, DMLogin, PswdChg;

{ TLoginManager }

constructor TLoginManager.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  { Establish connection to the data module code }
  LoginDM := TLoginDM.Create(Self);

  FApplicationID := cNoAppID;
  FUserID := cNoUserID;
  FUserFullName := '';
  FUserFirstName := '';
  FUserLastName := '';
  FNumAttemptsAllowed := DefNumAttempts;
end;

procedure TLoginManager.Connect;
begin
  ConnectDB(FMainDB, FUsername, FPassword);
end;

procedure TLoginManager.ConnectDB(DB: TDatabase; Username, Password: String);
begin
  if DB <> nil then
    with DB do
    begin
      Connected := False;
      LoginPrompt := False;
      Params.Values['USER NAME'] := Username;
      Params.Values['PASSWORD'] := Password;
      KeepConnection := True;
      Connected := True;
    end;
end;

procedure TLoginManager.Disconnect;
begin
  DisconnectDB(FMainDB);
end;

procedure TLoginManager.DisconnectDB(DB: TDatabase);
begin
  if DB <> nil then
    with DB do
    begin
      KeepConnection := False;
      Connected := False;
    end;
end;

procedure TLoginManager.Login(UserName, Password: String);
var
  Cancel: Boolean;
begin
  Logout;

  FUsername := Username;
  FPassword := Password;
  try
    { Any exception occurring within this block is considered a failed login attempt }

    Connect;
    FUserID := cNoUserID;
    LoginDM.GetUserValues(FUserID, FUserFirstName, FUserLastName,
                          FDateLastLogin, FPasswordExpired);
    FUserFullName := FUserFirstName + ' ' + FUserLastName;

    if Assigned(FOnLoggingIn) then
    begin
      FOnLoggingIn(Self, UserName, Password, Cancel);
      if Cancel then
      begin
        Disconnect;
        Exit;
      end;
    end;

    LoginDM.PostAuditTrail(evtLoginSuccessful, '');
    LoginDM.PostUserLoginDate;
    FNumFailedAttempts := 0;
  except
    on E: Exception do
    begin     { Failed login attempt }
      Application.HandleException(Self);
      Disconnect;
      if Assigned(FOnBadLogin) then
        FOnBadLogin(Self);

      { Count the number of failed attempts and shut down if the fail limit has been reached }
      Inc(FNumFailedAttempts);
      if FNumFailedAttempts >= NumAttemptsAllowed then
      begin
        MessageDlg(IntToStr(FNumFailedAttempts) +
                   ' login attempts have failed.  ' +
                   'Shutting down the application.',
                   mtError, [mbOk], 0);
        Application.Terminate;
      end;

      FUserID := cNoUserID;
      FUsername := '';
      FPassword := '';
      Exit;
    end;
  end;

  if Assigned(FOnLogin) then
    FOnLogin(Self, UserName, Password);
end;

procedure TLoginManager.Logout;
var
  Cancel: Boolean;
begin
  if FUserID <> cNoUserID then
  begin
    if Assigned(FOnLoggingOut) then
    begin
      Cancel := False;
      FOnLoggingOut(Self, Cancel);
      if Cancel then Exit;
    end;

    Disconnect;

    if Assigned(FOnLogout) then FOnLogout(Self);
    LoginDM.PostAuditTrail(evtLogout, '');
    FUserID := cNoUserID;
  end;
end;

procedure TLoginManager.SetMainDB(Value: TDatabase);
var
  I: Integer;
begin
  if Value <> FMainDB then
  begin
    FMainDB := Value;

    { Initialize the dataset components in the data module }
    for I := 0 to LoginDM.ComponentCount - 1 do
      if LoginDM.Components[I] is TDBDataSet then
        with TDBDataSet(LoginDM.Components[I]) do
        begin
          Active := False;
          DatabaseName := FMainDB.DatabaseName;
        end;
  end;
end;

initialization
  LoginManager := nil;
  LoginManager := TLoginManager.Create(Application);
end.

