{$A+,B-,C-,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R+,S-,T-,U-,V+,W-,X+,Y-,Z1}
{$MINSTACKSIZE $00004000}
{$MAXSTACKSIZE $00100000}
{$IMAGEBASE $00400000}
{$APPTYPE GUI}
unit Enque;

{-------------------------------------------------------------------------------}
{ This is the 32 bit version of the TEnque class.                               }
{-------------------------------------------------------------------------------}

interface

uses Windows, SysUtils;

type

  EEnqueException = class(Exception);

  TEnque = class
  private
    FExceptionOnError: Boolean;               { True by default. Raises exception if Access request fails (rather than returning False }
    FExclusiveAccess: THandle;                { Handle to Mutex object }
    FLastError: LongInt;
    FLastMessage: string;
    FName: string;
    FNumberOfReaders: THandle;                { Handle to semaphore object which indicates the number of readers }
    FReadersActive: LongInt;                  { Handle to manual reset event which indicates if there are any readers currently active }
    FReadLocks: Integer;                      { Number of read locks this process has obtained }
    FReadTimeout: DWORD;
    FReadWithoutIntegrity: Boolean;  
    FUpdateLocks: Integer;                    { Number of update locks this process has obtained }
    FWriteTimeout: DWORD;
    procedure CriticalError(Code: LongInt);
    procedure CriticalErrorFmt(Code: LongInt; Dets: array of const);
    function Error(Code: LongInt): Boolean;
    function ErrorFmt(Code: LongInt; Dets: array of const): Boolean;
    function FormatErrorString(Error: LongInt; Dets: array of const ): string;
    function GetHaveReadAccess: Boolean;
    function GetHaveUpdateAccess: Boolean;
    function GetNumReaders: Integer;
    procedure Lock;
    procedure IndicateReadersAreActive;
    procedure IndicateReadersAreInactive;
    function SetStatusString(Code: LongInt): string;
    procedure SetReadWithoutIntegrity(Value: Boolean);
    procedure UnLock;
  public
    { Constructors / Destructors }
    constructor Create(Name: string);
    destructor Destroy; override;
    { Methods }
    function GetReadAccess: Boolean;
    function GetUpdateAccess: Boolean;
    function ReleaseReadAccess: Boolean;
    function ReleaseUpdateAccess: Boolean;
    { Properties }
    property ExceptionOnError: Boolean read FExceptionOnError write FExceptionOnError;
    property HaveReadAccess: Boolean read GetHaveReadAccess;
    property HaveUpdateAccess: Boolean read GetHaveUpdateAccess;
    property LastError: LongInt read FLastError;
    property LastMessage: string read FLastMessage;
    property Name: string read FName write FName;
    property ReadLocks: Integer read FReadLocks;
    property ReadTimeOut: DWORD read FReadTimeout write FReadTimeout;
    property ReadWithoutIntegrity: Boolean read FReadWithoutIntegrity write SetReadWithoutIntegrity;
    property UpdateLocks: Integer read FUpdateLocks;
    property WriteTimeOut: DWORD read FWriteTimeout write FWriteTimeout;

    { DEMO/DEBUG PROPERTIES }

    property NumReaders: Integer read GetNumReaders;
  end;

  { TEnque error codes }

implementation

uses Forms,
     Dialogs;

{$R ENQRES.RES}
{$I EnqComm.INC}

const
  DEFAULT_TIMEOUT = INFINITE;

constructor TEnque.Create(Name: string);
var
  WrkName: array[0..255] of char;

begin

 { Ensure there is enough room for required suffix }

  FName := Copy(Name,1,253);

  FExclusiveAccess := CreateMutex(nil,FALSE,StrPCopy(WrkName,FName+MUTEX_SUFFIX));

  if FExclusiveAccess = 0 then
    CriticalErrorFmt(ENQ_CREATE_FAILED,['Mutex',FName,GetLastError]);

  FNumberOfReaders := CreateSemaphore(nil,MAX_SEMAPHORES,MAX_SEMAPHORES,StrPCopy(WrkName,FName+SEMAPHORE_SUFFIX));

  if FNumberOfReaders = 0 then
    CriticalErrorFmt(ENQ_CREATE_FAILED,['Semaphore',FName,GetLastError]);

  FReadersActive := CreateEvent(nil,TRUE,TRUE,StrPCopy(WrkName,FName+EVENT_SUFFIX));

  if FReadersActive = 0 then
    CriticalErrorFmt(ENQ_CREATE_FAILED,['Event',FName,GetLastError]);

  FExceptionOnError := True;
  FReadTimeout := DEFAULT_TIMEOUT;
  FWriteTimeout := DEFAULT_TIMEOUT;

end;

destructor TEnque.Destroy;
begin

  if FNumberOfReaders <> 0 then
    begin
      While FReadLocks > 0 do
        ReleaseReadAccess;
      CloseHandle(FNumberOfReaders);
    end;

  if FExclusiveAccess <> 0 then
    CloseHandle(FExclusiveAccess);

  if FReadersActive <> 0 then
    CloseHandle(FReadersActive);

end;

procedure TEnque.CriticalError(Code: LongInt);
begin
  Raise EEnqueException.Create(SetStatusString(Code));
end;

procedure TEnque.CriticalErrorFmt(Code: LongInt; Dets: array of const);
begin
  Raise EEnqueException.Create(FormatErrorString(Code,Dets));
end;

function TEnque.Error(Code: LongInt): Boolean;
begin
  SetStatusString(Code);
  if ExceptionOnError then
    Raise EEnqueException.Create(FLastMessage);
  Result := False;
end;

function TEnque.ErrorFmt(Code: LongInt; Dets: array of const): Boolean;
begin
  FormatErrorString(Code,Dets);
  if ExceptionOnError then
    Raise EEnqueException.Create(FLastMessage);
  Result := False;
end;

function TEnque.FormatErrorString(Error: LongInt; Dets: array of const ): string;
begin
  FLastMessage := Format(SetStatusString(Error),Dets);
  Result := FLastMessage;
end;

function TEnque.GetHaveReadAccess: Boolean;
begin
  Result := FReadLocks > 0;
end;

function TEnque.GetHaveUpdateAccess: Boolean;
begin
  Result := FUpdateLocks > 0;
end;

{ This method has been implemented for demonstation and debugging purposes only }

function TEnque.GetNumReaders: Integer;
var
  PreviousCount: LongInt;

begin

  { Obtain then release the semaphore object to put the current count in PreviousCount }
  { Note: Due to the pre-emptive multitasking mode of Win95/NT, this function may not  }
  { return the correct value. Because, after ReleaseSemaphore another task may call    }
  { WaitForSingleObject to change the number of readers.                               }

  WaitForSingleObject(FNumberOfReaders,0);
  ReleaseSemaphore(FNumberOfReaders,1,@PreviousCount);
  Result := MAX_SEMAPHORES - PreviousCount - 1;
end;

function TEnque.GetReadAccess: Boolean;
var
  WaitResult: LongInt;
  WaitObjects: array[0..1] of LongInt;

begin
  if FReadWithoutIntegrity then
    begin
      Result := True;
      Exit; { Note: Ealry exit from routine }
    end;

  WaitObjects[0] := FExclusiveAccess;
  WaitObjects[1] := FNumberOfReaders;
  WaitResult := WaitForMultipleObjects(2,@WaitObjects,TRUE,FReadTimeout);
  case WaitResult of
    WAIT_OBJECT_0,WAIT_ABANDONED:
      begin
        IndicateReadersAreActive;
        Inc(FReadLocks);
        ReleaseMutex(FExclusiveAccess);
        Result := True;
        SetStatusString(ENQ_OK);
      end;

    WAIT_TIMEOUT:
      Result := ErrorFmt(ENQ_TIMEOUT_READ,[FName]);
  else
    Result := ErrorFmt(ENQ_READ_FAILED,[FName]);
  end;
end;

function TEnque.GetUpdateAccess: Boolean;
var
  WaitResult: LongInt;
  WaitObjects: array[0..1] of LongInt;

begin
  if FReadlocks > 0 then
    Result := ErrorFmt(ENQ_READ_ACCESS_DENIES_UPDATE,[FName])
  else
    begin
      WaitObjects[0] := FExclusiveAccess;
      WaitObjects[1] := FReadersActive;
      WaitResult := WaitForMultipleObjects(2,@WaitObjects,TRUE,FWriteTimeout);
      case Waitresult of
        WAIT_OBJECT_0,WAIT_ABANDONED:
          begin
            Inc(FUpdateLocks);
            SetStatusString(ENQ_OK);
            Result := True;
          end;
        WAIT_TIMEOUT:
          Result := ErrorFmt(ENQ_TIMEOUT_UPDATE,[FName]);
      else
        Result := ErrorFmt(ENQ_UPDATE_FAILED,[FName]);
      end;
    end;
end;

{---------------------------------------------------------------------}
{ Lock - Used to ensure that, when we are releasing semaphores, we    }
{        have exclusive access. This allows us to safely determine    }
{        if we are the last reader -  no other threads can change the }
{        count while we have the lock. This ensures that the manual   }
{        reset event (FReadersActive) is set correctly.               }
{---------------------------------------------------------------------}

procedure TEnque.Lock;
var
  WaitResult: LongInt;

begin

  WaitResult := WaitForSingleObject(FExclusiveAccess,Max_Maintenance_Delay_Expected);

  case WaitResult of

    WAIT_OBJECT_0,WAIT_ABANDONED:
      ; { We are OK }

  else

      begin

        { If this situation arises, there must be a logic error in processing.  }
        { We call Lock when we are decreasing the usage count for the read      }
        { semaphore. Whenever readers are active the Mutex object should be     }
        { available (except for locks obtained internally by this class for     }
        { maintenance). If we timeout waiting for this object something must    }
        { have gone wrong. Hence, we call CriticalError which raises and        }
        { exception regardless of the users setting of the ExceptionOnError     }
        { property.                                                             }

        CriticalErrorFmt(ENQ_INTERNAL_LOCK_LOGIC_ERROR,[FName]);

      end;
  end;

end;

procedure TEnque.IndicateReadersAreActive;
begin
  ResetEvent(FReadersActive);
end;

procedure TEnque.IndicateReadersAreInactive;
begin
  SetEvent(FReadersActive);
end;

function TEnque.ReleaseReadAccess: Boolean;
var
  PreviousCount: LongInt;

begin
  Result := False;
  if FReadWithoutIntegrity then
    begin
      Result := True;
      Exit; { Note: Ealry exit from routine }
    end;

  if FReadLocks = 0 then
    Result := ErrorFmt(ENQ_NO_CURRENT_READ_ACCESS,[FName])
  else
    begin
      { The lock ensures that we get an accurate value for the number of readers }
      Lock;
      try
        Dec(FReadLocks);
        ReleaseSemaphore(FNumberOfReaders,1,@PreviousCount);
        { If this was the last reader, mark event as available }
        if PreviousCount = LAST_READER then
          IndicateReadersAreInactive;
        SetStatusString(ENQ_OK);
        Result := True;
      finally
        UnLock;
      end;
    end;
end;

function TEnque.ReleaseUpdateAccess: Boolean;
begin
  if FUpdateLocks = 0 then
    Result := ErrorFmt(ENQ_NO_CURRENT_Update_ACCESS,[FName])
  else
    begin
      Dec(FUpdateLocks);
      ReleaseMutex(FExclusiveAccess);
      SetStatusString(ENQ_OK);
      Result := True;
    end;
end;

{ To get the string resource, the 3 MSBits (representing the message severity)  }
{ are masked to get the 'simple' error number. The base value is then added to  }
{ get the string resource id.                                                   }
{ Note: For the error code of ENQ_OK the string is simply set to ''. As this    }
{ method is called for all successful calls with ENQ_OK this saves the overhead }
{ of attempting to load a non-existent string resource.                         }

function TEnque.SetStatusString(Code: LongInt): string;
begin
  FLastError := Code;
  if Code <> ENQ_OK then
    FLastMessage := LoadStr((Code and CUSTOMER_MSG_MASK) + ENQ_STRING_RESOURCE_BASE)
  else
    FLastMessage := '';
  Result := FLastMessage;
end;

procedure TEnque.SetReadWithoutIntegrity(Value: Boolean);
begin
  if FReadLocks > 0 then
    Raise Exception.Create('Cannot change ReadWithoutIntegrity. Object already has read locks.')
  else
    FReadWithoutIntegrity := Value;
end;

{---------------------------------------------------------------------}
{ UnLock - For every call to Lock you MUST ensure that this routine   }
{          is called. To safely do this, this routine is executed in  }
{          the finally block of a Try Finally End constuct.           }
{---------------------------------------------------------------------}

procedure TEnque.UnLock;
begin
  ReleaseMutex(FExclusiveAccess);
end;

end.

