{$A+,B-,D+,F-,G+,I+,K+,L+,N+,P+,Q-,R-,S+,T-,V+,W-,X+,Y+}
{$M 16384,8192}
unit Enque16;

{-------------------------------------------------------------------------------}
{ This is the 16 bit version of the TEnque class.                               }
{-------------------------------------------------------------------------------}

interface

uses WinTypes, WinProcs, SysUtils;

type

  EEnqueException = class(Exception);

  TEnque = class
  private
    FExceptionOnError: Boolean;               { Raises exception if Access request fails (rather than return False }
    FExclusiveAccess: THandle;                { Handle to Mutex object }
    FLastError: LongInt;
    FLastMessage: string;
    FName: string;
    FNumberOfReaders: THandle;                { Handle to semaphore object which indicates the number of readers }
    FReadersActive: LongInt;                  { Handle to manual reset event which indicates if readers currently active }
    FReadLocks: Integer;                      { Number of read locks this process has obtained }
    FReadTimeout: LongInt;
    FReadWithoutIntegrity: Boolean; 
    FUpdateLocks: Integer;                    { Number of update locks this process has obtained }
    FUpdateTimeout: LongInt;
    procedure CriticalError(Code: LongInt);
    procedure CriticalErrorFmt(Code: LongInt; Dets: array of const);
    function Error(Code: LongInt): Boolean;
    function ErrorFmt(Code: LongInt; Dets: array of const): Boolean;
    function FormatErrorString(Error: LongInt; Dets: array of const ): string;
    function GetHaveReadAccess: Boolean;
    function GetHaveUpdateAccess: Boolean;
    function GetNumReaders: Integer;
    procedure Lock;
    procedure IndicateReadersAreActive;
    procedure IndicateReadersAreInactive;
    function SetStatusString(Code: LongInt): string;
    procedure SetReadWithoutIntegrity(Value: Boolean);
    procedure UnLock;
  public
    { Constructors / Destructors }
    constructor Create(Name: string);
    destructor Destroy; override;
    { Methods }
    function GetReadAccess: Boolean;
    function GetUpdateAccess: Boolean;
    function ReleaseReadAccess: Boolean;
    function ReleaseUpdateAccess: Boolean;
    { Properties }
    property ExceptionOnError: Boolean read FExceptionOnError write FExceptionOnError;
    property HaveReadAccess: Boolean read GetHaveReadAccess;
    property HaveUpdateAccess: Boolean read GetHaveUpdateAccess;
    property LastError: LongInt read FLastError;
    property LastMessage: string read FLastMessage;
    property Name: string read FName write FName;
    property ReadLocks: Integer read FReadLocks;
    property ReadTimeOut: LongInt read FReadTimeout write FReadTimeout;
    property ReadWithoutIntegrity: Boolean read FReadWithoutIntegrity write SetReadWithoutIntegrity;
    property UpdateLocks: Integer read FUpdateLocks;
    property WriteTimeOut: LongInt read FUpdateTimeout write FUpdateTimeout;

    { DEMO/DEBUG PROPERTIES }

    property NumReaders: Integer read GetNumReaders;
  end;

implementation

uses Forms,
     Thunk_32,
     Win32Def,
     Dialogs;

{$R ENQRES16.RES}
{$I ENQCOMM.INC}

const
  DEFAULT_TIMEOUT = INFINITE32; { Use this as 16bit defn is $FFFF which is approx 65 secs }

constructor TEnque.Create(Name: string);
var
  WrkName: array[0..255] of char;

begin

 { Ensure there is enough room for required suffix }

  FName := Copy(Name,1,253);

  FExclusiveAccess := CreateMutex(nil,FALSE,StrPCopy(WrkName,FName+MUTEX_SUFFIX),idCreateMutex);

  if FExclusiveAccess = 0 then
    CriticalErrorFmt(ENQ_CREATE_FAILED,['Mutex',FName,GetLastError(idGetLastError)]);

  FNumberOfReaders := CreateSemaphore(nil,MAX_SEMAPHORES,MAX_SEMAPHORES,
                        StrPCopy(WrkName,FName+SEMAPHORE_SUFFIX),idCreateSemaphore);

  if FNumberOfReaders = 0 then
    CriticalErrorFmt(ENQ_CREATE_FAILED,['Semaphore',FName,GetLastError(idGetLastError)]);

  FReadersActive := CreateEvent(nil,TRUE,TRUE,StrPCopy(WrkName,FName+EVENT_SUFFIX),idCreateEvent);

  if FReadersActive = 0 then
    CriticalErrorFmt(ENQ_CREATE_FAILED,['Event',FName,GetLastError(idGetLastError)]);

  FExceptionOnError := True;
  FReadTimeout := DEFAULT_TIMEOUT;
  FUpdateTimeout := DEFAULT_TIMEOUT;

end;

destructor TEnque.Destroy;
begin

  if FNumberOfReaders <> 0 then
    begin
      While FReadLocks > 0 do
        ReleaseReadAccess;
      CloseHandle(FNumberOfReaders,idCloseHandle);
    end;

  if FExclusiveAccess <> 0 then
    CloseHandle(FExclusiveAccess,idCloseHandle);

  if FReadersActive <> 0 then
    CloseHandle(FReadersActive,idCloseHandle);

end;

procedure TEnque.CriticalError(Code: LongInt);
begin
  Raise EEnqueException.Create(SetStatusString(Code));
end;

procedure TEnque.CriticalErrorFmt(Code: LongInt; Dets: array of const);
begin
  Raise EEnqueException.Create(FormatErrorString(Code,Dets));
end;

function TEnque.Error(Code: LongInt): Boolean;
begin
  SetStatusString(Code);
  if ExceptionOnError then
    Raise EEnqueException.Create(FLastMessage);
  Result := False;
end;

function TEnque.ErrorFmt(Code: LongInt; Dets: array of const): Boolean;
begin
  FormatErrorString(Code,Dets);
  if ExceptionOnError then
    Raise EEnqueException.Create(FLastMessage);
  Result := False;
end;

function TEnque.FormatErrorString(Error: LongInt; Dets: array of const ): string;
begin
  FLastMessage := Format(SetStatusString(Error),Dets);
  Result := FLastMessage;
end;

function TEnque.GetHaveReadAccess: Boolean;
begin
  Result := FReadLocks > 0;
end;

function TEnque.GetHaveUpdateAccess: Boolean;
begin
  Result := FUpdateLocks > 0;
end;

{ This method has been implemented for demonstation and debugging purposes only }

function TEnque.GetNumReaders: Integer;
var
  PreviousCount: LongInt;

begin

  { Obtain then release the semaphore object to put the current count in PreviousCount }
  { Note: Due to the pre-emptive multitasking mode of Win95/NT, this function may not  }
  { return the correct value. Because, after ReleaseSemaphore another task may call    }
  { WaitForSingleObject to change the number of readers.                               }

  WaitForSingleObject(FNumberOfReaders,0,idWaitForSingleObject);
  ReleaseSemaphore(FNumberOfReaders,1,@PreviousCount,idReleaseSemaphore);
  Result := MAX_SEMAPHORES - PreviousCount - 1;
end;

function TEnque.GetReadAccess: Boolean;
var
  WaitResult: LongInt;
  WaitObjects: array[0..1] of LongInt;
  StartTicks: LongInt;

begin
  if FReadWithoutIntegrity then
    begin
      Result := True;
      Exit; { Note: Ealry exit from routine }
    end;

  StartTicks := GetTickCount;
  WaitObjects[0] := FExclusiveAccess;
  WaitObjects[1] := FNumberOfReaders;
  While True do
    begin
      WaitResult := WaitForMultipleObjects(2,@WaitObjects,TRUE,0,idWaitForMultipleObjects);
      case WaitResult of
        WAIT_OBJECT_0,WAIT_ABANDONED:
          begin
            IndicateReadersAreActive;
            Inc(FReadLocks);
            ReleaseMutex(FExclusiveAccess,idReleaseMutex);
            Result := True;
            SetStatusString(ENQ_OK);
            Break; { Note: Loop exit }
          end;
        else
          begin
            if (FReadTimeout <> INFINITE32) and (GetTickCount - StartTicks > FReadTimeout) then
              begin
                Result := ErrorFmt(ENQ_TIMEOUT_READ,[FName]);
                Break; { Note: Loop exit }
              end;
          end;
      end;
      Sleep(50,idSleep);
      Application.ProcessMessages;
    end;
end;

function TEnque.GetUpdateAccess: Boolean;
var
  WaitResult: LongInt;
  WaitObjects: array[0..1] of LongInt;
  StartTicks: LongInt;

begin
  if FReadlocks > 0 then
    Result := ErrorFmt(ENQ_READ_ACCESS_DENIES_UPDATE,[FName])
  else
    begin
      StartTicks := GetTickCount;
      WaitObjects[0] := FExclusiveAccess;
      WaitObjects[1] := FReadersActive;
      While True do
        begin
          WaitResult := WaitForMultipleObjects(2,@WaitObjects,TRUE,0,idWaitForMultipleObjects);
          case Waitresult of
            WAIT_OBJECT_0,WAIT_ABANDONED:
              begin
                Inc(FUpdateLocks);
                SetStatusString(ENQ_OK);
                Result := True;
                Break;
              end;
            else
              begin
                if (FUpdateTimeout <> INFINITE32) and (GetTickCount - StartTicks > FUpdateTimeout) then
                  begin
                    Result := ErrorFmt(ENQ_TIMEOUT_UPDATE,[FName]);
                    Break;
                  end;
              end;
          end;
          Sleep(50,idSleep);
          Application.ProcessMessages;
        end;
    end;
end;

{---------------------------------------------------------------------}
{ Lock - Used to ensure that, when we are releasing semaphores, we    }
{        have exclusive access. This allows us to safely determine    }
{        if we are the last reader -  no other threads can change the }
{        count while we have the lock. This ensures that the manual   }
{        reset event (FReadersActive) is set correctly.               }
{---------------------------------------------------------------------}

procedure TEnque.Lock;
var
  WaitResult: LongInt;
  StartTicks: LongInt;

begin
  StartTicks := GetTickCount;
  While True do
    begin
      WaitResult := WaitForSingleObject(FExclusiveAccess,0,idWaitForSingleObject);

      case WaitResult of
        WAIT_OBJECT_0,WAIT_ABANDONED:
            Break;
      else
        begin
          if GetTickCount - StartTicks > Max_Maintenance_Delay_Expected then

            { If this situation arises, there must be a logic error in    }
            { processing. We call Lock when we are decreasing the usage   }
            { count for the read semaphore. Whenever readers are active   }
            { the Mutex object should be available. If we timeout waiting }
            { for this object something must have gone wrong. Hence, we   }
            { call CriticalError which raises and exception regardless of }
            { of the users setting of the ExceptionOnError property.      }

            CriticalErrorFmt(ENQ_INTERNAL_TIMEOUT_LOGIC_ERROR,[FName]);

        end;
      end;
      Sleep(50,idSleep);
      Application.ProcessMessages;
    end;
end;

procedure TEnque.IndicateReadersAreActive;
begin
  ResetEvent(FReadersActive,idResetEvent);
end;

procedure TEnque.IndicateReadersAreInactive;
begin
  SetEvent(FReadersActive,idSetEvent);
end;

function TEnque.ReleaseReadAccess: Boolean;
var
  PreviousCount: LongInt;

begin
  Result := False;
  if FReadWithoutIntegrity then
    begin
      Result := True;
      Exit; { Note: Ealry exit from routine }
    end;

  if FReadLocks = 0 then
    Result := ErrorFmt(ENQ_NO_CURRENT_READ_ACCESS,[FName])
  else
    begin
      { The lock ensures that we get an accurate value for the number of readers }
      Lock;
      try
        Dec(FReadLocks);
        ReleaseSemaphore(FNumberOfReaders,1,@PreviousCount,idReleaseSemaphore);
        { If this was the last reader, mark event as available }
        if PreviousCount = LAST_READER then
          IndicateReadersAreInactive;
        SetStatusString(ENQ_OK);
        Result := True;
      finally
        UnLock;
      end;
    end;
end;

function TEnque.ReleaseUpdateAccess: Boolean;
begin
  if FUpdateLocks = 0 then
    Result := ErrorFmt(ENQ_NO_CURRENT_Update_ACCESS,[FName])
  else
    begin
      Dec(FUpdateLocks);
      ReleaseMutex(FExclusiveAccess,idReleaseMutex);
      SetStatusString(ENQ_OK);
      Result := True;
    end;
end;

{ To get the string resource, the 3 MSBits (representing the message severity)  }
{ are masked to get the 'simple' error number. The base value is then added to  }
{ get the string resource id.                                                   }
{ Note: For the error code of ENQ_OK the string is simply set to ''. As this    }
{ method is called for all successful calls with ENQ_OK this saves the overhead }
{ of attempting to load a non-existent string resource.                         }

function TEnque.SetStatusString(Code: LongInt): string;
begin
  FLastError := Code;
  if Code <> ENQ_OK then
    FLastMessage := LoadStr((Code and CUSTOMER_MSG_MASK) + ENQ_STRING_RESOURCE_BASE)
  else
    FLastMessage := '';
  Result := FLastMessage;
end;

procedure TEnque.SetReadWithoutIntegrity(Value: Boolean);
begin
  if FReadLocks > 0 then
    Raise Exception.Create('Cannot change ReadWithoutIntegrity. Object already has read locks.')
  else
    FReadWithoutIntegrity := Value;
end;

{---------------------------------------------------------------------}
{ UnLock - For every call to Lock you MUST ensure that this routine   }
{          is called. To safely do this, this routine is executed in  }
{          the finally block of a Try Finally End constuct.           }
{---------------------------------------------------------------------}

procedure TEnque.UnLock;
begin
  ReleaseMutex(FExclusiveAccess,idReleaseMutex);
end;

end.

