unit ODButton;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls;

type
  TOwnerDrawButton = class;

  TDrawButtonEvent = procedure(Button: TOwnerDrawButton;
    Canvas: TCanvas; Rect: TRect; State: TOwnerDrawState;
    Default: Boolean) of object;

  TOwnerDrawButton = class(TButton)
  private
    FCanvas: TCanvas;
    FColor: TColor;
    FDefault,
    FOwnerDraw: Boolean;
    FOnDrawButton: TDrawButtonEvent;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CNDrawItem(var Msg: TWMDrawItem); message cn_DrawItem;
    procedure WMLButtonDblClk(var Msg: TWMLButtonDblClk);
      message wm_LButtonDblClk;
    procedure DrawButton(Canvas: TCanvas; const Rect: TRect;
      State: TOwnerDrawState);
    procedure SetButtonStyle(ADefault: Boolean); override;
    procedure SetColor(Value: TColor);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetOwnerDraw(Value: Boolean);
  published
    property Color: TColor read FColor write SetColor default clBtnFace;
    property OwnerDraw: Boolean read FOwnerDraw write SetOwnerDraw;
    property OnDrawButton: TDrawButtonEvent
      read FOnDrawButton write FOnDrawButton;
  end;

procedure Register;

implementation

constructor TOwnerDrawButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCanvas := TCanvas.Create;
  FColor := clBtnFace;
end;

destructor TOwnerDrawButton.Destroy;
begin
 FCanvas.Free;
 inherited Destroy;
end;

procedure TOwnerDrawButton.CreateParams(var Params: TCreateParams);
const
  Styles: array[Boolean] of Longint = (0, bs_OwnerDraw);
begin
  inherited CreateParams(Params);
  with Params do
    Style := Style or Styles[FOwnerDraw];
end;

procedure TOwnerDrawButton.CNDrawItem(var Msg: TWMDrawItem);
var
  State: TOwnerDrawState;
begin
  with Msg.DrawItemStruct^ do
  begin
    FCanvas.Handle := HDC;
    FCanvas.Font := Font;
    FCanvas.Brush := Brush;
    FCanvas.Brush.Color := FColor;
    { The Value typecast to Word is for cross-platform }
    { compatibility. In Win32 ItemState is a Longint. The }
    { typecast gives the low word, which is what we want }
    State := TOwnerDrawState(WordRec(Word(ItemState)).Lo);
    DrawButton(FCanvas, rcItem, State)
  end;
end;

procedure TOwnerDrawButton.WMLButtonDblClk(var Msg: TWMLButtonDblClk);
begin
  Perform(wm_LButtonDown, Msg.Keys, Longint(Msg.Pos));
end;

procedure TOwnerDrawButton.DrawButton(Canvas: TCanvas;
  const Rect: TRect; State: TOwnerDrawState);
var
  R: TRect;
  Text: array[0..255] of Char;

  procedure FocusRect;
  begin
    if odFocused in State then
    begin
      R := Rect;
      InflateRect(R, -4, -4);
      Canvas.DrawFocusRect(R);
    end;
  end;

  procedure DefaultRect;
  begin
    if FDefault then
    begin
      Canvas.Brush.Color := clWindowFrame;
      Canvas.FrameRect(Rect);
    end;
  end;

begin
  if Assigned(FOnDrawButton) then
    FOnDrawButton(Self, Canvas, Rect, State, FDefault)
  else
  { This draws button as in a 16-bit app under Windows 95 }
  with Canvas do
  begin
    FillRect(Rect);
    R := Rect;
    { How big is the text? }
    DrawText(Handle, StrPCopy(Text, Caption), -1, R,
      dt_Center or dt_WordBreak or dt_CalcRect);
    { Center this calculated rectangle }
    OffsetRect(R, (Rect.Right - R.Right) div 2,
                  (Rect.Bottom - R.Bottom) div 2);
    { When depressed, text is slightly offset }
    if odSelected in State then
      OffsetRect(R, 1, 1);
    { Write caption }
    if Enabled then
      DrawText(Handle, Text, -1, R, dt_Center or dt_WordBreak)
    else
    begin
      Brush.Style := bsClear;
      OffsetRect(R, 1, 1);
      Font.Color := clBtnHighlight;
      DrawText(Handle, Text, -1, R, dt_Center or dt_WordBreak);
      OffsetRect(R, -1, -1);
      Font.Color := clBtnShadow;
      DrawText(Handle, Text, -1, R, dt_Center or dt_WordBreak);
      Brush.Style := bsSolid;
    end;
    if odSelected in State then
    begin
      { Draw dotted focus rect }
      FocusRect;
      { Draw outer black rect }
      DefaultRect;
      R := Rect;
      InflateRect(R, -1, -1);
      Brush.Color := clBtnShadow;
      { Draw inner gray rect }
      FrameRect(R);
      InflateRect(R, -1, -1);
      Brush.Color := clBtnFace;
      { Draw innermost light gray rect }
      FrameRect(R);
    end
    else
    begin
      R := Rect;
      if FDefault then
        { Can't do the frame rect here as for some }
        { reason it prevents the focus rect at the end }
        InflateRect(R, -1, -1);
      { Draw black lines on bottom and at right }
      Pen.Color := clWindowFrame;
      MoveTo(R.Right - 1, R.Top);
      LineTo(R.Right - 1, R.Bottom - 1);
      LineTo(R.Left - 1, R.Bottom - 1);
      { Draw light gray lines on top and at left }
      Pen.Color := clBtnFace;
      MoveTo(R.Left, R.Bottom - 2);
      LineTo(R.Left, R.Top);
      LineTo(R.Right - 2, R.Top);
      InflateRect(R, -1, -1);
      { Draw dark gray lines on bottom and at right }
      Pen.Color := clBtnShadow;
      MoveTo(R.Right - 1, R.Top);
      LineTo(R.Right - 1, R.Bottom - 1);
      LineTo(R.Left - 1, R.Bottom - 1);
      { Draw white lines on top and at left }
      Pen.Color := clBtnHighlight;
      MoveTo(R.Left, R.Bottom - 2);
      LineTo(R.Left, R.Top);
      LineTo(R.Right - 2, R.Top);
      { Do dotted focus rectange if need be }
      FocusRect;
      { Do outer black rectange if need be }
      DefaultRect;
    end;
  end;
end;

procedure TOwnerDrawButton.SetButtonStyle(ADefault: Boolean);
begin
  { Normally, when focus is received, TButton changes }
  { its style to give the thicker black border. But }
  { when owner drawn, you are not allowed other styles }
  if FOwnerDraw then
  begin
    FDefault := ADefault;
    Refresh;
  end
  else
    inherited SetButtonStyle(ADefault);
end;

procedure TOwnerDrawButton.SetColor(Value: TColor);
begin
  if FColor <> Value then
  begin
    FColor := Value;
    if not FOwnerDraw then
      OwnerDraw := True
    else
      InValidate;
  end;
end;

procedure TOwnerDrawButton.SetOwnerDraw(Value: Boolean);
begin
  if FOwnerDraw <> Value then
  begin
    FOwnerDraw := Value;
    RecreateWnd;
  end;
end;

procedure Register;
begin
  RegisterComponents('Clinic', [TOwnerDrawButton]);
end;

end.
