unit ExpComp;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs;

const
  aryFunctions: array[1..10] of string =
     ('ABS', 'ARCTAN', 'COS', 'EXP', 'FRAC', 'INT', 'LN', 'SIN', 'SQR', 'SQRT');

type
  TExpressionException = class(Exception);

  TOperators = set of char;

  OperandType = (Constant, Variable);

  TOperandInfo = packed record
    case OpType: OperandType of
       Constant: (OperandValue: Extended);
       Variable: (OperandName: string[30]);
  end;

  TSymbolInfo = packed record
    Name:      string[30];
    Operator:  Char;
    Operand:   array[1..2] of TOperandInfo;
    ApplyFunc: string[30];
  end;

  TOperand = class(TComponent)
  private
    FValue:    Extended;
    FOnChange: TNotifyEvent;
  protected
    procedure  SetValue(NewValue: Extended);
    procedure  Changed; virtual;
  public
    destructor Destroy; override;
    property   OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Value: Extended read FValue write SetValue;
  end;

  TRoundingStates = (rsAlwaysRoundDown, rsAlwaysRoundUp, rsNoRounding, rsStandardRounding);

  TExpression = class(TComponent)
  private
    FSymbols:           TList;
    FOperands:          TStringList;
    FSymCount:          Word;

    FExpression:        string;
    FNamedExpressions:  TStrings;
    FPrecision:         Word;
    FResult:            Extended;
    FRoundingState:     TRoundingStates;
    FRoundToNearest:    Extended;
    FSilentExceptions:  Boolean;
    FValidExpression:   Boolean;

    FBusy:              Boolean;

    FOnCalculated: TNotifyEvent;
    FOnException:  TExceptionEvent;

    function  Power(Base, Exponent: Extended): Extended;
    function  GenerateNewSymbol(Operand1, Operand2: string; Op: Char): string;
    procedure StripChar(var Str: string; Ch: string);
    function  PriorOperator(Expression: string; CurrOp: Word): Word;
    function  NextOperator(Expression: string; CurrOp: Word): Word;
    procedure ResolveUnaryOperators(var Str: string);
    procedure NormalizeOperator(var Str: string; SrcOp, DestOp: string);
    procedure NormalizeExpression(var Expression: string);
    function  IsConstantValue(Expression: string): Boolean;
    function  DecodeSubExpression(Expression: string): string;
    procedure DecodeExpression(var Expression: string);
    procedure ParseFunctions(var Expression: string);

    procedure SetExpression(Value: string);
    procedure SetFileName(Value: string);
    procedure SetPrecision(Value: Word);
    procedure SetRoundingState(Value: TRoundingStates);
    procedure SetRoundToNearest(Value: Extended);
    procedure SetNamedExpressions(Value: TStrings);

    function  GetFileName: string;

    procedure RegisterOperandChange(Sender: TObject);
    procedure RegisterException(Sender: TObject; E: Exception);
    procedure Calculated;
  protected
    procedure WaxAllSymbols;
    procedure WaxAllOperands;
    procedure AddOperand(Name: string; Value: Extended);
    procedure CalculateResult;

  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Loaded; override;

    function  OperandCount: Word;
    function  GetOperandAt(Index: Word): TOperand;
    function  GetNamedExpression(ExpName: string): string;
    procedure SetOperandValueByName(OpName: string; NewValue: Extended);

    property FileName:          string          read GetFileName        write SetFileName;
  published
    property Expression:        string          read FExpression        write SetExpression;
    property NamedExpressions:  TStrings        read FNamedExpressions  write SetNamedExpressions;
    property Precision:         Word            read FPrecision         write SetPrecision;
    property Result:            Extended        read FResult            write FResult;
    property RoundingState:     TRoundingStates read FRoundingState     write SetRoundingState;
    property RoundToNearest:    Extended        read FRoundToNearest    write SetRoundToNearest;
    property SilentExceptions:  Boolean         read FSilentExceptions  write FSilentExceptions;
    property ValidExpression:   Boolean         read FValidExpression   write FValidExpression;

    property OnCalculated:      TNotifyEvent    read FOnCalculated      write FOnCalculated;
    property OnException:       TExceptionEvent read FOnException       write FOnException;
  end;

implementation

destructor TOperand.Destroy;
begin
  FOnChange := nil;
  inherited Destroy;
end;

procedure TOperand.Changed;
begin
  if Assigned(FOnChange) then
     FOnChange(Self);
end;

procedure TOperand.SetValue(NewValue: Extended);
begin
  if (FValue = NewValue) then
      exit;

  FValue := NewValue;

  Changed;
end;

procedure TExpression.Calculated;
begin
  if Assigned(FOnCalculated) then
     FOnCalculated(Self);
end;

procedure TExpression.RegisterOperandChange(Sender: TObject);
begin
  try
    if (Sender as TOperand).Value <> -999999999 then
        CalculateResult;
  except
    on E: Exception do
       RegisterException(Self, E);
  end;
end;

procedure TExpression.RegisterException(Sender: TObject; E: Exception);
begin
  if not FSilentExceptions then
     Application.HandleException(Sender);

  if Assigned(FOnException) then
     FOnException(Self, E);
end;

constructor TExpression.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FSymbols           := TList.Create;
  FOperands          := TStringList.Create;
  FNamedExpressions  := TStringList.Create;

  FSymCount          := 0;
  FExpression        := '';
  FPrecision         := 0;
  FResult            := 0;
  FRoundingState     := rsNoRounding;
  FRoundToNearest    := 0;
  FSilentExceptions  := False;
  FValidExpression   := True;

  FBusy              := False;

  { Weird problem with Delphi 1.0; First call to StrToFloat in design-mode
    yields EConvertError, even on a valid floating-point number.  I issue
    a StrToFloat here in anticipation of the EConvertError so that future
    calls will be successful.                                             }

  try
    StrToFloat('0.1');
  except
    on EConvertError do;
  end;
end;

destructor TExpression.Destroy;
begin
  FOnCalculated := nil;
  FOnException  := nil;

  FNamedExpressions.Free;

  WaxAllSymbols;
  FSymbols.Free;

  WaxAllOperands;
  FOperands.Free;

  inherited Destroy;
end;

procedure TExpression.Loaded;
var
  TempStringList: TStringList;
  ExpCount: Integer;
  Name: string;
  Exp: string;
begin
  inherited Loaded;

  if (csDesigning in ComponentState) then
      exit;

  if (FNamedExpressions.Count <> 1) then
      exit;

  TempStringList := TStringList.Create;

  try
    try
      TempStringList.LoadFromFile(FNamedExpressions[0]);

      FNamedExpressions.Clear;

      for ExpCount := 0 to TempStringList.Count - 1 do
         begin
           Name := Copy(TempStringList[ExpCount], 1, Pos('=', TempStringList[ExpCount]) - 1);
           Exp  := Copy(TempStringList[ExpCount], Pos('=', TempStringList[ExpCount]) + 1, 255);

           FNamedExpressions.Add(UpperCase(Name));
           FNamedExpressions.Add(Exp);
         end;
    finally
      TempStringList.Free;
    end;
  except
    on E: Exception do
       RegisterException(Self, E);
  end;
end;

function TExpression.GetNamedExpression(ExpName: string): string;
var
  Idx: Integer;
begin
  Result := '';

  Idx := NamedExpressions.IndexOf(UpperCase(ExpName));

  if (Idx = -1) then
      exit;

  Result := NamedExpressions[Idx + 1];
end;

procedure TExpression.WaxAllSymbols;
var
  Idx: Integer;
begin
  if (FSymbols.Count > 0) then
      for Idx := 0 to FSymbols.Count - 1 do
        FreeMem(FSymbols[Idx], SizeOf(TSymbolInfo));

  FSymbols.Clear;
end;

procedure TExpression.WaxAllOperands;
var
  Idx: Integer;
begin
  if (FOperands.Count > 0) then
      for Idx := 0 to FOperands.Count - 1 do
        TOperand(FOperands.Objects[Idx]).Free;

  FOperands.Clear;
end;

procedure TExpression.AddOperand(Name: string; Value: Extended);
var
  i: Integer;
  Operand: TOperand;
begin
  i := FOperands.IndexOf(UpperCase(Name));

  if (i <> -1) then
      begin
        TOperand(FOperands.Objects[i]).Value := Value;
        exit;
      end;

  Operand := TOperand.Create(Self);

  Operand.Name     := Name;
  Operand.OnChange := RegisterOperandChange;

  FOperands.AddObject(Name, Operand);

  TOperand(FOperands.Objects[FOperands.Count - 1]).Value := Value;
end;

procedure TExpression.SetExpression(Value: string);
begin
  if FExpression = Value then
     exit;

  FExpression      := Value;

  FResult          := 0;
  FSymCount        := 0;
  FValidExpression := True;

  WaxAllSymbols;
  WaxAllOperands;

  try
    if (FExpression <> '') then
        begin
          try
            DecodeExpression(Value);
          except
            on Exception do
               begin
                 FValidExpression := False;
                 Raise;
               end;
          end;

          if (FOperands.Count = 0) then
              CalculateResult;
        end;
  except
    on E: Exception do
       RegisterException(Self, E);
  end;
end;

procedure TExpression.SetFileName(Value: string);
begin
  if (FNamedExpressions.Count > 0) then
      if FNamedExpressions[0] = ExtractFileName(Value) then
         exit;

  FNamedExpressions.Clear;

  FNamedExpressions.Add(ExtractFileName(Value));
end;

function TExpression.GetFileName: string;
begin
  Result := '';

  if (FNamedExpressions.Count <> 1) then
      exit;

  Result := FNamedExpressions[0];
end;

procedure TExpression.SetPrecision(Value: Word);
begin
  if (Value <> 0) AND ((Value < 2) OR (Value > 8)) then
      if (FSilentExceptions) then
          exit
      else
          raise TExpressionException.Create('Precision must either be 0 OR in the range of 2 to 8');

  if (FPrecision = Value) then
      exit;

  FPrecision := Value;

  try
    CalculateResult;
  except
    on E: Exception do
       RegisterException(Self, E);
  end;
end;

procedure TExpression.SetRoundingState(Value: TRoundingStates);
begin
  if (FRoundingState = Value) then
      exit;

  FRoundingState := Value;

  try
    CalculateResult;
  except
    on E: Exception do
       RegisterException(Self, E);
  end;
end;

procedure TExpression.SetRoundToNearest(Value: Extended);
begin
  if (Value < 0) or (Value > 0.99) then
      if (FSilentExceptions) then
          exit
      else
          raise TExpressionException.Create('RoundToNearest must be within range of 0 to .99');

  if (FRoundToNearest = Value) then
      exit;

  FRoundToNearest := Value;

  try
    CalculateResult;
  except
    on E: Exception do
       RegisterException(Self, E);
  end;
end;

procedure TExpression.SetNamedExpressions(Value: TStrings);
begin
  FNamedExpressions.Assign(Value);
end;

function TExpression.OperandCount: Word;
begin
  Result := FOperands.Count;
end;

function TExpression.GetOperandAt(Index: Word): TOperand;
begin
  if (FOperands.Count = 0) then
      Result := nil
  else
      Result := TOperand(FOperands.Objects[Index]);
end;

procedure TExpression.SetOperandValueByName(OpName: string; NewValue: Extended);
var
  Index: Integer;
  Ready: Boolean;
begin
  Index := FOperands.IndexOf(UpperCase(OpName));

  if (Index <> -1) then
      TOperand(FOperands.Objects[Index]).Value := NewValue;
end;

procedure TExpression.CalculateResult;
var
  Rslt:      Extended;
  i:         Integer;
  Op1, Op2:  Extended;
  Oper:      Char;
  Symbol:    ^TSymbolInfo;
begin
  if (FBusy) or (not FValidExpression) then
      exit;

  FBusy := True;

  try
    if (FOperands.Count > 0) then
        for i := 0 to FOperands.Count - 1 do
           if TOperand(FOperands.Objects[i]).Value = -999999999 then
              exit;

    Rslt := 0;

    for i := 0 to FSymbols.Count - 1 do
      begin
        Symbol := FSymbols[i];

        with Symbol^ do
          begin
            case Operand[1].OpType of
              Constant: Op1 := Operand[1].OperandValue;
              Variable: Op1 := TOperand(FOperands.Objects[FOperands.IndexOf(UpperCase(Operand[1].OperandName))]).Value;
            end;

            Oper := Operator;

            case Operand[2].OpType of
               Constant: Op2 := Operand[2].OperandValue;
               Variable: Op2 := TOperand(FOperands.Objects[FOperands.IndexOf(UpperCase(Operand[2].OperandName))]).Value;
            end;
          end;

        case Ord(Oper) of
  { Exp } Ord('^') : Rslt := Power(Op1, Op2);

  { Mul } Ord('*') : Rslt := Op1 * Op2;

  { Div } Ord('/') : Rslt := Op1 / Op2;

  { Mod } Ord('%') : Rslt := Frac(Op1 / Op2); { MOD is Integer only }

  { Add } Ord('+') : Rslt := Op1 + Op2;

  { Sub } Ord('-') : Rslt := Op1 - Op2;
          else
             raise TExpressionException.Create('Operator is not valid');
        end;

        if (Symbol^.ApplyFunc <> '') then
            begin
              if (Symbol^.ApplyFunc = 'ABS') then
                  Rslt := Abs(Rslt)
              else if (Symbol^.ApplyFunc = 'ARCTAN') then
                       Rslt := ArcTan(Rslt)
              else if (Symbol^.ApplyFunc = 'COS') then
                       Rslt := Cos(Rslt)
              else if (Symbol^.ApplyFunc = 'EXP') then
                       Rslt := Exp(Rslt)
              else if (Symbol^.ApplyFunc = 'FRAC') then
                       Rslt := Frac(Rslt)
              else if (Symbol^.ApplyFunc = 'INT') then
                       Rslt := Int(Rslt)
              else if (Symbol^.ApplyFunc = 'LN') then
                       Rslt := Ln(Rslt)
              else if (Symbol^.ApplyFunc = 'SIN') then
                       Rslt := Sin(Rslt)
              else if (Symbol^.ApplyFunc = 'SQRT') then
                       Rslt := Sqrt(Rslt)
              else if (Symbol^.ApplyFunc = 'SQR') then
                       Rslt := Sqr(Rslt);
            end;

        AddOperand(Symbol^.Name, Rslt);
      end;

    if (FPrecision > 0) then
        Rslt := Trunc(Rslt) + (System.Round(Frac(Rslt) * Power(10, FPrecision)) / Power(10, FPrecision));

    if (FRoundingState = rsNoRounding) then
        Result := Rslt
    else
        if (FRoundToNearest = 0) then
            begin
              if (FRoundingState = rsStandardRounding) then
                  Result := System.Round(Rslt)
              else
                  begin
                    Result := Trunc(Rslt);

                    if (Frac(Rslt) > 0) and (FRoundingState = rsAlwaysRoundUp) then
                        Result := Result + 1;
                  end;
            end
        else
            begin
              if (FRoundingState = rsStandardRounding) then
                  Result := Trunc(Rslt) + (System.Round(Frac(Rslt) / FRoundToNearest) * FRoundToNearest)
              else
                  begin
                    Result := Trunc(Rslt) + (Trunc(Frac(Rslt) / FRoundToNearest) * FRoundToNearest);

                    if (FRoundingState = rsAlwaysRoundUp) and (Frac(Rslt) > 0) then
                        Result := Result + FRoundToNearest;
                  end;
            end;
  finally
    FBusy := False;
    Calculated;
  end;
end;

function TExpression.Power(Base, Exponent: Extended): Extended;
begin
  if (Base = 0.0) then
      begin
        if (Exponent < 0.0) then
            raise EInvalidOp.Create('Zero raised to a negative power is undefined');

        if (Exponent = 0.0) then
            raise EInvalidOp.Create('Zero raised to Zero power (0 ^ 0) is undefined');

        Result := 0.0;
      end
  else
      begin
        if (Base < 0.0) and (Frac(Exponent) <> 0.0) then
            raise EInvalidOp.Create('Negative base raised to a non-integral power is undefined');

        Result := Exp(Ln(Abs(Base)) * Exponent);

        {  Negative Base with Odd Exponent  }

        if (Base < 0.0) and (Frac(Exponent / 2.0) <> 0.0) then
            Result := -Result;
      end;
end;

function TExpression.GenerateNewSymbol(Operand1, Operand2: string; Op: Char): string;
var
  Symbol:   ^TSymbolInfo;
  Operand:  TOperand;
  Op1Type:  OperandType;
  Op2Type:  OperandType;
  WorkExt:  Extended;
begin
  if (Operand1 = '') then
      Operand1 := '0';

  if (Operand2 = '') then
      Operand2 := '0';

  if (Op = '') then
      Op := '+';

  try
     WorkExt := StrToFloat(Operand1);
     Op1Type := Constant;
  except
     on EConvertError do
        Op1Type := Variable;
  end;

  try
     WorkExt := StrToFloat(Operand2);
     Op2Type := Constant;
  except
     on EConvertError do
        Op2Type := Variable;
  end;

  GetMem(Symbol, SizeOf(TSymbolInfo));

  with Symbol^ do
    begin
      with Operand[1] do
        begin
          OpType := Op1Type;

          case OpType of
             Constant: OperandValue := StrToFloat(Operand1);  { Value }

             Variable: begin
                         OperandName  := Operand1;            { Name  }

                         if Pos('SYS_GEN_SYM_', Operand1) = 0 then
                            AddOperand(Operand1, -999999999);
                       end;
          end;
        end;

      Operator := Op;

      with Operand[2] do
        begin
          OpType := Op2Type;

          case OpType of
             Constant: OperandValue := StrToFloat(Operand2);  { Value }

             Variable: begin
                         OperandName  := Operand2;            { Name  }

                         if Pos('SYS_GEN_SYM_', Operand2) = 0 then
                            AddOperand(Operand2, -999999999);
                       end;
          end;
        end;

      Inc(FSymCount);

      Name := 'SYS_GEN_SYM_' + IntToStr(FSymCount);
      Result := Name;

      try
        ApplyFunc := '';
      except
        on E: Exception do
           ShowMessage('Exception occurred; Msg is: ' + E.Message);
      end;
    end;

  try
    FSymbols.Add(Symbol);
  except
    on E: Exception do
       ShowMessage('Exception occurred; Msg is: ' + E.Message);
  end;
end;

procedure TExpression.StripChar(var Str: string; Ch: string);
var
  ChPos: byte;
begin
  ChPos := Pos(Ch, Str);

  while ChPos > 0 do
    begin
      Delete(Str, ChPos, 1);

      ChPos := Pos(Ch, Str);
    end;
end;

function TExpression.PriorOperator(Expression: string; CurrOp: Word): Word;
begin
  for Result := CurrOp - 1 downto 1 do
     if Expression[Result] in ['^', '*', '/', '%', '+', '-', '#'] then
        exit;
end;

function TExpression.NextOperator(Expression: string; CurrOp: Word): Word;
begin
  for Result := CurrOp + 1 to Length(Expression) do
     if Expression[Result] in ['^', '*', '/', '%', '+', '-', '#'] then
        exit;
end;

procedure TExpression.ResolveUnaryOperators(var Str: string);
var
  i:       Word;
  PriorOp: Word;
  NextOp:  Word;
  Operand: string;
begin
  {  Flank the entire expression with pound (#) delimiters.  }
  {  This is required for operator sub-parsing.              }

  Insert('#', Str, 1);
  Insert('#', Str, Length(Str)+1);

  {  Walk expression looking for unary operatpors            }

  i := 1;

  repeat
    if Str[i] in ['-', '+'] then
       begin
         PriorOp := PriorOperator(Str, i);

         if (i - PriorOp) = 1 then
             begin
               NextOp  := NextOperator(Str, i);
               Operand := Copy(Str, i, NextOp - i);

               Delete(Str, i, NextOp - i);
               Insert('(' + Operand + ')', Str, i);

               i   := (NextOp - i) + 2;
             end;
       end;

    Inc(i);

  until (i > Length(Str));

  {  Strip the pound (#) delimiters back out  }

  StripChar(Str, '#');
end;

procedure TExpression.NormalizeOperator(var Str: string; SrcOp, DestOp: string);
var
  OpPos: Word;
begin
  OpPos := Pos(SrcOP, Str);

  while OpPos > 0 do
    begin
      Delete(Str, OpPos, Length(SrcOp));
      Insert(DestOp, Str, OpPos);

      OpPos := Pos(SrcOp, Str);
    end;
end;

procedure TExpression.NormalizeExpression(var Expression: string);
var
  i:   Word;
  Cnt: Integer;
begin
  {  Are parentheses correctly paired?                          }

  if (Pos('(', Expression) > 0) or (Pos(')', Expression) > 0) then
      begin
        Cnt := 0;

        for i := 1 to Length(Expression) do
          begin
            if Expression[i] = '(' then
               Inc(Cnt);

            if Expression[i] = ')' then
               begin
                 Dec(Cnt);

                 if (Cnt < 0) then
                     break;
              end;
          end;

        if (Cnt <> 0) then
            raise TExpressionException.Create('Expression has un-matched parentheses');
      end;

  { Convert all operators to their single-character equivalent  }

  NormalizeOperator(Expression, '**',   '^');
  NormalizeOperator(Expression, 'DIV ', '/');
  NormalizeOperator(Expression, 'MOD ', '%');

  { Strip all blanks from Expression                            }

  StripChar(Expression, ' ');

  { Parse Function Calls                                        }

  ParseFunctions(Expression);

  {  Surround all unary operators with parentheses              }

  ResolveUnaryOperators(Expression);
end;

function TExpression.IsConstantValue(Expression: string): boolean;
var
  WorkExt: Extended;
begin
  try
     WorkExt := StrToFloat(Expression);
     Result := True;
  except
     on EConvertError do
        Result := False;
  end;
end;

procedure TExpression.ParseFunctions(var Expression: string);
var
  fncPos, TempPos: Integer;
  fncIdx: Integer;
  Cnt: Integer;
  Idx: Integer;
  LeftParen: Integer;
  RightParen: Integer;
  SubExp: string;
begin
  repeat
    fncPos := 0;

    for Idx := Low(aryFunctions) to High(aryFunctions) do
       begin
         TempPos := Pos(aryFunctions[Idx], Expression);

         if (TempPos = 0) then
             Continue;

         if (fncPos = 0) then
             begin
               fncPos := TempPos;
               fncIdx := Idx;
             end
         else
             if (TempPos < fncPos) then
                 begin
                   fncPos := TempPos;
                   fncIdx := Idx;
                 end;
       end;

    if (fncPos = 0) then
        Continue;

    LeftParen  := 0;
    RightParen := 0;

    Cnt := 0;

    Idx := fncPos;

    repeat
      if (Expression[Idx] = '(') then
          begin
            Inc(Cnt);

            if (LeftParen = 0) then
                LeftParen := Idx;
          end;

      if (Expression[Idx] = ')') then
          begin
            Dec(Cnt);
            RightParen := Idx;
          end;

      Inc(Idx);
    until ((LeftParen > 0) and (Cnt = 0)) or
           (Idx > Length(Expression));

    SubExp := Copy(Expression, LeftParen + 1, RightParen - LeftParen - 1);

    Delete(Expression, fncPos, RightParen - fncPos + 1);
    DecodeExpression(SubExp);
    Insert(SubExp, Expression, fncPos);

    if (FSymbols.Count > 0) then
        TSymbolInfo(FSymbols[FSymbols.Count - 1]^).ApplyFunc := aryFunctions[fncIdx];
  until (fncPos = 0);
end;

function TExpression.DecodeSubExpression(Expression: string): string;

  procedure CreateSymbol(var Expression: string; Op: Word);
  var
    PriorOp, NextOp: Word;
    Operand1, Operand2: string;
    Operator: Char;
  begin
    {  Determine Prior and Next operator (for operand determination)  }

    PriorOp  := PriorOperator(Expression, Op);
    NextOp   := NextOperator(Expression, Op);

    {  Parse out each operand and the operator  }

    Operand1 := Copy(Expression, PriorOp + 1, Op - PriorOp - 1);
    Operator := Copy(Expression, Op, 1)[1];
    Operand2 := Copy(Expression, Op + 1, NextOp - Op - 1);

    {  Delete everything from Prior to Next operator and insert  }
    {  in its place, a new, system-generated symbol name that    }
    {  represents the sub-expression.                            }

    Delete(Expression, PriorOp + 1, NextOp - PriorOp - 1);
    Insert(GenerateNewSymbol(Operand1, Operand2, Operator), Expression, PriorOp + 1);
  end;

  procedure ParseOperators(var Expression: string; Op: TOperators);
  var
    OpPos: Word;
  begin
    {  Parse out all occurences of requested operators.  }
    {  Operator precedence is evaluated left-to-right.   }

    OpPos := 1;

    repeat
      if Expression[OpPos] in Op then
         begin
           CreateSymbol(Expression, OpPos);
           OpPos := 1;
         end
      else
         Inc(OpPos);
    until (OpPos > Length(Expression));
  end;

begin
  {  if Expression is simply a constant value (i.e. -5), then    }
  {  generate a new symbol to represent the expression using     }
  {  0 (zero) and Expression with the "+" operator.              }

  if IsConstantValue(Expression) then
     begin
       Result := GenerateNewSymbol('0', Expression, '+');
       exit;
     end
  else
     if (Pos('^', Expression) = 0) and
        (Pos('*', Expression) = 0) and
        (Pos('/', Expression) = 0) and
        (Pos('%', Expression) = 0) and
        (Pos('+', Expression) = 0) and
        (Pos('-', Expression) = 0) then
         begin
           Result := GenerateNewSymbol('0', Expression, '+');
           exit;
         end;

  {  Flank the sub-expression with pound (#) delimiters.         }
  {  This is required for operator sub-parsing.                  }

  Insert('#', Expression, 1);
  Insert('#', Expression, Length(Expression)+1);

  {  Parse sub-expressions using operator precedence             }

  ParseOperators(Expression, ['^']);            { Exponentiation }
  ParseOperators(Expression, ['*', '/', '%']);  { Mult, Div, Mod }
  ParseOperators(Expression, ['+', '-']);       { Add, Sub       }

  {  Strip the pound (#) delimiters back out                     }

  StripChar(Expression, '#');

  {  Pass back the parsed expression                             }

  Result := Expression;
end;

procedure TExpression.DecodeExpression(var Expression: string);
var
  SubExp:     string;
  LeftParen:  Integer;
  RightParen: Integer;
  i:          Word;
  WorkExt:    Extended;
begin
  Expression := UpperCase(Expression);

  if Pos('Expression Not Found', Expression) > 0 then
     raise TExpressionException.Create('Requested Stored Expression does not exist');

  if Pos('#', Expression) > 0 then
     raise TExpressionException.Create('Expression contains system-reserved "#" character');

  if Pos('SYS_GEN_SYM', Expression) > 0 then
     raise TExpressionException.Create('Expression contains system-reserved "Sys_Gen_Sym" characters');

  NormalizeExpression(Expression);

 {  Parse each sub-expression until there are no remaining  }
 {  parentheses.                                            }

  repeat
    LeftParen  := 0;
    RightParen := 0;

    i := 1;

    repeat
      if (Expression[i] = '(') then
          LeftParen  := i;

      if (Expression[i] = ')') then
          RightParen := i;

      Inc(i);
    until (RightParen > 0) or
          (i > Length(Expression));

    if ((LeftParen  = 0) and (RightParen <> 0)) or
       ((LeftParen <> 0) and (RightParen  = 0)) then
        raise TExpressionException.Create('Expression has misplaced parentheses!');

    if (LeftParen = 0) and (RightParen = 0) then
        Expression := DecodeSubExpression(Expression)
    else
        begin
          SubExp := Copy(Expression, LeftParen + 1, RightParen - LeftParen - 1);

          Delete(Expression, LeftParen, RightParen - LeftParen + 1);
          Insert(DecodeSubExpression(SubExp), Expression, LeftParen);
        end;

  until (Pos('(', Expression) = 0) and
        (Pos(')', Expression) = 0);

 {  If the resulting expression still contains valid        }
 {  operators, then process the final sub-expression.       }

  if (Pos('^', Expression) > 0) or
     (Pos('*', Expression) > 0) or
     (Pos('/', Expression) > 0) or
     (Pos('%', Expression) > 0) or
     (Pos('+', Expression) > 0) or
     (Pos('-', Expression) > 0) then
      Expression := DecodeSubExpression(Expression);
end;

end.
