{ ------------------------------------------------------------------------- }
{ Description:                                                              }
{    Retrieves fonts assigned to the system screen elements.                }
{                                                                           }
{ Declaration:                                                              }
{    Function GetSysFont(FontParam: TSysFont; var ResFont: TFont): Boolean; }
{                                                                           }
{ Parameters:                                                               }
{    FontParam - Enumerated type that specifies the system font to          }
{    retrieve. This is one of the following values: sfCaptionFont,          }
{    sfSmCaptionFont, sfMenuFont, sfStatusFont, sfMessageFont, sfIconFont.  }
{                                                                           }
{    ResFont - A TFont variable that can be used to store the resulting     }
{    TFont object.                                                          }
{ ------------------------------------------------------------------------- }
unit SysFont;

interface

Uses
	SysUtils, Windows, Graphics;

type
	 TSysFont = (sfCaptionFont, sfSmCaptionFont, sfMenuFont,
			sfStatusFont, sfMessageFont, sfIconFont);


Function GetSysFont(FontParam: TSysFont; var ResFont: TFont): Boolean;
procedure TFontFromLogFont(const LogFont: TLogFont; var NewFont: TFont);

implementation

{ Gets the System font setting of the passed TSysFont parameter }
Function GetSysFont(FontParam: TSysFont; var ResFont: TFont): Boolean;
var
	TempNonClientMetrics: TNonClientMetrics;
	TempLogFont: TLogFont;
	ret: boolean;
begin
	Result := false;
	If FontParam = sfIconFont then
		{ The IconFont is retreived with a different method
			than all the other system fonts }
		ret := SystemParametersInfo(SPI_GETICONTITLELOGFONT,
				SizeOf(TempLogFont), @TempLogFont, 0)
	else
		begin
			TempNonClientMetrics.cbSize := SizeOf(TempNonClientMetrics);
			ret := SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0,
					@TempNonClientMetrics, 0);
		end;

	{ If call to SystemParametersInfo() not successful then exit}
	If not ret then exit;

	{ Load appropriate TempNonClientMetrics field into TempLogFont }
	case FontParam of
		sfCaptionFont:
			TempLogFont := TempNonClientMetrics.lfCaptionFont;
		sfSmCaptionFont:
			TempLogFont := TempNonClientMetrics.lfSmCaptionFont;
		sfMenuFont:
			TempLogFont := TempNonClientMetrics.lfMenuFont;
		sfStatusFont:
			TempLogFont := TempNonClientMetrics.lfStatusFont;
		sfMessageFont:
			TempLogFont := TempNonClientMetrics.lfMessageFont;
	end;

	{ Convert TempLogFont into a TFont }
	TFontFromLogFont(TempLogFont, ResFont);
	Result := true;
end;

{ Converts a LogFont structure to a Delphi TFont }
procedure TFontFromLogFont(const LogFont: TLogFont; var NewFont: TFont);
var
	Style: TFontStyles;
begin
	with LogFont do
	begin
		NewFont.Name := lfFaceName;
		NewFont.Height := lfHeight;
		Style := [];

		if lfWeight > FW_REGULAR then Include(Style, fsBold);
		if lfItalic <> 0 then Include(Style, fsItalic);
		if lfUnderline <> 0 then Include(Style, fsUnderline);
		if lfStrikeOut <> 0 then Include(Style, fsStrikeOut);

		NewFont.Style := Style;
	end;
end;

end.
