{*******************************************************}
{                                                       }
{       Delphi Visual Component Library                 }
{                                                       }
{       Copyright (c) 1996 Borland International        }
{                                                       }
{*******************************************************}

(*{$DEFINE TestVersion}*) { <--- remove outer comments for test version }

{ Revisions
  ---------
  05-14-96 GW V1.11 Fixed problems in DeleteChildren, revealed by
              proper Expand/Collapse behavior.
              (a) possible infinite loop in the delete loop
              (b) recursive call to DeleteChildren via the
                  Expanded := false statement.
  05-14-96 GW V1.10
              Added Const GWProperExpandCollapseBehaviorFix
              Changed to separate GWProgramExpand and Collapse
              Changed to global Const GWEnableSpecialNodesUpdating
  05-12-96 GW V1.09
              Addressed the problem of OnExpand(ing) and OnCollaps(ing)
              routines not getting called for Node.Expand and Node.Collapse.
  05-08-96 GW Many more refinements for Adding and Deleting
  05-07-96 GW Added GTreeView_ calls and GTVWin unit for perf monitoring
  05-06-96 GW Speed-up for deletion (added TTree2Node.GWCountEquals function)
  04-01-96 GW Revised comments
  03-30-96 GW Several mods to cut out redundant sorting
}
{--------------------------------------------------------
To locate revisions by GW, search for "GW"
All added variables and methods have been named with
GW somewhere in the name for easy finding, EXCEPT as follows:
-- DEFINE TestVersion switches in the GTVWin unit for counting
   calls to W95 controls.
-- names of my interception routines are GTreeView_xxx

All changed code has been commented out (not deleted) for
easy reference.
---------------------------------------------------------

Summary of speediest settings and best strategies
-------------------------------------------------
Settings:
  GWProperExpandCollapseBehaviorFix: Boolean = true;  -- default
  GWEnableSpecialNodesUpdating: Boolean = true;            -- default

Cases:
-----
Add:
  Top Level:      TreeNodes.AddChild(nil,S)
  Non-top-level:  TreeNodes.AddChild(AParentNode,S)

Delete:
  Top Level:      TreeNodes.Clear
  Non-top-level:  TreeNode.DeleteChildren

What to do in each case:
------------------------
Add -- Top-Level:
  TreeView.BeginUpdate
    Loop --  TreeView.Items.AddChild(nil,S);
  TreeView.EndUpdate

Add -- Non-top-Level **
  AParentNode.GWBeginNodeUpdate
    Loop --  TreeView.Items.AddChild(AParentNode,S);
  AParentNode.GWEndNodeUpdate

Del -- Top-Level
  TreeView.BeginUpdate
    TreeView.Items.Clear;
  TreeView.EndUpdate

Del -- Non-top-level
  TreeView.BeginUpdate
    TreeNode.DeleteChildren;
  TreeView.EndUpdate

** Counter-intuitive note: AddChild for non-top-level node is SLOWED by using
   TreeView.Begin/EndUpdate

----------------------------------
TTreeView Issues in Detail
(This blurb last revised 5-14-96)
----------------------------------

Slowness destroying large number of nodes
-----------------------------------------
1) TTree2Node.Destroy  sped up by introducing:
     TTree2Node.GWCountEquals(N: integer): Boolean; (much faster
       than counting all the nodes).

2) TTree2Node.Destroy sped up for special case of Node.DeleteChildren,
    and Nodes.Clear using:
    TTree2Nodes.GWParentDeleting as a flag to bypass Destroy's
    unnecessary parent-lookup and update activity.

No special application coding is required to use this.
Note: speed is further improved for deleting by using TreeView.Begin/End Update.

Slowness adding large number of nodes
-------------------------------------
1) Added features to bypass unnecessary sorting and repainting during
   bulk AddChilds.

  Special application coding required:

    -- call Node.GWBeginNodeUpdate before loop calling
         Tree2View.Items.AddChild(Node,S);
    -- after loop call Node.GWEndNodeUpdate

  Or for top level:
    Enable via: T2V.Items.GWEnableSpecialUpdating := true, then...
     T2V.Items.BeginUpdate;
       Loop
         T2V.Items.AddChild(nil,S);
     T2V.Items.EndUpdate;

2) Added mechanism to defeat erroneous sorting of complete view when setting
   Node.Text and Node.Data before Node (Item) has been added to tree.
     InternalAddObject: (SetText and SetData)
     Uses flag: TTree2Node.Itemized: Boolean;

3) Sped up TTree2Nodes.AddedNode by passing ParentNode to it
   so that it doesn't have to look it up.

Usually no OnExpanding/ed and OnCollapsing/ed during Node.Expand/Collapse
--------------------------------------------------------------------------
It seems that this is a problem with W95's CNNotifying of this VCL code.
Basically, when your code calls Node.Expand, W95 DOES call CNNotify, but
only the first time you Expand a particular Node, and not subsequently.
When your code calls Node.Collapse, W95 never calls CNNotify so far as I
could tell.

I revised the ExpandItem code to take responsibility for this chore, employing
flags Node.GWProgramExpand and Collapse.
--------------------------------------------------------------}

unit Tree2vw;

{$R-}

interface

uses Messages, Windows, SysUtils, CommCtrl, Classes, Controls, Forms,
  Menus, Graphics, StdCtrls, RichEdit, DsgnIntf;

Const
  T2VVerNum = 111;
  GWProperExpandCollapseBehaviorFix: Boolean = true;  { GW 5-14-96 }
  GWEnableSpecialNodesUpdating: Boolean = true;            { GW 5-14-96 }

type

{ TTree2Node }

  TCustomTree2View = class;
  TTree2Nodes = class;

  TNodeState = (nsCut, nsDropHilited, nsFocused, nsSelected, nsExpanded);
  TNodeAttachMode = (naAdd, naAddFirst, naAddChild, naAddChildFirst, naInsert);
  TAddMode = (taAddFirst, taAdd, taInsert);

  PNodeInfo = ^TNodeInfo;
  TNodeInfo = packed record
    ImageIndex: Integer;
    SelectedIndex: Integer;
    StateIndex: Integer;
    OverlayIndex: Integer;
    Data: Pointer;
    Count: Integer;
    Text: string[255];
  end;

  TTree2Node = class(TPersistent)
  private
    FOwner: TTree2Nodes;
    FText: string;
    FData: Pointer;
    FItemId: HTreeItem;
    FImageIndex: Integer;
    FSelectedIndex: Integer;
    FOverlayIndex: Integer;
    FStateIndex: Integer;
    FDeleting: Boolean;

    GWItemized: Boolean;  { GW added 5-7-96 Flag for sort control in create process}
    { GWAdding: Boolean;  {  GW added 3-30-96 }
    GWNodeUpdating: Boolean;         { GW added 5-7-96 }
    GWDeletingChildren: Boolean; { GW added 5-14-96 }
    GWProgramExpand,
    GWProgramCollapse: smallint; { GW 5-14-96 signals expand/collapse via ExpandItem}

    procedure ExpandItem(Expand: Boolean; Recurse: Boolean);
    function GetAbsoluteIndex: Integer;
    function GetExpanded: Boolean;
    function GetLevel: Integer;
    function GetParent: TTree2Node;
    function GetChildren: Boolean;
    function GetCut: Boolean;
    function GetDropTarget: Boolean;
    function GetFocused: Boolean;
    function GetIndex: Integer;
    function GetItem(Index: Integer): TTree2Node;
    function GetSelected: Boolean;
    function GetState(NodeState: TNodeState): Boolean;
    function GetCount: Integer;
    function GetTree2View: TCustomTree2View;
    function HasVisibleParent: Boolean;
    procedure InternalMove(ParentNode, Node: TTree2Node; HItem: HTreeItem;
      AddMode: TAddMode);
    function IsEqual(Node: TTree2Node): Boolean;
    function IsNodeVisible: Boolean;
    procedure ReadData(Stream: TStream; Info: PNodeInfo);
    procedure SetChildren(Value: Boolean);
    procedure SetCut(Value: Boolean);
    procedure SetData(Value: Pointer);
    procedure SetDropTarget(Value: Boolean);
    procedure SetItem(Index: Integer; Value: TTree2Node);
    procedure SetExpanded(Value: Boolean);
    procedure SetFocused(Value: Boolean);
    procedure SetImageIndex(Value: Integer);
    procedure SetOverlayIndex(Value: Integer);
    procedure SetSelectedIndex(Value: Integer);
    procedure SetSelected(Value: Boolean);
    procedure SetStateIndex(Value: Integer);
    procedure SetText(const S: string);
    procedure WriteData(Stream: TStream; Info: PNodeInfo);
  public
    constructor Create(AOwner: TTree2Nodes);
    destructor Destroy; override;
    function AlphaSort: Boolean;
    procedure Assign(Source: TPersistent); override;
    procedure Collapse(Recurse: Boolean);
    function CustomSort(SortProc: TTVCompare; Data: Longint): Boolean;
    procedure Delete;
    procedure DeleteChildren;
    function DisplayRect(TextOnly: Boolean): TRect;
    function EditText: Boolean;
    procedure EndEdit(Cancel: Boolean);
    procedure Expand(Recurse: Boolean);
    function GetFirstChild: TTree2Node;
    function GetHandle: HWND;
    function GetLastChild: TTree2Node;
    function GetNext: TTree2Node;
    function GetNextChild(Value: TTree2Node): TTree2Node;
    function GetNextSibling: TTree2Node;
    function GetNextVisible: TTree2Node;
    function GetPrev: TTree2Node;
    function GetPrevChild(Value: TTree2Node): TTree2Node;
    function GetPrevSibling: TTree2Node;
    function GetPrevVisible: TTree2Node;
    function HasAsParent(Value: TTree2Node): Boolean;
    function IndexOf(Value: TTree2Node): Integer;
    procedure MakeVisible;
    procedure MoveTo(Destination: TTree2Node; Mode: TNodeAttachMode);
    property AbsoluteIndex: Integer read GetAbsoluteIndex;
    property Count: Integer read GetCount;

    function GWCountEquals(N: integer): Boolean; { *** GW Added 05-06-96 ***}
    procedure GWBeginNodeUpdate;                 { *** GW Added 05-07-96 ***}
    procedure GWEndNodeUpdate;                   { *** GW Added 05-07-96 ***}

    property Cut: Boolean read GetCut write SetCut;
    property Data: Pointer read FData write SetData;
    property Deleting: Boolean read FDeleting;
    property Focused: Boolean read GetFocused write SetFocused;
    property DropTarget: Boolean read GetDropTarget write SetDropTarget;
    property Selected: Boolean read GetSelected write SetSelected;
    property Expanded: Boolean read GetExpanded write SetExpanded;
    property Handle: HWND read GetHandle;
    property HasChildren: Boolean read GetChildren write SetChildren;
    property ImageIndex: Integer read FImageIndex write SetImageIndex;
    property Index: Integer read GetIndex;
    property IsVisible: Boolean read IsNodeVisible;
    property Item[Index: Integer]: TTree2Node read GetItem write SetItem; default;
    property ItemId: HTreeItem read FItemId;
    property Level: Integer read GetLevel;
    property OverlayIndex: Integer read FOverlayIndex write SetOverlayIndex;
    property Owner: TTree2Nodes read FOwner;
    property Parent: TTree2Node read GetParent;
    property SelectedIndex: Integer read FSelectedIndex write SetSelectedIndex;
    property StateIndex: Integer read FStateIndex write SetStateIndex;
    property Text: string read FText write SetText;
    property Tree2View: TCustomTree2View read GetTree2View;
  end;

{ TTree2Nodes }

  TTree2Nodes = class(TPersistent)
  private
    FOwner: TCustomTree2View;
    FUpdateCount: Integer;
    { GWUpdating: Boolean; }
    GWParentDeleting: integer;
    procedure AddedNode(NewNode: TTree2Node; GWParentNode: TTree2Node);
    function GetHandle: HWND;
    function GetNodeFromIndex(Index: Integer): TTree2Node;
    procedure ReadData(Stream: TStream);
    procedure Repaint(Node: TTree2Node);
    procedure WriteData(Stream: TStream);
    procedure GWEndUpdateInternal;    { GW added 5-8-96 }
  protected
    function AddItem(Parent, Target: HTreeItem; const Item: TTVItem;
      AddMode: TAddMode): HTreeItem;
    function InternalAddObject(Node: TTree2Node; const S: string;
      Ptr: Pointer; AddMode: TAddMode): TTree2Node;
    procedure DefineProperties(Filer: TFiler); override;
    function CreateItem(Node: TTree2Node): TTVItem;
    function GetCount: Integer;
    procedure SetItem(Index: Integer; Value: TTree2Node);
    procedure SetUpdateState(Updating: Boolean);
  public
    function GWItemsUpdating: Boolean;
    {GW 5-8-96 True iff GWEnableSpecialUpdating and FUpdateCount > 0 }
    constructor Create(AOwner: TCustomTree2View);
    destructor Destroy; override;
    function AddChildFirst(Node: TTree2Node; const S: string): TTree2Node;
    function AddChild(Node: TTree2Node; const S: string): TTree2Node;
    function AddChildObjectFirst(Node: TTree2Node; const S: string;
      Ptr: Pointer): TTree2Node;
    function AddChildObject(Node: TTree2Node; const S: string; Ptr: Pointer): TTree2Node;
    function AddFirst(Node: TTree2Node; const S: string): TTree2Node;
    function Add(Node: TTree2Node; const S: string): TTree2Node;
    function AddObjectFirst(Node: TTree2Node; const S: string;
      Ptr: Pointer): TTree2Node;
    function AddObject(Node: TTree2Node; const S: string;
      Ptr: Pointer): TTree2Node;

    procedure Assign(Source: TPersistent); override;
    procedure BeginUpdate;
    procedure Clear;
    procedure Delete(Node: TTree2Node);
    procedure EndUpdate;
    function GetFirstNode: TTree2Node;
    function GetNode(ItemId: HTreeItem): TTree2Node;
    function Insert(Node: TTree2Node; const S: string): TTree2Node;
    function InsertObject(Node: TTree2Node; const S: string;
      Ptr: Pointer): TTree2Node;
    property Count: Integer read GetCount;
    property Handle: HWND read GetHandle;
    property Item[Index: Integer]: TTree2Node read GetNodeFromIndex; default;
    property Owner: TCustomTree2View read FOwner;

  end;

{ TCustomTree2View }

  THitTest = (htAbove, htBelow, htNowhere, htOnItem, htOnButton,
    htOnIcon, htOnIndent, htOnLabel, htOnRight,
    htOnStateIcon, htToLeft, htToRight);
  THitTests = set of THitTest;
  ETree2ViewError = class(Exception);

  TTVChangingEvent = procedure(Sender: TObject; Node: TTree2Node;
    var AllowChange: Boolean) of object;
  TTVChangedEvent = procedure(Sender: TObject; Node: TTree2Node) of object;
  TTVEditingEvent = procedure(Sender: TObject; Node: TTree2Node;
    var AllowEdit: Boolean) of object;
  TTVEditedEvent = procedure(Sender: TObject; Node: TTree2Node; var S: string) of object;
  TTVExpandingEvent = procedure(Sender: TObject; Node: TTree2Node;
    var AllowExpansion: Boolean) of object;
  TTVCollapsingEvent = procedure(Sender: TObject; Node: TTree2Node;
    var AllowCollapse: Boolean) of object;
  TTVExpandedEvent = procedure(Sender: TObject; Node: TTree2Node) of object;
  TTVCompareEvent = procedure(Sender: TObject; Node1, Node2: TTree2Node;
    Data: Integer; var Compare: Integer) of object;

  TSortType = (stNone, stData, stText, stBoth);

  TCustomTree2View = class(TWinControl)
  private
    FShowLines: Boolean;
    FShowRoot: Boolean;
    FShowButtons: Boolean;
    FBorderStyle: TBorderStyle;
    FReadOnly: Boolean;
    FImages: TImageList;
    FStateImages: TImageList;
    FImageChangeLink: TChangeLink;
    FStateChangeLink: TChangeLink;
    FDragImage: TImageList;
    FTree2Nodes: TTree2Nodes;
    FSortType: TSortType;
    FSaveItems: TStringList;
    FSaveTopIndex: Integer;
    FSaveIndex: Integer;
    FSaveIndent: Integer;
    FHideSelection: Boolean;
    FMemStream: TMemoryStream;
    FEditInstance: Pointer;
    FDefEditProc: Pointer;
    FEditHandle: HWND;
    FDragged: Boolean;
    FRClicked: Boolean;
    FLastDropTarget: TTree2Node;
    FDragNode: TTree2Node;
    FOnEditing: TTVEditingEvent;
    FOnEdited: TTVEditedEvent;
    FOnExpanded: TTVExpandedEvent;
    FOnExpanding: TTVExpandingEvent;
    FOnCollapsed: TTVExpandedEvent;
    FOnCollapsing: TTVCollapsingEvent;
    FOnChanging: TTVChangingEvent;
    FOnChange: TTVChangedEvent;
    FOnCompare: TTVCompareEvent;
    FOnDeletion: TTVExpandedEvent;
    FOnGetImageIndex: TTVExpandedEvent;
    FOnGetSelectedIndex: TTVExpandedEvent;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMCtl3DChanged(var Message: TMessage); message CM_CTL3DCHANGED;
    procedure CMDrag(var Message: TCMDrag); message CM_DRAG;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure EditWndProc(var Message: TMessage);
    procedure DoDragOver(Source: TDragObject; X, Y: Integer);
    procedure GetImageIndex(Node: TTree2Node);
    procedure GetSelectedIndex(Node: TTree2Node);
    function GetDropTarget: TTree2Node;
    function GetIndent: Integer;
    function GetNodeFromItem(const Item: TTVItem): TTree2Node;
    function GetSelection: TTree2Node;
    function GetTopItem: TTree2Node;
    procedure ImageListChange(Sender: TObject);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetButtonStyle(Value: Boolean);
    procedure SetDropTarget(Value: TTree2Node);
    procedure SetHideSelection(Value: Boolean);
    procedure SetImageList(Value: HImageList; Flags: Integer);
    procedure SetIndent(Value: Integer);
    procedure SetImages(Value: TImageList);
    procedure SetLineStyle(Value: Boolean);
    procedure SetReadOnly(Value: Boolean);
    procedure SetRootStyle(Value: Boolean);
    procedure SetSelection(Value: TTree2Node);
    procedure SetSortType(Value: TSortType);
    procedure SetStateImages(Value: TImageList);
    procedure SetStyle(Value: Integer; UseStyle: Boolean);
    procedure SetTree2Nodes(Value: TTree2Nodes);
    procedure SetTopItem(Value: TTree2Node);
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMRButtonDown(var Message: TWMRButtonDown); message WM_RBUTTONDOWN;
  protected
    function CanEdit(Node: TTree2Node): Boolean; dynamic;
    function CanChange(Node: TTree2Node): Boolean; dynamic;
    function CanCollapse(Node: TTree2Node): Boolean; dynamic;
    function CanExpand(Node: TTree2Node): Boolean; dynamic;
    procedure Change(Node: TTree2Node); dynamic;
    procedure Collapse(Node: TTree2Node); dynamic;
    function CreateNode: TTree2Node; virtual;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    procedure DoEndDrag(Target: TObject; X, Y: Integer); dynamic;
    procedure DoStartDrag(var DragObject: TDragObject); override;
    procedure Edit(const Item: TTVItem); dynamic;
    procedure Expand(Node: TTree2Node); dynamic;
    function GetDragImages: TCustomImageList; override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure SetDragMode(Value: TDragMode); override;
    procedure WndProc(var Message: TMessage); override;
    property OnEditing: TTVEditingEvent read FOnEditing write FOnEditing;
    property OnEdited: TTVEditedEvent read FOnEdited write FOnEdited;
    property OnExpanding: TTVExpandingEvent read FOnExpanding write FOnExpanding;
    property OnExpanded: TTVExpandedEvent read FOnExpanded write FOnExpanded;
    property OnCollapsing: TTVCollapsingEvent read FOnCollapsing write FOnCollapsing;
    property OnCollapsed: TTVExpandedEvent read FOnCollapsed write FOnCollapsed;
    property OnChanging: TTVChangingEvent read FOnChanging write FOnChanging;
    property OnChange: TTVChangedEvent read FOnChange write FOnChange;
    property OnCompare: TTVCompareEvent read FOnCompare write FOnCompare;
    property OnDeletion: TTVExpandedEvent read FOnDeletion write FOnDeletion;
    property OnGetImageIndex: TTVExpandedEvent read FOnGetImageIndex write FOnGetImageIndex;
    property OnGetSelectedIndex: TTVExpandedEvent read FOnGetSelectedIndex write FOnGetSelectedIndex;
    property ShowButtons: Boolean read FShowButtons write SetButtonStyle default True;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property ShowLines: Boolean read FShowLines write SetLineStyle default True;
    property ShowRoot: Boolean read FShowRoot write SetRootStyle default True;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly default False;
    property Indent: Integer read GetIndent write SetIndent;
    property Items: TTree2Nodes read FTree2Nodes write SetTree2Nodes;
    property SortType: TSortType read FSortType write SetSortType default stNone;
    property HideSelection: Boolean read FHideSelection write SetHideSelection default True;
    property Images: TImageList read FImages write SetImages;
    property StateImages: TImageList read FStateImages write SetStateImages;
  public
    GWDebugMemo: TMemo;          { GW 3-30-96 }
    function  GWVerNum: integer; { GW 5-14-96 }
    
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AlphaSort: Boolean;
    function CustomSort(SortProc: TTVCompare; Data: Longint): Boolean;
    procedure FullCollapse;
    procedure FullExpand;
    function GetHitTestInfoAt(X, Y: Integer): THitTests;
    function GetNodeAt(X, Y: Integer): TTree2Node;
    function IsEditing: Boolean;
    procedure LoadFromFile(const FileName: string);
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToFile(const FileName: string);
    procedure SaveToStream(Stream: TStream);
    property DropTarget: TTree2Node read GetDropTarget write SetDropTarget;
    property Selected: TTree2Node read GetSelection write SetSelection;
    property TopItem: TTree2Node read GetTopItem write SetTopItem;
  end;

  TTree2View = class(TCustomTree2View)
  published
    property ShowButtons;
    property BorderStyle;
    property DragCursor;
    property ShowLines;
    property ShowRoot;
    property ReadOnly;
    property DragMode;
    property HideSelection;
    property Indent;
    property Items;
    property OnEditing;
    property OnEdited;
    property OnExpanding;
    property OnExpanded;
    property OnCollapsing;
    property OnCompare;
    property OnCollapsed;
    property OnChanging;
    property OnChange;
    property OnDeletion;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property Align;
    property Enabled;
    property Font;
    property Color;
    property ParentColor;
    property ParentCtl3D;
    property Ctl3D;
    property SortType;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnClick;
    property OnEnter;
    property OnExit;
    property OnDragDrop;
    property OnDragOver;
    property OnStartDrag;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnDblClick;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property PopupMenu;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property Images;
    property StateImages;
  end;

procedure Register;

implementation

uses Printers, Consts, ComStrs, NodeEdit, comctrls, LibConst
{$IFDEF TestVersion}
, GWWinTV
{$ENDIF}
;

{ TTree2Node }
function DefaultTree2ViewSort(Node1, Node2: TTree2Node; lParam: Integer): Integer; stdcall;
begin
  with Node1 do
    if Assigned(Tree2View.OnCompare) then
      Tree2View.OnCompare(Tree2View, Node1, Node2, lParam, Result)
    else Result := lstrcmp(PChar(Node1.Text), PChar(Node2.Text));
end;

procedure Tree2ViewError(MsgID: Integer);
begin
  raise ETree2ViewError.CreateRes(MsgID);
end;

constructor TTree2Node.Create(AOwner: TTree2Nodes);
begin
  inherited Create;
  FOverlayIndex := -1;
  FStateIndex := -1;
  FOwner := AOwner;

  FDeleting               := false;  { GW added 5-14-96 }
  GWItemized              := false;  { GW added 5-7-96 }
  GWNodeUpdating          := false;  { GW added 5-7-96 }
  GWDeletingChildren      := false;
  GWProgramExpand         := 0; { GW added 5-14-96 }
  GWProgramCollapse       := 0; { GW added 5-14-96 }
end;

destructor TTree2Node.Destroy;
var
  Node: TTree2Node;
  CheckValue: Integer;
begin
  FDeleting := True;

  { GW 5-7-96 added to speed up child deletes }
  If (FOwner <> nil) and (FOwner.GWParentDeleting = 0) then
  Begin { no need to do this if parent is unconcerned }
    Node := Parent;
    if (Node <> nil) and (not Node.Deleting) then
    begin
      if Node.IndexOf(Self) <> -1 then CheckValue := 1
      else CheckValue := 0;
      { if Node.Count = CheckValue then }
      if Node.GWCountEquals(CheckValue) then { GW 05-06-96 }
      begin
        Node.Expanded := False;
        Node.HasChildren := False;
      end;
    end;
  end;

{$IFDEF TestVersion}
  if ItemId <> nil then GTreeView_DeleteItem(Handle, ItemId);
{$ELSE}
  if ItemId <> nil then TreeView_DeleteItem(Handle, ItemId);
{$ENDIF}
  Data := nil;
  inherited Destroy;
end;

function TTree2Node.GetHandle: HWND;
begin
  Result := Tree2View.Handle;
end;

function TTree2Node.GetTree2View: TCustomTree2View;
begin
  Result := Owner.Owner;
end;

function TTree2Node.HasAsParent(Value: TTree2Node): Boolean;
begin
  if Self = Value then Result := True
  else if Parent <> nil then Result := Parent.HasAsParent(Value)
  else Result := False;
end;

procedure TTree2Node.SetText(const S: string);
var
  Item: TTVItem;
begin
  FText := S;
  with Item do
  begin
    mask := TVIF_TEXT;
    hItem := ItemId;
    pszText := LPSTR_TEXTCALLBACK;
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Item);
{$ELSE}
  TreeView_SetItem(Handle, Item);
{$ENDIF}
  if Tree2View.SortType in [stText, stBoth] then
  begin
    If GWItemized then  { GW 5-7-96... indicates this node actually in tree }
    Begin
      if Parent <> nil then Parent.AlphaSort
      else Tree2View.AlphaSort;  { no parent means top-level node }
    end;
  end;
end;

procedure TTree2Node.SetData(Value: Pointer);
begin
  FData := Value;
  { GW 5-7-96... No point doing the following if:
    a) Node not in tree (ie: never been "itemized")
    b) Node is in process of deletion
  }
  If GWItemized and not Deleting then
  Begin
    if (Tree2View.SortType in [stData, stBoth]) and Assigned(Tree2View.OnCompare) then
    begin
      Begin
        if Parent <> nil then Parent.AlphaSort
        else Tree2View.AlphaSort; { no parent means top-level node }
      end;
    end;
  end;
end;

function TTree2Node.GetState(NodeState: TNodeState): Boolean;
var
  Item: TTVItem;
begin
  Result := False;
  with Item do
  begin
    mask := TVIF_STATE;
    hItem := ItemId;
{$IFDEF TestVersion}
    if GTreeView_GetItem(Handle, Item) then
  {$ELSE}
    if TreeView_GetItem(Handle, Item) then
  {$ENDIF}  
      case NodeState of
        nsCut: Result := (state and TVIS_CUT) <> 0;
        nsFocused: Result := (state and TVIS_FOCUSED) <> 0;
        nsSelected: Result := (state and TVIS_SELECTED) <> 0;
        nsExpanded: Result := (state and TVIS_EXPANDED) <> 0;
        nsDropHilited: Result := (state and TVIS_DROPHILITED) <> 0;
      end;
  end;
end;

procedure TTree2Node.SetImageIndex(Value: Integer);
var
  Item: TTVItem;
begin
  FImageIndex := Value;
  with Item do
  begin
    mask := TVIF_IMAGE;
    hItem := ItemId;
    iImage := I_IMAGECALLBACK;
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Item);
{$ELSE}
  TreeView_SetItem(Handle, Item);
{$ENDIF}  
end;

procedure TTree2Node.SetSelectedIndex(Value: Integer);
var
  Item: TTVItem;
begin
  FSelectedIndex := Value;
  with Item do
  begin
    mask := TVIF_SELECTEDIMAGE;
    hItem := ItemId;
    iSelectedImage := I_IMAGECALLBACK;
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Item);
{$ELSE}
  TreeView_SetItem(Handle, Item);
{$ENDIF}
end;

procedure TTree2Node.SetOverlayIndex(Value: Integer);
var
  Item: TTVItem;
begin
  FOverlayIndex := Value;
  with Item do
  begin
    mask := TVIF_STATE;
    stateMask := TVIS_OVERLAYMASK;
    hItem := ItemId;
    state := IndexToOverlayMask(OverlayIndex + 1);
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Item);
{$ELSE}
  TreeView_SetItem(Handle, Item);
{$ENDIF}  
end;

procedure TTree2Node.SetStateIndex(Value: Integer);
var
  Item: TTVItem;
begin
  FStateIndex := Value;
  if Value >= 0 then Dec(Value);
  with Item do
  begin
    mask := TVIF_STATE;
    stateMask := TVIS_STATEIMAGEMASK;
    hItem := ItemId;
    state := IndexToStateImageMask(Value + 1);
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Item);
{$ELSE}
  TreeView_SetItem(Handle, Item);
{$ENDIF}
end;

procedure TTree2Node.ExpandItem(Expand: Boolean; Recurse: Boolean);
var
  Flag: Integer;
  Node: TTree2Node;
  TV: TCustomTree2View;
Label skipit;
begin
  if Recurse then
  {=========== Recursion actions ================================}
  begin
    Node := Self;
    repeat
      Node.ExpandItem(Expand, False);
      Node := Node.GetNext;
    until (Node = nil) or not Node.HasAsParent(Self);
  end
  {=========== The actual expansion/collapse ==============================
    GW 5-12-96
    Modified to carry out the CanExpand/Collapse and Expanded/Collapsed
    logic, since the W95 control apparently usually doesn't notify us via
    CNNotify when called through TreeView_Expand.
    But... because it SOMETIMES does, we need the
    GWProgramExpand/Collapse to PREVENT the "normal" notification from
    redundantly calling CanExpand/Collapse and Expanded/Collapsed
  ========================================================================}

  else begin
    If GWProperExpandCollapseBehaviorFix then
    Begin
      {-------------------------------------------
        Do the fixed behavior
      -------------------------------------------}
      TV := Tree2View;
      If TV = nil then goto skipit;

      if Expand then { try to expand }
      Begin
        Inc(GWProgramExpand);
        try
          if not Expanded then
          Begin
            if TV.CanExpand(self) then
            Begin
              Flag := TVE_EXPAND;
              {$IFDEF TestVersion}
              GTreeView_Expand(Handle, ItemId, Flag);
              {$ELSE}
              TreeView_Expand(Handle, ItemId, Flag);
              {$ENDIF}
              TV.Expand(self);
            end; { if }
          end; { if }
        finally
          Dec(GWProgramExpand);
        end; { try }

      end else
      Begin     { try to collapse }
        Inc(GWProgramCollapse);
        try
          If Expanded then
          Begin  { Expanded, see if we should collapse }
            if TV.CanCollapse(self) then
            Begin
              Flag := TVE_COLLAPSE;
              {$IFDEF TestVersion}
              GTreeView_Expand(Handle, ItemId, Flag);
              {$ELSE}
              TreeView_Expand(Handle, ItemId, Flag);
              {$ENDIF}
              TV.Collapse(self);
            end; { if }
          end; { if }
        finally
          Dec(GWProgramCollapse);
        end; { try }
      end; { if }
    end else
    Begin
      {------------------------------------------
        Do the unmodified behavior. This code
        included in case someone finds that
        the modified behavior has a side-effect.
        GW 5-14-96
      ------------------------------------------}
      GWProgramExpand   := 0; { redundant but just in case }
      GWProgramCollapse := 0; { redundant but just in case }
      { original code follows...}
      if Expand then Flag := TVE_EXPAND
        else Flag := TVE_COLLAPSE;
      {$IFDEF TestVersion}
      GTreeView_Expand(Handle, ItemId, Flag);
      {$ELSE}
      TreeView_Expand(Handle, ItemId, Flag);
      {$ENDIF}
    end;
  end;
  {=============================================================}
skipit:
end;

procedure TTree2Node.Expand(Recurse: Boolean);
begin
  ExpandItem(True, Recurse);
end;

procedure TTree2Node.Collapse(Recurse: Boolean);
begin
  ExpandItem(False, Recurse);
end;

function TTree2Node.GetExpanded: Boolean;
begin
  Result := GetState(nsExpanded);
end;

procedure TTree2Node.SetExpanded(Value: Boolean);
begin
  if Value then Expand(False)
  else Collapse(False);
end;

function TTree2Node.GetSelected: Boolean;
begin
  Result := GetState(nsSelected);
end;

procedure TTree2Node.SetSelected(Value: Boolean);
begin
{$IFDEF TestVersion}
  if Value then GTreeView_SelectItem(Handle, ItemId)
  else if Selected then GTreeView_SelectItem(Handle, nil);
{$ELSE}
  if Value then TreeView_SelectItem(Handle, ItemId)
  else if Selected then TreeView_SelectItem(Handle, nil);
{$ENDIF}
end;

function TTree2Node.GetCut: Boolean;
begin
  Result := GetState(nsCut);
end;

procedure TTree2Node.SetCut(Value: Boolean);
var
  Item: TTVItem;
  Template: Integer;
begin
  if Value then Template := -1
  else Template := 0;
  with Item do
  begin
    mask := TVIF_STATE;
    hItem := ItemId;
    stateMask := TVIS_CUT;
    state := stateMask and Template;
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Item);
{$ELSE}
  TreeView_SetItem(Handle, Item);
{$ENDIF}  
end;

function TTree2Node.GetDropTarget: Boolean;
begin
  Result := GetState(nsDropHilited);
end;

procedure TTree2Node.SetDropTarget(Value: Boolean);
begin
{$IFDEF TestVersion}
  if Value then GTreeView_SelectDropTarget(Handle, ItemId)
  else if DropTarget then GTreeView_SelectDropTarget(Handle, nil);
{$ELSE}
  if Value then TreeView_SelectDropTarget(Handle, ItemId)
  else if DropTarget then TreeView_SelectDropTarget(Handle, nil);
{$ENDIF}
end;

function TTree2Node.GetChildren: Boolean;
var
  Item: TTVItem;
begin
  Item.mask := TVIF_CHILDREN;
  Item.hItem := ItemId;
{$IFDEF TestVersion}
  if GTreeView_GetItem(Handle, Item) then Result := Item.cChildren > 0
{$ELSE}
  if TreeView_GetItem(Handle, Item) then Result := Item.cChildren > 0
{$ENDIF}
  else Result := False;
end;

procedure TTree2Node.SetFocused(Value: Boolean);
var
  Item: TTVItem;
  Template: Integer;
begin
  if Value then Template := -1
  else Template := 0;
  with Item do
  begin
    mask := TVIF_STATE;
    hItem := ItemId;
    stateMask := TVIS_FOCUSED;
    state := stateMask and Template;
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Item);
{$ELSE}
  TreeView_SetItem(Handle, Item);
{$ENDIF}
end;

function TTree2Node.GetFocused: Boolean;
begin
  Result := GetState(nsFocused);
end;

procedure TTree2Node.SetChildren(Value: Boolean);
var
  Item: TTVItem;
begin
  with Item do
  begin
    mask := TVIF_CHILDREN;
    hItem := ItemId;
    cChildren := Ord(Value);
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Item);
{$ELSE}
  TreeView_SetItem(Handle, Item);
{$ENDIF}
end;

function TTree2Node.GetParent: TTree2Node;
begin
  with FOwner do
{$IFDEF TestVersion}
    Result := GetNode(GTreeView_GetParent(Handle, ItemId));
  {$ELSE}
    Result := GetNode(TreeView_GetParent(Handle, ItemId));
  {$ENDIF}
end;

function TTree2Node.GetNextSibling: TTree2Node;
begin
  (***
  If Tree2View.GWDebugMemo <> nil then
    Tree2View.GWDebugMemo.Lines.Append('GNS ');
  ***)
  with FOwner do
{$IFDEF TestVersion}
    Result := GetNode(GTreeView_GetNextSibling(Handle, ItemId));
  {$ELSE}
    Result := GetNode(TreeView_GetNextSibling(Handle, ItemId));
  {$ENDIF}
end;

function TTree2Node.GetPrevSibling: TTree2Node;
begin
  with FOwner do
{$IFDEF TestVersion}
    Result := GetNode(GTreeView_GetPrevSibling(Handle, ItemId));
  {$ELSE}
    Result := GetNode(TreeView_GetPrevSibling(Handle, ItemId));
  {$ENDIF}
end;

function TTree2Node.GetNextVisible: TTree2Node;
begin
  if IsVisible then
    with FOwner do
{$IFDEF TestVersion}
      Result := GetNode(GTreeView_GetNextVisible(Handle, ItemId))
    {$ELSE}
      Result := GetNode(TreeView_GetNextVisible(Handle, ItemId))
    {$ENDIF}
  else Result := nil;
end;

function TTree2Node.GetPrevVisible: TTree2Node;
begin
  with FOwner do
{$IFDEF TestVersion}
    Result := GetNode(GTreeView_GetPrevVisible(Handle, ItemId));
  {$ELSE}
    Result := GetNode(TreeView_GetPrevVisible(Handle, ItemId));
  {$ENDIF}
end;

function TTree2Node.GetNextChild(Value: TTree2Node): TTree2Node;
begin
  if Value <> nil then Result := Value.GetNextSibling
  else Result := nil;
end;

function TTree2Node.GetPrevChild(Value: TTree2Node): TTree2Node;
begin
  if Value <> nil then Result := Value.GetPrevSibling
  else Result := nil;
end;

function TTree2Node.GetFirstChild: TTree2Node;
begin
  with FOwner do
{$IFDEF TestVersion}
    Result := GetNode(GTreeView_GetChild(Handle, ItemId));
  {$ELSE}
    Result := GetNode(TreeView_GetChild(Handle, ItemId));
  {$ENDIF}  
end;

function TTree2Node.GetLastChild: TTree2Node;
var
  Node: TTree2Node;
begin
  Result := GetFirstChild;
  if Result <> nil then
  begin
    Node := Result;
    repeat
      Result := Node;
      Node := Result.GetNextSibling;
    until Node = nil;
  end;
end;

function TTree2Node.GetNext: TTree2Node;
var
  NodeID, ParentID: HTreeItem;
  Handle: HWND;
begin
  (***
  If Tree2View.GWDebugMemo <> nil then
    Tree2View.GWDebugMemo.Lines.Append('GN');
  ***)  
  Handle := FOwner.Handle;
{$IFDEF TestVersion}
  NodeID := GTreeView_GetChild(Handle, ItemId);
  if NodeID = nil then NodeID := GTreeView_GetNextSibling(Handle, ItemId);
{$ELSE}
  NodeID := TreeView_GetChild(Handle, ItemId);
  if NodeID = nil then NodeID := TreeView_GetNextSibling(Handle, ItemId);
{$ENDIF}

  ParentID := ItemId;
  while (NodeID = nil) and (ParentID <> nil) do
  begin
{$IFDEF TestVersion}
    ParentID := GTreeView_GetParent(Handle, ParentID);
    NodeID := GTreeView_GetNextSibling(Handle, ParentID);
  {$ELSE}
    ParentID := TreeView_GetParent(Handle, ParentID);
    NodeID := TreeView_GetNextSibling(Handle, ParentID);
  {$ENDIF}
  end;
  Result := FOwner.GetNode(NodeID);
end;

function TTree2Node.GetPrev: TTree2Node;
var
  Node: TTree2Node;
begin
  Result := GetPrevSibling;
  if Result <> nil then
  begin
    Node := Result;
    repeat
      Result := Node;
      Node := Result.GetLastChild;
    until Node = nil;
  end else
    Result := Parent;
end;

function TTree2Node.GetAbsoluteIndex: Integer;
var
  Node: TTree2Node;
begin
  Result := -1;
  Node := Self;
  while Node <> nil do
  begin
    Inc(Result);
    Node := Node.GetPrev;
  end;
end;

function TTree2Node.GetIndex: Integer;
var
  Node: TTree2Node;
begin
  Result := -1;
  Node := Self;
  while Node <> nil do
  begin
    Inc(Result);
    Node := Node.GetPrevSibling;
  end;
end;

function TTree2Node.GetItem(Index: Integer): TTree2Node;
begin
  Result := GetFirstChild;
  while (Result <> nil) and (Index > 0) do
  begin
    Result := GetNextChild(Result);
    Dec(Index);
  end;
  if Result = nil then Tree2ViewError(SListIndexError);
end;

procedure TTree2Node.SetItem(Index: Integer; Value: TTree2Node);
begin
  item[Index].Assign(Value);
end;

function TTree2Node.IndexOf(Value: TTree2Node): Integer;
var
  Node: TTree2Node;
begin
  Result := -1;
  Node := GetFirstChild;
  while (Node <> nil) do
  begin
    Inc(Result);
    if Node = Value then Break;
    Node := GetNextChild(Node);
  end;
  if Node = nil then Result := -1;
end;

function TTree2Node.GetCount: Integer;
var
  Node: TTree2Node;
begin
  Result := 0;
  Node := GetFirstChild;
  while Node <> nil do
  begin
    Inc(Result);
    Node := Node.GetNextChild(Node);
  end;
end;

{---------------------------------------------------}
function TTree2Node.GWCountEquals(N: integer): Boolean;
{ Added GW 05-06-96 to speed up large Deletes       }
{---------------------------------------------------}
var
  Node: TTree2Node;
  Cnt: integer;
Label done;
Begin
  Cnt := 0;
  Result := false;
  Node := GetFirstChild;
  while Node <> nil do
  begin
    Inc(Cnt);
    Node := Node.GetNextChild(Node);
    If Cnt > N then goto done;
  end;
  If Cnt = N then Result := true;
done:
end;

{----------------------------------}
procedure TTree2Node.GWBeginNodeUpdate;  { GW added 5-7-96 }
{----------------------------------}
Begin
  GWNodeUpdating := true;
end;

{---------------------------------}
procedure TTree2Node.GWEndNodeUpdate;   { GW added 5-7-96 }
{---------------------------------}
Var
  MyTreeNodes: TTree2Nodes;
Begin
  GWNodeUpdating := false;
  if fOwner <> nil then
  Begin
    MyTreeNodes := fOwner;
    MyTreeNodes.Repaint(self);
    { do the re-sort }
    if MyTreeNodes.owner.SortType in [stText, stBoth] then AlphaSort;
  end;
end;

procedure TTree2Node.EndEdit(Cancel: Boolean);
begin
{$IFDEF TestVersion}
  GTreeView_EndEditLabelNow(Handle, Cancel);
{$ELSE}
  TreeView_EndEditLabelNow(Handle, Cancel);
{$ENDIF}
end;

procedure TTree2Node.InternalMove(ParentNode, Node: TTree2Node;
  HItem: HTreeItem; AddMode: TAddMode);
var
  I: Integer;
  NodeId: HTreeItem;
  Tree2ViewItem: TTVItem;
  Children: Boolean;
  IsSelected: Boolean;
begin
  if (AddMode = taInsert) and (Node <> nil) then
    NodeId := Node.ItemId else
    NodeId := nil;
  Children := HasChildren;
  IsSelected := Selected;
  if (Parent <> nil) and (Parent.Count = 1) then
  begin
    Parent.Expanded := False;
    Parent.HasChildren := False;
  end;
  with Tree2ViewItem do
  begin
    mask := TVIF_PARAM;
    hItem := ItemId;
    lParam := 0;
  end;
{$IFDEF TestVersion}
  GTreeView_SetItem(Handle, Tree2ViewItem);
{$ELSE}
  TreeView_SetItem(Handle, Tree2ViewItem);
{$ENDIF}
  with Owner do
    HItem := AddItem(HItem, NodeId, CreateItem(Self), AddMode);
  if HItem = nil then
    raise EOutOfResources.CreateRes(sInsertError);
  for I := Count - 1 downto 0 do
    Item[I].InternalMove(Self, nil, HItem, taAddFirst);
{$IFDEF TestVersion}
  GTreeView_DeleteItem(Handle, ItemId);
{$ELSE}
  TreeView_DeleteItem(Handle, ItemId);
{$ENDIF}
  FItemId := HItem;
  Assign(Self);
  HasChildren := Children;
  Selected := IsSelected;
end;

procedure TTree2Node.MoveTo(Destination: TTree2Node; Mode: TNodeAttachMode);
var
  AddMode: TAddMode;
  Node: TTree2Node;
  HItem: HTreeItem;
  OldOnChanging: TTVChangingEvent;
  OldOnChange: TTVChangedEvent;
begin
  OldOnChanging := Tree2View.OnChanging;
  OldOnChange := Tree2View.OnChange;
  Tree2View.OnChanging := nil;
  Tree2View.OnChange := nil;
  try
    if (Destination = nil) or not Destination.HasAsParent(Self) then
    begin
      AddMode := taAdd;
      if (Destination <> nil) and not (Mode in [naAddChild, naAddChildFirst]) then
        Node := Destination.Parent else
        Node := Destination;
      case Mode of
        naAdd,
        naAddChild: AddMode := taAdd;
        naAddFirst,
        naAddChildFirst: AddMode := taAddFirst;
        naInsert:
          begin
            Destination := Destination.GetPrevSibling;
            if Destination = nil then AddMode := taAddFirst
            else AddMode := taInsert;
          end;
      end;
      if Node <> nil then
        HItem := Node.ItemId else
        HItem := nil;
      InternalMove(Node, Destination, HItem, AddMode);
    end;
  finally
    Tree2View.OnChanging := OldOnChanging;
    Tree2View.OnChange := OldOnChange;
  end;
end;

procedure TTree2Node.MakeVisible;
begin
{$IFDEF TestVersion}
  GTreeView_EnsureVisible(Handle, ItemId);
{$ELSE}
  TreeView_EnsureVisible(Handle, ItemId);
{$ENDIF}  
end;

function TTree2Node.GetLevel: Integer;
var
  Node: TTree2Node;
begin
  Result := 0;
  Node := Parent;
  while Node <> nil do
  begin
    Inc(Result);
    Node := Node.Parent;
  end;
end;

function TTree2Node.IsNodeVisible: Boolean;
var
  Rect: TRect;
begin
{$IFDEF TestVersion}
  Result := GTreeView_GetItemRect(Handle, ItemId, Rect, True);
{$ELSE}
  Result := TreeView_GetItemRect(Handle, ItemId, Rect, True);
{$ENDIF}  
end;

function TTree2Node.HasVisibleParent: Boolean;
begin
  Result := (Parent <> nil) and (Parent.Expanded);
end;

function TTree2Node.EditText: Boolean;
begin
{$IFDEF TestVersion}
  Result := GTreeView_EditLabel(Handle, ItemId) <> 0;
{$ELSE}
  Result := TreeView_EditLabel(Handle, ItemId) <> 0;
{$ENDIF}  
end;

function TTree2Node.DisplayRect(TextOnly: Boolean): TRect;
begin
  FillChar(Result, SizeOf(Result), 0);
{$IFDEF TestVersion}
  GTreeView_GetItemRect(Handle, ItemId, Result, TextOnly);
{$ELSE}
  TreeView_GetItemRect(Handle, ItemId, Result, TextOnly);
{$ENDIF}  
end;

function TTree2Node.AlphaSort: Boolean;
begin
  Result := CustomSort(nil, 0);
end;

function TTree2Node.CustomSort(SortProc: TTVCompare; Data: Longint): Boolean;
var
  SortCB: TTVSortCB;
begin
  (***
  If Tree2View.GWDebugMemo <> nil then
    Tree2View.GWDebugMemo.Lines.Append('<'+FText+'.CustomSort>');
  ***)  
  with SortCB do
  begin
    if not Assigned(SortProc) then lpfnCompare := @DefaultTree2ViewSort
    else lpfnCompare := SortProc;
    hParent := ItemId;
    lParam := Data;
  end;
{$IFDEF TestVersion}
  Result := GTreeView_SortChildrenCB(Handle, SortCB, 0);
{$ELSE}
  Result := TreeView_SortChildrenCB(Handle, SortCB, 0);
{$ENDIF}
end;

procedure TTree2Node.Delete;
begin
  if not Deleting then Free;
end;

procedure TTree2Node.DeleteChildren;
var
  Node: TTree2Node;
Label skipit;
begin
  If GWDeletingChildren then goto skipit; { fix recursion via W95 control and }

  { GW 5-7-96 Temporarily set GWParentDeleting flag
    to speed up child Node.Destroy by skipping child's parent update }
  if (fOwner <> nil) then Inc(fOwner.GWParentDeleting);
  try
    {-----------------------------------
     5-14-96 original code can get hung up
    repeat
      Node := GetFirstChild;
      if Node <> nil then Node.Delete;
    until Node = nil;
    ------------------------------------}
    { GW 5-14-96 Infinite hangup if
      Node1.Delete deletes last child, precipitating...
        ParentNode.Collapse... in which OnCollapsing may call...
          DeleteChildren... which will be unable to delete Node1
                            due to Deleting flag.
      So rewrite to make sure that all deletable children get deleted
      ...presumably only one... but it's bad to ever get stuck here! }

    {-------- 5-14-96 Revamped deletechildren loop ----------}
    GWDeletingChildren := true;
    Node := GetFirstChild;
    While Node <> nil do
    Begin
      Node.Delete;
      Node := GetFirstChild;
      While (Node <> nil) and Node.Deleting do
        Node := GetNextChild(Node);
    end;
    {-------- 5-14-96 End Revamped deletechildren loop ----------}

  finally
    { GW 5-7-96 Now update self to compensate for skipped child
              Node.Destroy parent processing }
    if (fOwner <> nil) then Dec(fOwner.GWParentDeleting);
    Expanded := False;      { we get recursion here without GWDeletingChildren }
    HasChildren := False;
    GWDeletingChildren := false;
  end;
skipit:
end;

procedure TTree2Node.Assign(Source: TPersistent);
var
  Node: TTree2Node;
begin
  if Source is TTree2Node then
  begin
    Node := TTree2Node(Source);
    Text := Node.Text;
    Data := Node.Data;
    ImageIndex := Node.ImageIndex;
    SelectedIndex := Node.SelectedIndex;
    StateIndex := Node.StateIndex;
    OverlayIndex := Node.OverlayIndex;
    Focused := Node.Focused;
    DropTarget := Node.DropTarget;
    Cut := Node.Cut;
    HasChildren := Node.HasChildren;
  end
  else inherited Assign(Source);
end;

function TTree2Node.IsEqual(Node: TTree2Node): Boolean;
begin
  Result := (Text = Node.Text) and (Data = Node.Data);
end;

procedure TTree2Node.ReadData(Stream: TStream; Info: PNodeInfo);
var
  I, Size, ItemCount: Integer;
begin
  Stream.ReadBuffer(Size, SizeOf(Size));
  Stream.ReadBuffer(Info^, Size);
  Text := Info^.Text;
  ImageIndex := Info^.ImageIndex;
  SelectedIndex := Info^.SelectedIndex;
  StateIndex := Info^.StateIndex;
  OverlayIndex := Info^.OverlayIndex;
  Data := Info^.Data;
  ItemCount := Info^.Count;
  for I := 0 to ItemCount - 1 do
    with Owner.AddChild(Self, '') do ReadData(Stream, Info);
end;

procedure TTree2Node.WriteData(Stream: TStream; Info: PNodeInfo);
var
  I, Size, L, ItemCount: Integer;
begin
  L := Length(Text);
  if L > 255 then L := 255;
  Size := SizeOf(TNodeInfo) + L - 255;
  Info^.Text := Text;
  Info^.ImageIndex := ImageIndex;
  Info^.SelectedIndex := SelectedIndex;
  Info^.OverlayIndex := OverlayIndex;
  Info^.StateIndex := StateIndex;
  Info^.Data := Data;
  ItemCount := Count;
  Info^.Count := ItemCount;
  Stream.WriteBuffer(Size, SizeOf(Size));
  Stream.WriteBuffer(Info^, Size);
  for I := 0 to ItemCount - 1 do Item[I].WriteData(Stream, Info);
end;

{ TTree2Nodes }

constructor TTree2Nodes.Create(AOwner: TCustomTree2View);
begin
  inherited Create;
  FOwner := AOwner;
  { GWUpdating            := false;}
  GWParentDeleting        := 0;      { GW 5-7-96 }
end;

destructor TTree2Nodes.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TTree2Nodes.GetCount: Integer;
begin
{$IFDEF TestVersion}
  if Owner.HandleAllocated then Result := GTreeView_GetCount(Handle)
{$ELSE}
  if Owner.HandleAllocated then Result := TreeView_GetCount(Handle)
{$ENDIF}  
  else Result := 0;
end;

function TTree2Nodes.GetHandle: HWND;
begin
  Result := Owner.Handle;
end;

procedure TTree2Nodes.Delete(Node: TTree2Node);
begin
  if (Node.ItemId = nil) and Assigned(Owner.FOnDeletion) then
    Owner.FOnDeletion(Self, Node);
  Node.Delete;
end;

procedure TTree2Nodes.Clear;
begin
  if Owner.HandleAllocated then
  Begin
    try
      Inc(GWParentDeleting);
      {$IFDEF TestVersion}
      GTreeView_DeleteAllItems(Handle);
      {$ELSE}
      TreeView_DeleteAllItems(Handle);
      {$ENDIF}
    finally
      Dec(GWParentDeleting);
    end;
  end;
end;

function TTree2Nodes.AddChildFirst(Node: TTree2Node; const S: string): TTree2Node;
begin
  Result := AddChildObjectFirst(Node, S, nil);
end;

function TTree2Nodes.AddChildObjectFirst(Node: TTree2Node; const S: string;
  Ptr: Pointer): TTree2Node;
begin
  Result := InternalAddObject(Node, S, Ptr, taAddFirst);
end;

function TTree2Nodes.AddChild(Node: TTree2Node; const S: string): TTree2Node;
begin
  Result := AddChildObject(Node, S, nil);
end;

function TTree2Nodes.AddChildObject(Node: TTree2Node; const S: string; Ptr: Pointer): TTree2Node;
begin
  Result := InternalAddObject(Node, S, Ptr, taAdd);
end;

function TTree2Nodes.AddFirst(Node: TTree2Node; const S: string): TTree2Node;
begin
  Result := AddObjectFirst(Node, S, nil);
end;

function TTree2Nodes.AddObjectFirst(Node: TTree2Node; const S: string;
  Ptr: Pointer): TTree2Node;
begin
  if Node <> nil then Node := Node.Parent;
  Result := InternalAddObject(Node, S, Ptr, taAddFirst);
end;

function TTree2Nodes.Add(Node: TTree2Node; const S: string): TTree2Node;
begin  { node = nil means top level }
  Result := AddObject(Node, S, nil);
end;

procedure TTree2Nodes.Repaint(Node: TTree2Node);
var
  R: TRect;
begin
  while (Node <> nil) and not Node.IsVisible do Node := Node.Parent;
  if Node <> nil then
  begin
    R := Node.DisplayRect(False);
    InvalidateRect(Owner.Handle, @R, True);
  end;
end;

function TTree2Nodes.AddObject(Node: TTree2Node; const S: string;
  Ptr: Pointer): TTree2Node;
begin { add object last in list containing Node }
  if Node <> nil then Node := Node.Parent;
  Result := InternalAddObject(Node, S, Ptr, taAdd);
end;

function TTree2Nodes.Insert(Node: TTree2Node; const S: string): TTree2Node;
begin
  Result := InsertObject(Node, S, nil);
end;

procedure TTree2Nodes.AddedNode(NewNode: TTree2Node; GWParentNode: TTree2Node);
{ updates GWParentNode }
begin
  { ParentNode := NewNode.Parent }

  if GWParentNode <> nil then
  begin
    GWParentNode.HasChildren := True;
    If not GWParentNode.GWNodeUpdating then Repaint(GWParentNode);
  end;
end;

function TTree2Nodes.InsertObject(Node: TTree2Node; const S: string;
  Ptr: Pointer): TTree2Node;
  { insert after Node}
var
  Item, ItemId: HTreeItem;
  Parent: TTree2Node;
  AddMode: TAddMode;
begin
  Result := Owner.CreateNode;
  try
    Parent := nil;  { GW 5-7-96 }
    Item := nil;
    ItemId := nil;
    AddMode := taInsert;
    if Node <> nil then
    begin
      Parent := Node.Parent;
      if Parent <> nil then Item := Parent.ItemId;
      Node := Node.GetPrevSibling;
      if Node <> nil then ItemId := Node.ItemId
      else AddMode := taAddFirst;
    end;
    Result.Data := Ptr;
    Result.Text := S;
    Item := AddItem(Item, ItemId, CreateItem(Result), AddMode);
    if Item = nil then
      raise EOutOfResources.CreateRes(sInsertError);
    Result.FItemId := Item;
    AddedNode(Result, Parent { GW });
  except
    Result.Free;
    raise;
  end;
end;

function TTree2Nodes.InternalAddObject(Node: TTree2Node; const S: string;
  Ptr: Pointer; AddMode: TAddMode): TTree2Node;
  { add object as last child of Node, or at top level if Node = nil }
var
  Item: HTreeItem;
begin
  Result := Owner.CreateNode;
  try
    if Node <> nil then Item := Node.ItemId
    else Item := nil;
    Result.Data := Ptr;
    Result.Text := S;
    Item := AddItem(Item, nil, CreateItem(Result), AddMode);
    if Item = nil then
      raise EOutOfResources.CreateRes(sInsertError);
    Result.FItemId := Item;
    AddedNode(Result, Node);

    {------------------------------------------
      Sort logic GW 5-8-96
    -------------------------------------------}
    if owner.SortType in [stText, stBoth] then
    Begin
      If Node = nil then
      Begin { no parent, top level }
        If not {TTreeNodes.}GWItemsUpdating then owner.AlphaSort;
      end else
      Begin { sort the parent's children }
        if not Node.GWNodeUpdating then Node.AlphaSort;
      end;
    end;

  except
    Result.Free;
    raise;
  end;
end;

function TTree2Nodes.CreateItem(Node: TTree2Node): TTVItem;
begin
  with Result do
  begin
    mask := TVIF_TEXT or TVIF_PARAM or TVIF_IMAGE or TVIF_SELECTEDIMAGE;
    lParam := Longint(Node);
    pszText := LPSTR_TEXTCALLBACK;
    iImage := I_IMAGECALLBACK;
    iSelectedImage := I_IMAGECALLBACK;
  end;
  Node.GWItemized := true;
end;

function TTree2Nodes.AddItem(Parent, Target: HTreeItem;
  const Item: TTVItem; AddMode: TAddMode): HTreeItem;
var
  InsertStruct: TTVInsertStruct;
begin
  with InsertStruct do
  begin
    hParent := Parent;
    case AddMode of
      taAddFirst:
        hInsertAfter := TVI_FIRST;
      taAdd:
        hInsertAfter := TVI_LAST;
      taInsert:
        hInsertAfter := Target;
    end;
  end;
  InsertStruct.item := Item;
{$IFDEF TestVersion}
  Result := GTreeView_InsertItem(Handle, InsertStruct);
{$ELSE}
  Result := TreeView_InsertItem(Handle, InsertStruct);
{$ENDIF}
end;

function TTree2Nodes.GetFirstNode: TTree2Node;
begin
{$IFDEF TestVersion}
  Result := GetNode(GTreeView_GetRoot(Handle));
{$ELSE}
  Result := GetNode(TreeView_GetRoot(Handle));
{$ENDIF}
end;

function TTree2Nodes.GetNodeFromIndex(Index: Integer): TTree2Node;
begin
  (***
  If fOwner.GWDebugMemo <> nil then
    fOwner.GWDebugMemo.Lines.Append('GNFI '+inttostr(Index));
  ***)  
  Result := GetFirstNode;
  while (Index <> 0) and (Result <> nil) do
  begin
    Result := Result.GetNext;
    Dec(Index);
  end;
  if Result = nil then Tree2ViewError(sInvalidIndex);
end;

function TTree2Nodes.GetNode(ItemId: HTreeItem): TTree2Node;
var
  Item: TTVItem;
begin
  with Item do
  begin
    hItem := ItemId;
    mask := TVIF_PARAM;
  end;
{$IFDEF TestVersion}
  if GTreeView_GetItem(Handle, Item) then Result := TTree2Node(Item.lParam)
{$ELSE}
  if TreeView_GetItem(Handle, Item) then Result := TTree2Node(Item.lParam)
{$ENDIF}  
  else Result := nil;
end;

procedure TTree2Nodes.SetItem(Index: Integer; Value: TTree2Node);
begin
  GetNodeFromIndex(Index).Assign(Value);
end;

procedure TTree2Nodes.BeginUpdate;
begin
  if FUpdateCount = 0 then SetUpdateState(True);
  Inc(FUpdateCount);
end;

procedure TTree2Nodes.SetUpdateState(Updating: Boolean);
Var
  RefreshDone: Boolean;
begin
  SendMessage(Handle, WM_SETREDRAW, Ord(not Updating), 0);
  RefreshDone := false;

  { note, "if not Updating" logic assumes that SetUpdateState only
    called from BeginUpdate or EndUpdate, currently true.}
  if not Updating then { end of update. Entire block: GW 5-8-96  }
  Begin
    if GWEnableSpecialNodesUpdating and (owner <> nil) then
    Begin
      if owner.SortType in [stText, stBoth] then
      Begin
        owner.AlphaSort; {sort top level }
        RefreshDone := true;
      end;
    end;
    if not RefreshDone then Owner.Refresh;
  end;
end;

procedure TTree2Nodes.EndUpdate;
begin
  Dec(FUpdateCount);
  if FUpdateCount = 0 then SetUpdateState(False);
end;

{--------------------------------------------}
function TTree2Nodes.GWItemsUpdating: Boolean; { GW 5-8-96 }
{--------------------------------------------}
Begin
  Result := GWEnableSpecialNodesUpdating and (FUpdateCount > 0);
end;
{---------------------------------}
procedure TTree2Nodes.GWEndUpdateInternal;   { GW added 5-7-96 }
{---------------------------------}
Begin
  { GWUpdating := false; }
end;

{---------------------------------}
procedure TTree2Nodes.Assign(Source: TPersistent);
var
  Tree2Nodes: TTree2Nodes;
  MemStream: TMemoryStream;
begin
  if Source is TTree2Nodes then
  begin
    Tree2Nodes := TTree2Nodes(Source);
    Clear;
    MemStream := TMemoryStream.Create;
    try
      Tree2Nodes.WriteData(MemStream);
      MemStream.Position := 0;
      ReadData(MemStream);
    finally
      MemStream.Free;
    end;
  end
  else inherited Assign(Source);
end;

procedure TTree2Nodes.DefineProperties(Filer: TFiler);

  function WriteNodes: Boolean;
  var
    I: Integer;
    Nodes: TTree2Nodes;
  begin
    Nodes := TTree2Nodes(Filer.Ancestor);
    if (Nodes <> nil) and (Nodes.Count = Count) then
      for I := 0 to Count - 1 do
      begin
        Result := not Item[I].IsEqual(Nodes[I]);
        if Result then Break;
      end
    else Result := Count > 0;
  end;

begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadData, WriteData, WriteNodes);
end;

procedure TTree2Nodes.ReadData(Stream: TStream);
var
  I, Count: Integer;
  NodeInfo: TNodeInfo;
begin
  Clear;
  Stream.ReadBuffer(Count, SizeOf(Count));
  for I := 0 to Count - 1 do
    Add(nil, '').ReadData(Stream, @NodeInfo);
end;

procedure TTree2Nodes.WriteData(Stream: TStream);
var
  I: Integer;
  Node: TTree2Node;
  NodeInfo: TNodeInfo;
begin
  I := 0;
  Node := GetFirstNode;
  while Node <> nil do
  begin
    Inc(I);
    Node := Node.GetNextSibling;
  end;
  Stream.WriteBuffer(I, SizeOf(I));
  Node := GetFirstNode;
  while Node <> nil do
  begin
    Node.WriteData(Stream, @NodeInfo);
    Node := Node.GetNextSibling;
  end;
end;

type
  TTree2Strings = class(TStrings)
  private
    FOwner: TTree2Nodes;
  protected
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
    function GetObject(Index: Integer): TObject; override;
    procedure PutObject(Index: Integer; AObject: TObject); override;
    procedure SetUpdateState(Updating: Boolean); override;
  public
    constructor Create(AOwner: TTree2Nodes);
    function Add(const S: string): Integer; override;
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
    property Owner: TTree2Nodes read FOwner;
  end;

constructor TTree2Strings.Create(AOwner: TTree2Nodes);
begin
  inherited Create;
  FOwner := AOwner;
end;

function TTree2Strings.Get(Index: Integer): string;
const
  TAB = Chr(9);
var
  Level, I: Integer;
  Node: TTree2Node;
begin
  Result := '';
  Node := Owner.GetNodeFromIndex(Index);
  Level := Node.Level;
  for I := 0 to Level - 1 do Result := Result + TAB;
  Result := Result + Node.Text;
end;

function TTree2Strings.GetObject(Index: Integer): TObject;
begin
  Result := Owner.GetNodeFromIndex(Index).Data;
end;

procedure TTree2Strings.PutObject(Index: Integer; AObject: TObject);
begin
  Owner.GetNodeFromIndex(Index).Data := AObject;
end;

function TTree2Strings.GetCount: Integer;
begin
  Result := Owner.Count;
end;

procedure TTree2Strings.Clear;
begin
  Owner.Clear;
end;

procedure TTree2Strings.Delete(Index: Integer);
begin
  Owner.GetNodeFromIndex(Index).Delete;
end;

procedure TTree2Strings.SetUpdateState(Updating: Boolean);
begin
  SendMessage(Owner.Handle, WM_SETREDRAW, Ord(not Updating), 0);
  if not Updating then Owner.Owner.Refresh;
end;

function TTree2Strings.Add(const S: string): Integer;
var
  Level, OldLevel, I: Integer;
  NewStr: string;
  Node: TTree2Node;

  function GetBufStart(Buffer: PChar; var Level: Integer): PChar;
  begin
    Level := 0;
    while Buffer^ in [' ', #9] do
    begin
      Inc(Buffer);
      Inc(Level);
    end;
    Result := Buffer;
  end;

begin
  Result := GetCount;
  if (Length(S) = 1) and (S[1] = Chr($1A)) then Exit;
  Node := nil;
  OldLevel := 0;
  NewStr := GetBufStart(PChar(S), Level);
  if Result > 0 then
  begin
    Node := Owner.GetNodeFromIndex(Result - 1);
    OldLevel := Node.Level;
  end;
  if (Level > OldLevel) or (Node = nil) then
  begin
    if Level - OldLevel > 1 then Tree2ViewError(sInvalidLevel);
  end
  else begin
    for I := OldLevel downto Level do
    begin
      Node := Node.Parent;
      if (Node = nil) and (I - Level > 0) then
        Tree2ViewError(sInvalidLevel);
    end;
  end;
  Owner.AddChild(Node, NewStr);
end;

procedure TTree2Strings.Insert(Index: Integer; const S: string);
begin
  with Owner do
    Insert(GetNodeFromIndex(Index), S);
end;

{ TCustomTree2View }

constructor TCustomTree2View.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csCaptureMouse] + [csDisplayDragImage];
  Width := 121;
  Height := 97;
  TabStop := True;
  ParentColor := False;
  FTree2Nodes := TTree2Nodes.Create(Self);
  FBorderStyle := bsSingle;
  FShowButtons := True;
  FShowRoot := True;
  FShowLines := True;
  FHideSelection := True;
  FDragImage := TImageList.CreateSize(32, 32);
  FSaveIndent := -1;
  FEditInstance := MakeObjectInstance(EditWndProc);
  FImageChangeLink := TChangeLink.Create;
  FImageChangeLink.OnChange := ImageListChange;
  FStateChangeLink := TChangeLink.Create;
  FStateChangeLink.OnChange := ImageListChange;

  Tag := GWVerNum;  { GW, so we can identify version easily }
end;

destructor TCustomTree2View.Destroy;
begin
  Items.Free;
  FSaveItems.Free;
  FDragImage.Free;
  FMemStream.Free;
  FreeObjectInstance(FEditInstance);
  FImageChangeLink.Free;
  FStateChangeLink.Free;
  inherited Destroy;
end;

procedure TCustomTree2View.CreateParams(var Params: TCreateParams);
const
  BorderStyles: array[TBorderStyle] of Integer = (0, WS_BORDER);
  LineStyles: array[Boolean] of Integer = (0, TVS_HASLINES);
  RootStyles: array[Boolean] of Integer = (0, TVS_LINESATROOT);
  ButtonStyles: array[Boolean] of Integer = (0, TVS_HASBUTTONS);
  EditStyles: array[Boolean] of Integer = (TVS_EDITLABELS, 0);
  HideSelections: array[Boolean] of Integer = (TVS_SHOWSELALWAYS, 0);
  DragStyles: array[TDragMode] of Integer = (TVS_DISABLEDRAGDROP, 0);
begin
  InitCommonControls;
  inherited CreateParams(Params);
  CreateSubClass(Params, WC_TreeVIEW);
  with Params do
  begin
    Style := Style or LineStyles[FShowLines] or BorderStyles[FBorderStyle] or
      RootStyles[FShowRoot] or ButtonStyles[FShowButtons] or
      EditStyles[FReadOnly] or HideSelections[FHideSelection] or
      DragStyles[DragMode];
    if Ctl3D and (FBorderStyle = bsSingle) then
      ExStyle := Params.ExStyle or WS_EX_CLIENTEDGE;
  end;
end;

procedure TCustomTree2View.CreateWnd;
begin
  inherited CreateWnd;
  if FMemStream <> nil then
  begin
    Items.ReadData(FMemStream);
    FMemStream.Destroy;
    FMemStream := nil;
    SetTopItem(Items.GetNodeFromIndex(FSaveTopIndex));
    FSaveTopIndex := 0;
    SetSelection(Items.GetNodeFromIndex(FSaveIndex));
    FSaveIndex := 0;
  end;
  if FSaveIndent <> -1 then Indent := FSaveIndent;
  if (Images <> nil) and Images.HandleAllocated then
    SetImageList(Images.Handle, TVSIL_NORMAL);
  if (StateImages <> nil) and StateImages.HandleAllocated then
    SetImageList(StateImages.Handle, TVSIL_STATE);
end;

procedure TCustomTree2View.DestroyWnd;
var
  Node: TTree2Node;
begin
  if Items.Count > 0 then
  begin
    FMemStream := TMemoryStream.Create;
    Items.WriteData(FMemStream);
    FMemStream.Position := 0;
    Node := GetTopItem;
    if Node <> nil then FSaveTopIndex := Node.AbsoluteIndex;
    Node := Selected;
    if Node <> nil then FSaveIndex := Node.AbsoluteIndex;
  end;
  FSaveIndent := Indent;
  inherited DestroyWnd;
end;

procedure TCustomTree2View.EditWndProc(var Message: TMessage);
begin
  try
    with Message do
    begin
      case Msg of
        WM_KEYDOWN,
        WM_SYSKEYDOWN: if DoKeyDown(TWMKey(Message)) then Exit;
        WM_CHAR: if DoKeyPress(TWMKey(Message)) then Exit;
        WM_KEYUP,
        WM_SYSKEYUP: if DoKeyUp(TWMKey(Message)) then Exit;
        CN_KEYDOWN,
        CN_CHAR, CN_SYSKEYDOWN,
        CN_SYSCHAR:
          begin
            WndProc(Message);
            Exit;
          end;
      end;
      Result := CallWindowProc(FDefEditProc, FEditHandle, Msg, WParam, LParam);
    end;
  except
    Application.HandleException(Self);
  end;
end;

procedure TCustomTree2View.CMColorChanged(var Message: TMessage);
begin
  inherited;
  RecreateWnd;
end;

procedure TCustomTree2View.CMCtl3DChanged(var Message: TMessage);
begin
  inherited;
  if FBorderStyle = bsSingle then RecreateWnd;
end;

function TCustomTree2View.AlphaSort: Boolean;
var
  I: Integer;
begin
  if HandleAllocated then
  begin
    Result := CustomSort(nil, 0);
    (****** redundant GW 3-30-96 **************
    for I := 0 to Items.Count - 1 do
      with Items[I] do
        if HasChildren then AlphaSort;
    ********************************************)
  end
  else Result := False;
end;

function TCustomTree2View.CustomSort(SortProc: TTVCompare; Data: Longint): Boolean;
var
  SortCB: TTVSortCB;
  I: Integer;
  Node: TTree2Node;
begin
  (***
  if GWDebugMemo <> nil then
       GWDebugMemo.Lines.Append('< T2V.CustomSort >');  { GW 3-30-96 }
  ***)     
  Result := False;
  if HandleAllocated then
  begin
    with SortCB do
    begin
      if not Assigned(SortProc) then lpfnCompare := @DefaultTree2ViewSort
      else lpfnCompare := SortProc;
      hParent := TVI_ROOT;
      lParam := Data;
{$IFDEF TestVersion}
      Result := GTreeView_SortChildrenCB(Handle, SortCB, 0);
    {$ELSE}
      Result := TreeView_SortChildrenCB(Handle, SortCB, 0);
    {$ENDIF}
    end;

    (***** original code very slow ****************
    for I := 0 to Items.Count - 1 do
    begin
      Node := Items[I];
      if Node.HasChildren then Node.CustomSort(SortProc, Data);
    end;
    **********************************************)
    
    {------------------------------------------
    GW 5-7-96 Should be able to replace with the following
      Note: for some reason above algorithm sorts every
      node individually. Why not use TreeView_SortChildren
      with recurse = true?
      Anyhow, version below at least avoids slow
      Items[] (TTreeNodes.GetNodeFromIndex)
    -------------------------------------------}
    If Items.Count > 0 then
    Begin
      Node := Items[0];
      While Node <> nil do
      Begin
        if Node.HasChildren then Node.CustomSort(SortProc, Data);
        Node := Node.GetNext;
      end;
    end;
    (**************************)
  end;
end;
{
  if Result = nil then Tree2ViewError(sInvalidIndex);
}

procedure TCustomTree2View.SetSortType(Value: TSortType);
begin
  if SortType <> Value then
  begin
    FSortType := Value;
    if ((SortType in [stData, stBoth]) and Assigned(OnCompare)) or
      (SortType in [stText, stBoth]) then
      AlphaSort;
  end;
end;

procedure TCustomTree2View.SetStyle(Value: Integer; UseStyle: Boolean);
var
  Style: Integer;
begin
  if HandleAllocated then
  begin
    Style := GetWindowLong(Handle, GWL_STYLE);
    if not UseStyle then Style := Style and not Value
    else Style := Style or Value;
    SetWindowLong(Handle, GWL_STYLE, Style);
  end;
end;

procedure TCustomTree2View.SetBorderStyle(Value: TBorderStyle);
begin
  if BorderStyle <> Value then
  begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

procedure TCustomTree2View.SetDragMode(Value: TDragMode);
begin
  if Value <> DragMode then
    SetStyle(TVS_DISABLEDRAGDROP, Value = dmManual);
  inherited;
end;

procedure TCustomTree2View.SetButtonStyle(Value: Boolean);
begin
  if ShowButtons <> Value then
  begin
    FShowButtons := Value;
    SetStyle(TVS_HASBUTTONS, Value);
  end;
end;

procedure TCustomTree2View.SetLineStyle(Value: Boolean);
begin
  if ShowLines <> Value then
  begin
    FShowLines := Value;
    SetStyle(TVS_HASLINES, Value);
  end;
end;

procedure TCustomTree2View.SetRootStyle(Value: Boolean);
begin
  if ShowRoot <> Value then
  begin
    FShowRoot := Value;
    SetStyle(TVS_LINESATROOT, Value);
  end;
end;

procedure TCustomTree2View.SetReadOnly(Value: Boolean);
begin
  if ReadOnly <> Value then
  begin
    FReadOnly := Value;
    SetStyle(TVS_EDITLABELS, not Value);
  end;
end;

procedure TCustomTree2View.SetHideSelection(Value: Boolean);
begin
  if HideSelection <> Value then
  begin
    FHideSelection := Value;
    SetStyle(TVS_SHOWSELALWAYS, not Value);
  end;
end;

function TCustomTree2View.GetNodeAt(X, Y: Integer): TTree2Node;
var
  HitTest: TTVHitTestInfo;
begin
  with HitTest do
  begin
    pt.X := X;
    pt.Y := Y;
{$IFDEF TestVersion}
    if GTreeView_HitTest(Handle, HitTest) <> nil then
  {$ELSE}
    if TreeView_HitTest(Handle, HitTest) <> nil then
  {$ENDIF}  
      Result := Items.GetNode(HitTest.hItem)
    else Result := nil;
  end;
end;

function TCustomTree2View.GetHitTestInfoAt(X, Y: Integer): THitTests;
var
  HitTest: TTVHitTestInfo;
begin
  Result := [];
  with HitTest do
  begin
    pt.X := X;
    pt.Y := Y;
{$IFDEF TestVersion}
    GTreeView_HitTest(Handle, HitTest);
  {$ELSE}
    TreeView_HitTest(Handle, HitTest);
  {$ENDIF}
    if (flags and TVHT_ABOVE) <> 0 then Include(Result, htAbove);
    if (flags and TVHT_BELOW) <> 0 then Include(Result, htBelow);
    if (flags and TVHT_NOWHERE) <> 0 then Include(Result, htNowhere);
    if (flags and TVHT_ONITEM) <> 0 then Include(Result, htOnItem);
    if (flags and TVHT_ONITEMBUTTON) <> 0 then Include(Result, htOnButton);
    if (flags and TVHT_ONITEMICON) <> 0 then Include(Result, htOnIcon);
    if (flags and TVHT_ONITEMINDENT) <> 0 then Include(Result, htOnIndent);
    if (flags and TVHT_ONITEMLABEL) <> 0 then Include(Result, htOnLabel);
    if (flags and TVHT_ONITEMRIGHT) <> 0 then Include(Result, htOnRight);
    if (flags and TVHT_ONITEMSTATEICON) <> 0 then Include(Result, htOnStateIcon);
    if (flags and TVHT_TOLEFT) <> 0 then Include(Result, htToLeft);
    if (flags and TVHT_TORIGHT) <> 0 then Include(Result, htToRight);
  end;
end;

procedure TCustomTree2View.SetTree2Nodes(Value: TTree2Nodes);
begin
  Items.Assign(Value);
end;

procedure TCustomTree2View.SetIndent(Value: Integer);
begin
{$IFDEF TestVersion}
  if Value <> Indent then GTreeView_SetIndent(Handle, Value);
{$ELSE}
  if Value <> Indent then TreeView_SetIndent(Handle, Value);
{$ENDIF}  
end;

function TCustomTree2View.GetIndent: Integer;
begin
{$IFDEF TestVersion}
  Result := GTreeView_GetIndent(Handle)
{$ELSE}
  Result := TreeView_GetIndent(Handle)
{$ENDIF}  
end;

procedure TCustomTree2View.FullExpand;
var
  Node: TTree2Node;
begin
  Node := Items.GetFirstNode;
  while Node <> nil do
  begin
    Node.Expand(True);
    Node := Node.GetNextSibling;
  end;
end;

procedure TCustomTree2View.FullCollapse;
var
  Node: TTree2Node;
begin
  Node := Items.GetFirstNode;
  while Node <> nil do
  begin
    Node.Collapse(True);
    Node := Node.GetNextSibling;
  end;
end;

procedure TCustomTree2View.Loaded;
begin
  inherited Loaded;
  if csDesigning in ComponentState then FullExpand;
end;

function TCustomTree2View.GetTopItem: TTree2Node;
begin
  if HandleAllocated then
{$IFDEF TestVersion}
    Result := Items.GetNode(GTreeView_GetFirstVisible(Handle))
  {$ELSE}
    Result := Items.GetNode(TreeView_GetFirstVisible(Handle))
  {$ENDIF}  
  else Result := nil;
end;

procedure TCustomTree2View.SetTopItem(Value: TTree2Node);
begin
  if HandleAllocated and (Value <> nil) then
{$IFDEF TestVersion}
    GTreeView_SelectSetFirstVisible(Handle, Value.ItemId);
  {$ELSE}
    TreeView_SelectSetFirstVisible(Handle, Value.ItemId);
  {$ENDIF}  
end;

function TCustomTree2View.GetSelection: TTree2Node;
begin
  if HandleAllocated then
{$IFDEF TestVersion}
    Result := Items.GetNode(GTreeView_GetSelection(Handle))
  {$ELSE}
    Result := Items.GetNode(TreeView_GetSelection(Handle))
  {$ENDIF}  
  else Result := nil;
end;

procedure TCustomTree2View.SetSelection(Value: TTree2Node);
begin
  if Value <> nil then Value.Selected := True
{$IFDEF TestVersion}
  else GTreeView_SelectItem(Handle, nil);
{$ELSE}
  else TreeView_SelectItem(Handle, nil);
{$ENDIF}  
end;

function TCustomTree2View.GetDropTarget: TTree2Node;
begin
  if HandleAllocated then
  begin
{$IFDEF TestVersion}
    Result := Items.GetNode(GTreeView_GetDropHilite(Handle));
  {$ELSE}
    Result := Items.GetNode(TreeView_GetDropHilite(Handle));
  {$ENDIF}  
    if Result = nil then Result := FLastDropTarget;
  end
  else Result := nil;
end;

procedure TCustomTree2View.SetDropTarget(Value: TTree2Node);
begin
  if HandleAllocated then
    if Value <> nil then Value.DropTarget := True
{$IFDEF TestVersion}
    else GTreeView_SelectDropTarget(Handle, nil);
  {$ELSE}
    else TreeView_SelectDropTarget(Handle, nil);
  {$ENDIF}  
end;

function TCustomTree2View.GetNodeFromItem(const Item: TTVItem): TTree2Node;
begin
  with Item do
    if (state and TVIF_PARAM) <> 0 then Result := Pointer(lParam)
    else Result := Items.GetNode(hItem);
end;

function TCustomTree2View.IsEditing: Boolean;
begin
{$IFDEF TestVersion}
  Result := GTreeView_GetEditControl(Handle) <> 0;
{$ELSE}
  Result := TreeView_GetEditControl(Handle) <> 0;
{$ENDIF}  
end;

procedure TCustomTree2View.CNNotify(var Message: TWMNotify);
var
  Node: TTree2Node;
begin
  with Message.NMHdr^ do
  Begin
    If GWDebugMemo <> nil then
    GWDebugMemo.Lines.Append('CNNotify '+IntToStr(code));

    case code of
      TVN_BEGINDRAG:
        begin
          FDragged := True;
          with PNMTreeView(Pointer(Message.NMHdr))^ do
            FDragNode := GetNodeFromItem(ItemNew);
        end;
      TVN_BEGINLABELEDIT:
        begin
          with PTVDispInfo(Pointer(Message.NMHdr))^ do
            if Dragging or not CanEdit(GetNodeFromItem(item)) then
              Message.Result := 1;
          if Message.Result = 0 then
          begin
{$IFDEF TestVersion}
            FEditHandle := GTreeView_GetEditControl(Handle);
          {$ELSE}
            FEditHandle := TreeView_GetEditControl(Handle);
          {$ENDIF}
            FDefEditProc := Pointer(GetWindowLong(FEditHandle, GWL_WNDPROC));
            SetWindowLong(FEditHandle, GWL_WNDPROC, LongInt(FEditInstance));
          end;
        end;
      TVN_ENDLABELEDIT:
        with PTVDispInfo(Pointer(Message.NMHdr))^ do
          Edit(item);
      TVN_ITEMEXPANDING:
        with PNMTreeView(Pointer(Message.NMHdr))^ do
        begin
          Node := GetNodeFromItem(ItemNew);

          If GWProperExpandCollapseBehaviorFix then
          Begin {-------------------------- fixed code }
            if (action = TVE_EXPAND) and (Node.GWProgramExpand = 0) then
            Begin
              if not CanExpand(Node) then Message.Result := 1;
            end;

            if (action = TVE_COLLAPSE) and (Node.GWProgramCollapse = 0) then
            Begin
              if not CanCollapse(Node) then Message.Result := 1;
            end;
            { otherwise... OK to proceed, Result := 0, which is default }
          end else
          Begin {--------------------- do original code }
            if (action = TVE_EXPAND) and not CanExpand(Node) then
              Message.Result := 1
            else if (action = TVE_COLLAPSE) and
              not CanCollapse(Node) then Message.Result := 1;
          end;
        end;
      TVN_ITEMEXPANDED:
        with PNMTreeView(Pointer(Message.NMHdr))^ do
        begin
          Node := GetNodeFromItem(itemNew);
          If GWProperExpandCollapseBehaviorFix  then
          Begin { GW 5-12-96 }
            if (action = TVE_EXPAND) and (Node.GWProgramExpand = 0)
              then Expand(Node); { else do it later in ExpandItem }

            if (action = TVE_COLLAPSE) and (Node.GWProgramCollapse = 0)
              then Collapse(Node); { else do it later in ExpandItem }
          end else
          Begin { original code for normal manual expand/collapse }
            if (action = TVE_EXPAND) then Expand(Node)
            else if (action = TVE_COLLAPSE) then Collapse(Node);
          end;
        end;
      TVN_SELCHANGING:
        with PNMTreeView(Pointer(Message.NMHdr))^ do
          if not CanChange(GetNodeFromItem(itemNew)) then
            Message.Result := 1;
      TVN_SELCHANGED:
        with PNMTreeView(Pointer(Message.NMHdr))^ do
          Change(GetNodeFromItem(itemNew));
      TVN_DELETEITEM:
        begin
          with PNMTreeView(Pointer(Message.NMHdr))^ do
            Node := GetNodeFromItem(itemOld);
          if Node <> nil then
          begin
            Node.FItemId := nil;
            Items.Delete(Node);
          end;
        end;
      TVN_SETDISPINFO:
        with PTVDispInfo(Pointer(Message.NMHdr))^ do
        begin
          Node := GetNodeFromItem(item);
          if (Node <> nil) and ((item.mask and TVIF_TEXT) <> 0) then
            Node.Text := item.pszText;
        end;
      TVN_GETDISPINFO:
        with PTVDispInfo(Pointer(Message.NMHdr))^ do
        begin
          Node := GetNodeFromItem(item);
          if Node <> nil then
          begin
            if (item.mask and TVIF_TEXT) <> 0 then
              StrLCopy(item.pszText, PChar(Node.Text), item.cchTextMax);
            if (item.mask and TVIF_IMAGE) <> 0 then
            begin
              GetImageIndex(Node);
              item.iImage := Node.ImageIndex;
            end;
            if (item.mask and TVIF_SELECTEDIMAGE) <> 0 then
            begin
              GetSelectedIndex(Node);
              item.iSelectedImage := Node.SelectedIndex;
            end;
          end;
        end;
      NM_RCLICK: FRClicked := True;
    end;
  end;
end;

function TCustomTree2View.GetDragImages: TCustomImageList;
begin
  if FDragImage.Count > 0 then
    Result := FDragImage else
    Result := nil;
end;

procedure TCustomTree2View.WndProc(var Message: TMessage);
begin
  if not (csDesigning in ComponentState) and ((Message.Msg = WM_LBUTTONDOWN) or
    (Message.Msg = WM_LBUTTONDBLCLK)) and not Dragging and (DragMode = dmAutomatic) then
  begin
    if not IsControlMouseMsg(TWMMouse(Message)) then
    begin
      ControlState := ControlState + [csLButtonDown];
      Dispatch(Message);
    end;
  end
  else inherited WndProc(Message);
end;

function  TCustomTree2View.GWVerNum: integer; { GW 5-14-96 }
Begin
  Result := T2VVerNum;
end;

procedure TCustomTree2View.DoStartDrag(var DragObject: TDragObject);
var
  ImageHandle: HImageList;
  DragNode: TTree2Node;
  P: TPoint;
begin
  inherited DoStartDrag(DragObject);
  DragNode := FDragNode;
  FLastDropTarget := nil;
  FDragNode := nil;
  if DragNode = nil then
  begin
    GetCursorPos(P);
    with ScreenToClient(P) do DragNode := GetNodeAt(X, Y);
  end;
  if DragNode <> nil then
  begin
{$IFDEF TestVersion}
    ImageHandle := GTreeView_CreateDragImage(Handle, DragNode.ItemId);
  {$ELSE}
    ImageHandle := TreeView_CreateDragImage(Handle, DragNode.ItemId);
  {$ENDIF}  
    if ImageHandle <> 0 then
      with FDragImage do
      begin
        Handle := ImageHandle;
        SetDragImage(0, 2, 2);
      end;
  end;
end;

procedure TCustomTree2View.DoEndDrag(Target: TObject; X, Y: Integer);
begin
  inherited DoEndDrag(Target, X, Y);
  FLastDropTarget := nil;
end;

procedure TCustomTree2View.CMDrag(var Message: TCMDrag);
begin
  inherited;
  if Message.Result <> 0 then
    with Message, DragRec^ do
      case DragMessage of
        dmDragMove: with ScreenToClient(Pos) do DoDragOver(Source, X, Y);
        dmDragLeave:
          begin
            TDragObject(Source).HideDragImage;
            FLastDropTarget := DropTarget;
            DropTarget := nil;
            TDragObject(Source).ShowDragImage;
          end;
        dmDragDrop: FLastDropTarget := nil;
      end;
end;

procedure TCustomTree2View.DoDragOver(Source: TDragObject; X, Y: Integer);
var
  Node: TTree2Node;
begin
  Node := GetNodeAt(X, Y);
  if (Node <> nil) and
    ((Node <> DropTarget) or (Node = FLastDropTarget)) then
  begin
    FLastDropTarget := nil;
    TDragObject(Source).HideDragImage;
    Node.DropTarget := True;
    TDragObject(Source).ShowDragImage;
  end;
end;

procedure TCustomTree2View.GetImageIndex(Node: TTree2Node);
begin
  if Assigned(FOnGetImageIndex) then FOnGetImageIndex(Self, Node);
end;

procedure TCustomTree2View.GetSelectedIndex(Node: TTree2Node);
begin
  if Assigned(FOnGetSelectedIndex) then FOnGetSelectedIndex(Self, Node);
end;

function TCustomTree2View.CanChange(Node: TTree2Node): Boolean;
begin
  Result := True;
  if Assigned(FOnChanging) then FOnChanging(Self, Node, Result);
end;

procedure TCustomTree2View.Change(Node: TTree2Node);
begin
  if Assigned(FOnChange) then FOnChange(Self, Node);
end;

procedure TCustomTree2View.Expand(Node: TTree2Node);
begin
  if Assigned(FOnExpanded) then FOnExpanded(Self, Node);
end;

function TCustomTree2View.CanExpand(Node: TTree2Node): Boolean;
begin
  Result := True;
  if Assigned(FOnExpanding) then FOnExpanding(Self, Node, Result);
end;

procedure TCustomTree2View.Collapse(Node: TTree2Node);
begin
  if Assigned(FOnCollapsed) then FOnCollapsed(Self, Node);
end;

function TCustomTree2View.CanCollapse(Node: TTree2Node): Boolean;
begin
  Result := True;
  if Assigned(FOnCollapsing) then FOnCollapsing(Self, Node, Result);
end;

function TCustomTree2View.CanEdit(Node: TTree2Node): Boolean;
begin
  Result := True;
  if Assigned(FOnEditing) then FOnEditing(Self, Node, Result);
end;

procedure TCustomTree2View.Edit(const Item: TTVItem);
var
  S: string;
  Node: TTree2Node;
begin
  with Item do
    if pszText <> nil then
    begin
      S := pszText;
      Node := GetNodeFromItem(Item);
      if Assigned(FOnEdited) then FOnEdited(Self, Node, S);
      if Node <> nil then Node.Text := S;
    end;
end;

function TCustomTree2View.CreateNode: TTree2Node;
begin
  Result := TTree2Node.Create(Items);
end;

procedure TCustomTree2View.SetImageList(Value: HImageList; Flags: Integer);
begin
{$IFDEF TestVersion}
  if HandleAllocated then GTreeView_SetImageList(Handle, Value, Flags);
{$ELSE}
  if HandleAllocated then TreeView_SetImageList(Handle, Value, Flags);
{$ENDIF}
end;

procedure TCustomTree2View.ImageListChange(Sender: TObject);
var
  ImageHandle: HImageList;
begin
  if HandleAllocated then
  begin
    ImageHandle := TImageList(Sender).Handle;
    if Sender = Images then
      SetImageList(ImageHandle, TVSIL_NORMAL)
    else if Sender = StateImages then
      SetImageList(ImageHandle, TVSIL_STATE);
  end;
end;

procedure TCustomTree2View.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = Images then Images := nil;
    if AComponent = StateImages then StateImages := nil;
  end;
end;

procedure TCustomTree2View.SetImages(Value: TImageList);
begin
  if Images <> nil then
    Images.UnRegisterChanges(FImageChangeLink);
  FImages := Value;
  if Images <> nil then
  begin
    Images.RegisterChanges(FImageChangeLink);
    SetImageList(Images.Handle, TVSIL_NORMAL)
  end
  else SetImageList(0, TVSIL_NORMAL);
end;

procedure TCustomTree2View.SetStateImages(Value: TImageList);
begin
  if StateImages <> nil then
    StateImages.UnRegisterChanges(FStateChangeLink);
  FStateImages := Value;
  if StateImages <> nil then
  begin
    StateImages.RegisterChanges(FStateChangeLink);
    SetImageList(StateImages.Handle, TVSIL_STATE)
  end
  else SetImageList(0, TVSIL_STATE);
end;

procedure TCustomTree2View.LoadFromFile(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TCustomTree2View.LoadFromStream(Stream: TStream);
begin
  with TTree2Strings.Create(Items) do
    try
      LoadFromStream(Stream);
    finally
      Free;
  end;
end;

procedure TCustomTree2View.SaveToFile(const FileName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TCustomTree2View.SaveToStream(Stream: TStream);
begin
  with TTree2Strings.Create(Items) do
    try
      SaveToStream(Stream);
    finally
      Free;
  end;
end;

procedure TCustomTree2View.WMRButtonDown(var Message: TWMRButtonDown);
var
  MousePos: TPoint;
begin
  FRClicked := False;
  inherited;
  if FRClicked then
  begin
    GetCursorPos(MousePos);
    with PointToSmallPoint(ScreenToClient(MousePos)) do
      Perform(WM_RBUTTONUP, 0, MakeLong(X, Y));
  end;
end;

procedure TCustomTree2View.WMLButtonDown(var Message: TWMLButtonDown);
var
  Node: TTree2Node;
  MousePos: TPoint;
begin
  FDragged := False;
  FDragNode := nil;
  try
    inherited;
    if DragMode = dmAutomatic then
    begin
      SetFocus;
      if not FDragged then
      begin
        GetCursorPos(MousePos);
        with PointToSmallPoint(ScreenToClient(MousePos)) do
          Perform(WM_LBUTTONUP, 0, MakeLong(X, Y));
      end
      else begin
        Node := GetNodeAt(Message.XPos, Message.YPos);
        if Node <> nil then
        begin
          Node.Focused := True;
          Node.Selected := True;
          BeginDrag(False);
        end;
      end;
    end;
  finally
    FDragNode := nil;
  end;
end;

{ TTreeNodesProperty }
type
  TTree2NodesProperty = class(TClassProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

procedure TTree2NodesProperty.Edit;
begin
  if EditTreeViewItems(TTreeNodes(GetOrdValue)) then Modified;
end;

function TTree2NodesProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paReadOnly];
end;

{ Registration }
procedure Register;
begin
  RegisterComponents(LoadStr(srWin95), [TTree2View]);
  RegisterPropertyEditor(TypeInfo(TTree2Nodes), nil, '', TTree2NodesProperty);
end;

end.
